/**
 *  \file       acmds/acut_context.c
 *  \brief      \$Id: acut_context.c 1182 2009-05-15 13:59:20Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1182 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

/** \addtogroup acut
 * \{
 */

#include "acut_context.h"
#include "debug.h"
#include <stdlib.h>
#include <string.h>

#define MSGQUE msgque

/*****************************************************************************/
/*                                                                           */
/*                               Request Handler                             */
/*                                                                           */
/*****************************************************************************/

static void CutHelp ( const char * ) __attribute__ ((noreturn));

/// \brief display help using \b -h or \b --help command-line option
/// \param base the executable usually: <tt>basename(argv[0])</tt>
static void
CutHelp ( const char * base  )
{
    fprintf(stderr, "usage: %s [OPTION]... [ARGUMENT]...\n", base);
    fputs("\n", stderr);
    fprintf(stderr, "  The %s tool works as a filter between a msgque client and a msgque server\n", base);
    fputs("  using the option '-f' to extract columns.\n", stderr);
    fputs("\n", stderr);
    fputs(MqHelp (base), stderr);
    fputs("\n", stderr);
    fprintf(stderr, "  %s [OPTION]:\n", base);
    fputs("    -f NUM...        the column number(s)  (default: 'empty'\n", stderr);
    fputs("      NUM            a single column with '0' < NUM < '255'\n", stderr);
    fputs("      NUM1,NUM2      multiple columns, use ',' as delimiter\n", stderr);
    fputs("      NUM1-NUM2      multiple columns, from NUM1 up to NUM2\n", stderr);
    fputs("      NUM-           multiple columns, from NUM up to the end\n", stderr);
    fputs("      -NUM           multiple columns, from the beginn up to NUM\n", stderr);
    fputs("    -h, --help       print this help\n", stderr);
    fputs("\n", stderr);

    exit(EXIT_SUCCESS);
}

/// \tool_FTR
static enum MqErrorE
CutFilter (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqBufferS * buf;
  struct MqReadS * const read = msgque->read;
  struct MqSendS * const send = msgque->send;
  struct CutCtxS * const context = ((struct CutCtxS*)msgque->config.myCtx);
  MQ_SIZE const end = MqReadGetNumItems(read);
  MQ_INT i;
  
  MqSendSTART (send);
  for (i=0; i<end ; i++) {
    MqErrorCheck (MqReadU (read, &buf));
    if (context->fields[i]) {
	MqSendU (send, buf);
    }
  }
  MqErrorCheck (MqSendFTR (send, MQ_TIMEOUT10));
  return MQ_OK;

error:
  return MqErrorStack (msgque->error);
}

/*****************************************************************************/
/*                                                                           */
/*                                context_init                               */
/*                                                                           */
/*****************************************************************************/

/// \tool_Delete
static void
CutDelete (
  struct MqS ** msgqueP
)
{
  if (!(*msgqueP))
    return;
  free ((*msgqueP)->config.myCtx);
  MqDelete (msgqueP);
}

/// \tool_Create
enum MqErrorE
CutCreate (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS ** msgqueP
)
{
  MQ_STR fieldsC, end, start;
  MQ_INT index, last=-1;
  struct CutCtxS *context = MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (struct CutCtxS));
  struct MqS * msgque;
  struct MqErrorS * error;

  // add my config data
  config->myCtx = (struct MqContextS *) context;
  config->fHelp = CutHelp;
  config->fFTR  = CutFilter;
  config->isFilter = MQ_YES;
  config->isServer = MQ_YES;
  config->fParentCreate = CutCreate;
  config->fParentDelete = CutDelete;

  // create the Message-Queue
  MqErrorCheck (MqCreate (config, argv, alfa, msgqueP));
  msgque = (*msgqueP);
  error = msgque->error;

  // get and parse the fields
  fieldsC=NULL;
  MqBufferLCheckOptionC (error, argv, "-f", &fieldsC);
  if (!fieldsC) {
    MqErrorC (error, __func__, 1, "option argument '-f' is required");
    goto error;
  }

  for (start=fieldsC; *start ; start=end) {
    index = (MQ_INT) strtol (start, &end, 10);
    if (index > 255) {
	MqErrorV (error, __func__, 1, 
	    "field index '%i' is larger than the maximum allowed index '255'", index);
	goto error;
    }
    if (start == end) 
	index = 255;
    if (index < 0) {
	last = 0;
	index = (index * -1);
    }
    if (*end == ',' || *end == '\0') {
	if (*end == ',') end++;
	if (last == -1) {
	    context->fields[index] = 1;
	} else {
	    for (; last <= index; last++) {
		context->fields[last] = 1;
	    }
	    last = -1;
	}
    } else if (*end == '-') {
	if (last != -1) {
	  MqErrorC (error, __func__, 1, 
	      "invalid '-' character found, use #-#, -# or #- instead");
	  goto error;
	}
	end++;
	if (*end) {
	    last = index;
	} else {
	    for (; index <= 255; index++) {
		context->fields[index] = 1;
	    }
	}
    } else {
	MqErrorV (error, __func__, 1, 
	    "invalid field character '%c' found, only 'NUMBER', ',' or '-' are allowed", *end);
	goto error;
    }
  }

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(*msgqueP, &argv));

  // everything is OK
  return MQ_OK;

error:
  MqErrorStack ((*msgqueP)->error);
  return MqErrorSAppendC((*msgqueP)->error, "use '-h' or '--help' for usage");
}

/** \} acut */

