/**
 *  \file       acmds/asort_context.c
 *  \brief      \$Id: asort_context.c 1182 2009-05-15 13:59:20Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1182 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

/** \addtogroup asort
 * \{
 */

#include "asort_context.h"
#include "optionL.h"
#include "debug.h"
#include <errno.h>
#include <stdlib.h>
#include <string.h>

/// link to the MqErrorS object
#define MQ_ERROR_S context->msgque->error
#define MSGQUE context->msgque

/*****************************************************************************/
/*                                                                           */
/*                               Request Handler                             */
/*                                                                           */
/*****************************************************************************/

static void SortHelp ( const char * ) __attribute__ ((noreturn));

/// \brief display help using \b -h or \b --help command-line option
/// \param base the exesortable usually: <tt>basename(argv[0])</tt>
static void
SortHelp ( const char * base  )
{
    fprintf(stderr, "usage: %s [OPTION]... [ARGUMENT]...\n", base);
    fputs("\n", stderr);
    fprintf(stderr, "  The %s tool works as a filter between a msgque client and a msgque server\n", base);
    fputs("  using the option '-# TYPE' to identify a column as sort index.\n", stderr);
    fputs("\n", stderr);
    fputs(MqHelp (base), stderr);
    fputs("\n", stderr);
    fprintf(stderr, "  %s [OPTION]:\n", base);
    fputs("    -NUM TYPE        the column number(s)  (default: 'empty'\n", stderr);
    fputs("      NUM            a single column with '0' < NUM < '255'\n", stderr);
    fputs("      TYPE           the sort type of the column C,I,W,D\n", stderr);
    fputs("       C             sort as string\n", stderr);
    fputs("       I,W           sort as I)nt or W)ide integer\n", stderr);
    fputs("       D             sort as double precision decimal\n", stderr);
    fputs("    -h, --help       print this help\n", stderr);
    fputs("\n", stderr);

    exit(EXIT_SUCCESS);
}

/// \brief compare function for qsort (man qsort)
/// \param left left site of the comparison
/// \param right right site of the comparison
/// \return \b <0 for smaller, \b 0 for equal or >0 for larger
static int
SortComp (
    const void *left,
    const void *right
)
{
  struct MqBufferS * l = ((struct SortKeyS *) left)->key;
  struct MqBufferS * r = ((struct SortKeyS *) right)->key;

  // get the right 'compare' function

  switch (l->type) {
    case MQ_STRT: return strcmp (l->cur.C, r->cur.C);
    case MQ_INTT: return (MqBufU2INT(l->cur) > MqBufU2INT(r->cur) ? 1 : MqBufU2INT(l->cur) < MqBufU2INT(r->cur) ? -1 : 0);
    case MQ_WIDT: return (MqBufU2WID(l->cur) > MqBufU2WID(r->cur) ? 1 : MqBufU2WID(l->cur) < MqBufU2WID(r->cur) ? -1 : 0);
    case MQ_DBLT: return (MqBufU2DBL(l->cur) > MqBufU2DBL(r->cur) ? 1 : MqBufU2DBL(l->cur) < MqBufU2DBL(r->cur) ? -1 : 0);
    default: return 0;
  }
  return 0;
}

/// \brief service-handler for the system token \b _EOF used in filters after
///   the last line received
/// \msgque
/// \param data not used, allways \b NULL
/// \retMqErrorE
static enum MqErrorE
SortEOF (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct SortCtxS * context = (struct SortCtxS *)msgque->config.myCtx;
  struct SortKeyS * start = context->sort;
  struct SortKeyS * end   = context->sort + context->sort_cursize;
  struct MqSendS * const send = msgque->send;

  // sort the data
  qsort(context->sort, context->sort_cursize, sizeof(struct SortKeyS), SortComp);

  for (; start < end; start++) {
    MqSendSTART (send);
    MqSendL (send, start->line);
    MqErrorCheck (MqSendFTR (send, MQ_TIMEOUT10));
    MqBufferLDelete (&start->line);
    start->key = NULL;
  }

  // cleanup after sending
  context->sort_cursize = 0;

  return MQ_OK;

error: 
  return MqErrorStack (MQ_ERROR_S);
}

/// \tool_FTR
static enum MqErrorE
SortFilter (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct SortCtxS * context = (struct SortCtxS *)msgque->config.myCtx;
  struct MqReadS * const read = msgque->read;

  struct MqBufferLS * line;
  MQ_SIZE sort_size = context->sort_size;
  MQ_SIZE sort_cursize = context->sort_cursize;
  struct SortKeyS * sortkey;
  struct MqBufferS * key;

  // create the list of all items in MqReadS
  MqErrorCheck (MqReadL (read, &line));

  // check if sort key is large enough
  if (sort_cursize >= sort_size) {
    sort_size = (sort_size+sort_size*1/3);
    context->sort = MqSysRealloc (msgque->error, context->sort, sort_size * sizeof (struct SortKeyS));
    context->sort_size = sort_size;
  }

  // add the idx'th column to the search key
  if (context->idx > line->cursize) {
    return MqErrorV (msgque->error, __func__, 1, 
	"can not get column '" MQ_FORMAT_Z "' from input line-number '" 
	    MQ_FORMAT_Z "'", context->idx, sort_cursize );
  }

  // save the current line into the sort key
  sortkey = &context->sort[sort_cursize];
  key = line->data[context->idx];

  // save and cast the key into the given type
  sortkey->key = key;
  sortkey->line = line;
  if (key->type != context->type) {
    MqErrorCheck( MqBufferCastTo (msgque->error, key, context->type));
  }
  context->sort_cursize++;

  return MQ_OK;

error:
  return MqErrorStack (msgque->error);
}

/*****************************************************************************/
/*                                                                           */
/*                                context_init                               */
/*                                                                           */
/*****************************************************************************/

/// \tool_Delete
static void
SortDelete (
  struct MqS ** const msgqueP
)
{
  struct SortCtxS *context;
  if (!(*msgqueP))
    return;
  context = (struct SortCtxS *) (*msgqueP)->config.myCtx;
  MqSysFree (context->sort);
  MqSysFree (context);
  MqDelete (msgqueP);
}

/// \tool_Create
enum MqErrorE
SortCreate (
  struct MqConfigS * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS ** msgqueP
)
{
  struct SortCtxS * context = MqSysCalloc (NULL, 1, sizeof (struct SortCtxS));
  struct MqErrorS * error;
  OptLSP sortcols;
  MQ_STR key;
  char type;

  // add my config data
  config->myCtx = (struct MqContextS *) context;
  config->fHelp = SortHelp;
  config->fFTR  = SortFilter;
  config->fEOF  = SortEOF;
  config->isFilter = MQ_YES;
  config->isServer = MQ_YES;
  config->fParentCreate = SortCreate;
  config->fParentDelete = SortDelete;

  // create the Message-Queue
  MqErrorCheck (MqCreate (config, argv, alfa, msgqueP));
  error = (*msgqueP)->error;

  // initialize the sort index
  context->sort = (struct SortKeyS *) MqSysMalloc (error, 10 * sizeof(struct SortKeyS));
  context->sort_size = 10;

  // check for the sort key
  MqErrorCheck ( OptLCreate (error, argv, &sortcols));
  if (sortcols->cursize != 1) {
    OptLDelete (&sortcols);
    MqErrorC (error, __func__, -1,
		"expect exact 'one' sort-column of format '-NUM TYPE'");
    goto error;
  }

  MqBufferGetC(error, sortcols->data[0].val,&key);
  type = *key;

  if (strlen (key) > 1 || ( type != 'C' && type != 'I' && type != 'W' && type != 'D')) {
    MqErrorV (error, __func__, 1, 
		"invalid type '%s', expect 'C,I,W or D'", key);
    // !attention! the !key! frum upper code is part of the "sortcols" memory
    OptLDelete (&sortcols);
    goto error;
  }

  context->idx = sortcols->data[0].idx;

  switch (type) {
    case 'C': context->type = MQ_STRT;	break;
    case 'I': context->type = MQ_INTT;	break;
    case 'W': context->type = MQ_WIDT;	break;
    case 'D': context->type = MQ_DBLT;	break;
  }

  // cleanup
  OptLDelete (&sortcols);

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(*msgqueP, &argv));

  return MQ_OK;

error:
  MqErrorStack ((*msgqueP)->error);
  return MqErrorSAppendC((*msgqueP)->error, "use '-h' or '--help' for usage");
}

/** \} asort */

