/**
 *  \file       acmds/asplit.c
 *  \brief      \$Id: asplit.c 1182 2009-05-15 13:59:20Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1182 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

/** \ingroup tools
 *  \defgroup asplit asplit
 *  \{
 *  \brief asplit \asplit_desc
 *
 *  \verbinclude asplit.help
 */

/// use the GNU redline extension
#define _GNU_SOURCE
/// solaris needs this for f(un)lockfile
#define __EXTENSIONS__

#include "msgque.h"
#include "config.h"
#include "debug.h"

#include <stdlib.h>
#include <string.h>

/*****************************************************************************/
/*                                                                           */
/*                               getline function                            */
/*                                                                           */
/*****************************************************************************/

#if !HAVE_GETLINE
#ifndef DOXYGEN_SHOULD_SKIP_THIS

#include <errno.h>
#include <stdio.h>

#if defined(_MSC_VER)
# define size_t unsigned int
# define ssize_t unsigned int
#endif

#ifndef SIZE_MAX
# define SIZE_MAX ((size_t) -1)
#endif
#ifndef SSIZE_MAX
# define SSIZE_MAX ((ssize_t) (SIZE_MAX / 2))
#endif
#if !HAVE_FLOCKFILE
# undef flockfile
# define flockfile(x) ((void) 0)
#endif
#if !HAVE_FUNLOCKFILE
# undef funlockfile
# define funlockfile(x) ((void) 0)
#endif

/* Read up to (and including) a DELIMITER from FP into *LINEPTR (and
   NUL-terminate it).  *LINEPTR is a pointer returned from malloc (or
   NULL), pointing to *N characters of space.  It is realloc'ed as
   necessary.  Returns the number of characters read (not including
   the null terminator), or -1 on error or EOF.  */

static ssize_t
getline (char **lineptr, size_t *n, FILE *fp)
{
  int delimiter = '\n';
  ssize_t result = -1;
  size_t cur_len = 0;

  if (lineptr == NULL || n == NULL || fp == NULL)
    {
      errno = EINVAL;
      return -1;
    }

  flockfile (fp);

  if (*lineptr == NULL || *n == 0)
    {
      *n = 120;
      *lineptr = (char *) malloc (*n);
      if (*lineptr == NULL)
    {
      result = -1;
      goto unlock_return;
    }
    }

  for (;;)
    {
      int i;

      i = getc (fp);
      if (i == EOF)
    {
      result = -1;
      break;
    }

      /* Make enough space for len+1 (for final NUL) bytes.  */
      if (cur_len + 1 >= *n)
    {
      size_t needed_max =
        SSIZE_MAX < SIZE_MAX ? (size_t) SSIZE_MAX + 1 : SIZE_MAX;
      size_t needed = 2 * *n + 1;   /* Be generous. */
      char *new_lineptr;

      if (needed_max < needed)
        needed = needed_max;
      if (cur_len + 1 >= needed)
        {
          result = -1;
          goto unlock_return;
        }

      new_lineptr = (char *) realloc (*lineptr, needed);
      if (new_lineptr == NULL)
        {
          result = -1;
          goto unlock_return;
        }

      *lineptr = new_lineptr;
      *n = needed;
    }

      (*lineptr)[cur_len] = i;
      cur_len++;

      if (i == delimiter)
    break;
    }
  (*lineptr)[cur_len] = '\0';
  result = cur_len ? cur_len : result;

 unlock_return:
  funlockfile (fp);
  return result;
}
#endif
#endif /* HAVE_GETLINE */

/*****************************************************************************/
/*                                                                           */
/*                                     main                                  */
/*                                                                           */
/*****************************************************************************/

static void SplitHelp ( const char * ) __attribute__ ((noreturn));

/// \brief display help using \b -h or \b --help command-line option
/// \param base the executable usually: <tt>basename(argv[0])</tt>
static void
SplitHelp ( const char * base  )
{
    fprintf(stderr, "usage: %s [OPTION]... [ARGUMENT]...\n", base);
    fputs("\n", stderr);
    fprintf(stderr, " The %s tool expect input data from stdin and is sending output data as\n", base);
    fputs(" package to a msgque server. For every line from stdin an output package is\n", stderr);
    fputs(" icreated by splitting the input line using the the delimiter -d.\n", stderr);
    fputs("\n", stderr);
    fprintf(stderr, "  %s [ARGUMENT]... syntax:\n", base);
    fprintf(stderr, "    ... | %s [OPTION]... %c aserver [OPTION]... [ARGUMENT]\n", base, MQ_ALFA);
    fputs("\n", stderr);
    fputs(MqHelp (NULL), stderr);
    fputs("\n", stderr);
    fprintf(stderr,"  %s [OPTION]:\n", base);
    fputs("    -d CHAR          the delimiter character                  (default: ' ')\n", stderr);
    fputs("    -h, --help       print this help\n", stderr);
    fputs("\n", stderr);

    exit(EXIT_SUCCESS);
}

/// \brief split the input-line(s) from \e fp into a msgque package using delimiter \e delimiter
/// \msgque
/// \param delimiter split the input line into fields using \e delimiter
/// \param fp using input-stream fp (by default stdout)
/// \retMqErrorE
static enum MqErrorE
split (
  struct MqS * const msgque,
  char delimiter,
  FILE *fp
)
{
  struct MqSendS * const send = msgque->send;
  size_t len = 0;
  ssize_t read;
  char *line=NULL, *cur, *cep;

  while ((read = getline(&line, &len, fp)) != -1) {
    MqSendSTART (send);
    for (cur=line; *cur != '\0'; cur=cep+1) {
      for (cep=cur; *cep != delimiter && *cep != '\n'; cep++);
      *cep = '\0';
      MqSendC (send, cur);
    }
    MqErrorCheck (MqSendFTR (send, MQ_TIMEOUT10));
  }
  if (line) free (line);
  MqErrorCheck (MqSendEOF (send, MQ_TIMEOUT10));
  return MQ_OK;

error:
  return MqErrorStack (msgque->error);
} 

/// \tool_Create
enum MqErrorE
SplitCreate (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS ** msgqueP
)
{
  MQ_STR defdelimiter="\t";		// the delimiter-char to split from
  MQ_STR delimiter = defdelimiter;

  // add my config data
  config->fHelp = SplitHelp;

  // create Msgque
  MqErrorCheck (MqCreate (config, argv, alfa, msgqueP));

  // get the delimiter character
  MqBufferLCheckOptionC ((*msgqueP)->error, argv, "-d", &delimiter);

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(*msgqueP, &argv));

  // do the splitting
  MqErrorCheck(split(*msgqueP, delimiter[0], stdin));

  // cleanup
  if (delimiter != defdelimiter) MqSysFree(delimiter);

  return MQ_OK;

error:
  if (delimiter != defdelimiter) MqSysFree(delimiter);
  MqErrorStack((*msgqueP)->error);
  return MqErrorSAppendC((*msgqueP)->error, "use '-h' or '--help' for usage");
}

/// \brief main entry-point for the tool
/// \param argc the number of command-line arguments
/// \param argv the command-line arguments as an array of strings
/// \return the exit number
int
SplitMain (
  int argc,
  char **argv
)
{
  // the parent-context
  struct MqS * msgque;
  struct MqBufferLS *largv, *lalfa;

  // split command line
  MqBufferLCreateArgv (argc, argv, &largv, &lalfa);

  // call context create
  SplitCreate (MqConfigInit(), largv, lalfa, &msgque);

  // exit 
  MqExit (msgque);

  return 0;
}

/** \} asplit */

