/**
 *  \file       src/buffer.c
 *  \brief      \$Id: buffer.c 1183 2009-05-18 09:58:23Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1183 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "buffer.h"
#include "error.h"
#include "sys.h"
#include "log.h"

#include <errno.h>
#include <math.h>
#include <limits.h>
#include <float.h>

static struct MqBufferS *
sBufferSetA1 ( struct MqBufferS * const, union MqBufferAtomU const, enum MqTypeE const);
static struct MqBufferS *
sBufferSetA2 ( struct MqBufferS * const, union MqBufferAtomU const, enum MqTypeE const);
static struct MqBufferS *
sBufferSetA4 ( struct MqBufferS * const, union MqBufferAtomU const, enum MqTypeE const);
static struct MqBufferS *
sBufferSetA8 ( struct MqBufferS * const, union MqBufferAtomU const, enum MqTypeE const);

// can be (*((MQ_INT const *)(bin)))
MQ_INLINE MQ_INT pByte2INT (
  //MQ_BINB const * const bin
  MQ_STRB const * const bin
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_INT i;
  memcpy(&i,bin,sizeof(MQ_INT));
  return i;
#else
  return (*((MQ_INT const *)(bin)));
#endif
}

// can be ((*buf->cur.A).Y)
MQ_INLINE MQ_BYT MqBufU2BYT (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_ATO a;
  a.B[0] = (*buf.A).B[0];
  return a.Y;
#else
  return (*buf.A).Y;
#endif
}

// can be ((*buf->cur.A).O)
MQ_INLINE MQ_BOL MqBufU2BOL (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_ATO a;
  a.B[0] = (*buf.A).B[0];
  return a.O;
#else
  return (*buf.A).O;
#endif
}

// can be ((*buf->cur.A).S)
MQ_INLINE MQ_SRT MqBufU2SRT (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_ATO a;
  a.B[0] = (*buf.A).B[0];
  a.B[1] = (*buf.A).B[1];
  return a.S;
#else
  return (*buf.A).S;
#endif
}

// can be ((*buf->cur.A).I)
MQ_INLINE MQ_INT MqBufU2INT (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_ATO a;
  a.B[0] = (*buf.A).B[0];
  a.B[1] = (*buf.A).B[1];
  a.B[2] = (*buf.A).B[2];
  a.B[3] = (*buf.A).B[3];
  return a.S;
#else
  return (*buf.A).I;
#endif
}

// can be ((*buf->cur.A).F)
MQ_INLINE MQ_FLT MqBufU2FLT (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_ATO a;
  a.B[0] = (*buf.A).B[0];
  a.B[1] = (*buf.A).B[1];
  a.B[2] = (*buf.A).B[2];
  a.B[3] = (*buf.A).B[3];
  return a.F;
#else
  return (*buf.A).F;
#endif
}

// can be ((*buf->cur.A).W)
MQ_INLINE MQ_WID MqBufU2WID (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_WID w;
  memcpy(&w,buf.B,sizeof(MQ_WID));
  return w;
#else
  return (*buf.A).W;
#endif
}

// can be (*((MQ_DBL const *)(bin)))
MQ_INLINE MQ_DBL MqBufU2DBL (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_DBL d;
  memcpy(&d,buf.B,sizeof(MQ_DBL));
  return d;
#else
  return (*buf.A).D;
#endif
}

// can be (*in->cur.A)
MQ_INLINE MQ_ATO MqBufU2ATO (
  union MqBufferU buf
) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
  MQ_ATO p;
  memcpy(&p,buf.B,sizeof(MQ_ATO));
  return p;
#else
  return *buf.A;
#endif
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_init                                  */
/*                                                                           */
/*****************************************************************************/

struct MqBufferS *
MqBufferCreate (
  MQ_SIZE size
)
{
  register struct MqBufferS * const buf = (struct MqBufferS * const) MqSysMalloc (MQ_ERROR_PANIC, sizeof (*buf));

  // +1 to allow strings with len=strlen(str) fit into the buffer with an additional
  // '\0' at the end
  if (size > MQ_BLS_SIZE) {
    buf->data = (MQ_BIN) MqSysCalloc (MQ_ERROR_PANIC, (size+1), sizeof (*buf->data));
    buf->size = size;
  } else {
    buf->data = (MQ_BIN) memset(buf->bls,'\0',MQ_BLS_SIZE);
    buf->size = MQ_BLS_SIZE;
  }
  buf->cursize = 0;
  buf->numItems = 0;
  buf->cur.B = buf->data;
  buf->alloc = MQ_DYNAMIC;
  buf->type = MQ_STRT;

  return buf;
}

/*
/// \brief create a new \e MqBufferS with \a size and a external memory \b not managed by \libmsgque
/// \error
/// \param data the external memory
/// \param size the size of the buffer to create
/// \attention be aware of the following restrictions:
///   - the user is responsible for creating and deleting the external memory, \libmsgque
///     is responsibilty for for creating and deleting the #MqBufferS object
///   - the external memory is \b read \b only, every modification of the object
///     (appending or setting) will convert the object into an dynamic memory object and
///     the original memory will \b not be used anymore
///   - the #MqBufferS object have to be deleted with #MqBufferDelete
///   - type type of the new object is #MQ_BINT
///   .
struct MqBufferS * MqBufferCreateBinary (
  struct MqErrorS * const error,
  MQ_BIN data,
  const MQ_SIZE size
);
struct MqBufferS *
MqBufferCreateBinary (
  register struct MqErrorS * const error,
  MQ_BIN data,
  const MQ_SIZE size
)
{
  struct MqBufferS * buf = MqSysMalloc (error, sizeof (*buf));

  buf->data = data;
  buf->size = size;
  buf->cursize = 0;
  buf->numItems = 0;
  buf->cur.B = data;
  buf->alloc = MQ_STATIC;
  buf->type = MQ_BINT;

  return buf;
}
*/

struct MqBufferS *
pBufferCreateRef (
  struct MqBufferS const * const buf
)
{
  struct MqBufferS * const ret = (struct MqBufferS *) MqSysMalloc (MQ_ERROR_PANIC, sizeof (*ret));
  memcpy (ret, buf, sizeof (*ret));
  ret->alloc = MQ_STATIC;
  return ret;
}

void
MqBufferDelete (
  struct MqBufferS ** const bufP
)
{
  struct MqBufferS *buf;
  if (unlikely (bufP == NULL || *bufP == NULL)) return;
  buf = *bufP;

  if (buf->data && buf->alloc == MQ_DYNAMIC && buf->data != buf->bls) {
    MqSysFree (buf->data);
  }
  MqSysFree (*bufP);
}

MQ_INLINE void
MqBufferReset (
  struct MqBufferS * const buf
)
{
  buf->cur.B = buf->data;
  buf->cursize = 0;
  buf->numItems = 0;
}

/*****************************************************************************/
/*                                                                           */
/*                           buffer_create_type                              */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE struct MqBufferS * 
MqBufferCreateY (
  MQ_BYT const val
)
{
  union MqBufferAtomU ato;
  ato.Y = val;
  return sBufferSetA1(MqBufferCreate (0), ato, MQ_BYTT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateO (
  MQ_BOL const val
)
{
  union MqBufferAtomU ato;
  ato.O = val;
  return sBufferSetA1(MqBufferCreate (0), ato, MQ_BOLT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateS (
  MQ_SRT const val
)
{
  union MqBufferAtomU ato;
  ato.S = val;
  return sBufferSetA2(MqBufferCreate (0), ato, MQ_SRTT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateI (
  MQ_INT const val
)
{
  union MqBufferAtomU ato;
  ato.I = val;
  return sBufferSetA4(MqBufferCreate (0), ato, MQ_INTT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateF (
  MQ_FLT const val
)
{
  union MqBufferAtomU ato;
  ato.F = val;
  return sBufferSetA4(MqBufferCreate (0), ato, MQ_FLTT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateW (
  MQ_WID const val
)
{
  union MqBufferAtomU ato;
  ato.W = val;
  return sBufferSetA8(MqBufferCreate (0), ato, MQ_WIDT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateD (
  MQ_DBL const val
)
{
  union MqBufferAtomU ato;
  ato.D = val;
  return sBufferSetA8(MqBufferCreate (0), ato, MQ_DBLT);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateC (
  MQ_STRB const * const val
)
{
  return MqBufferSetC(MqBufferCreate (0), val);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateB (
  MQ_BINB const * const val,
  MQ_SIZE const len
)
{
  return MqBufferSetB(MqBufferCreate (0), val, len);
}

MQ_INLINE struct MqBufferS * 
MqBufferCreateU (
  struct MqBufferS const * const val
)
{
  return MqBufferSetU(MqBufferCreate (0), val);
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_atom                                  */
/*                                                                           */
/*****************************************************************************/

void
_pBufferNewSize (
  register struct MqBufferS * const buf,
  MQ_STRB const * const header,
  MQ_SIZE const newSize
)
{

/*
if (newSize == buf->size+1) {
  printLog (msgque, "buffer SIZE warning: newSize<%i> = buf->size<%i> + 1 !!\n",
    newSize, buf->size);
}
*/

  MQ_SIZE size = newSize + (newSize / 3);  // + 1/3 reserve
  // don't use buf->cursize, because this can be invalid at the time this function is called
  MQ_SIZE cursize = (buf->cur.B - buf->data);
  if (unlikely (buf->alloc == MQ_DYNAMIC && buf->size > MQ_BLS_SIZE)) {
    // the buf->size+1 is used to fit a string with '\0' and strlen() size into the buffer
    buf->data = (MQ_BIN) MqSysRealloc (MQ_ERROR_PANIC, buf->data, size+1);
  } else {
    buf->data = (MQ_BIN)MqSysMalloc (MQ_ERROR_PANIC, size+1);
    buf->alloc = MQ_DYNAMIC;
  }
  buf->size = size;
  buf->cur.B = buf->data + cursize;
  memset (buf->cur.B, '\0', (buf->size - cursize));

  // additional debugging output
  //MqDLogV (msgque, header, 5, "increase buffer size to " MQ_FORMAT_Z " bytes\n", buf->size);
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_get                                   */
/*                                                                           */
/*****************************************************************************/

#define RANGE_E(txt,type) \
    MqErrorSGenV(error, __func__, MQ_ERROR, -1, \
        #txt " ERROR, can not convert '%s' into an '" #type "' object", buf->cur.C)

#define TYPE_E(ttype,stype) \
    MqErrorSGenV(error, __func__, MQ_ERROR, -1, \
        "type ERROR, it is not allowed to convert an '%s' object into an '%s' object", \
	MqLogTypeName(stype), MqLogTypeName(ttype))


/// \brief MQ_BYT maximum
#define MQ_BYT_MAX      SCHAR_MAX
/// \brief MQ_BYT minimum
#define MQ_BYT_MIN      SCHAR_MIN

/// \brief MQ_SRT maximum
#define MQ_SRT_MAX      SHRT_MAX
/// \brief MQ_SRT minimum
#define MQ_SRT_MIN      SHRT_MIN

/// \brief MQ_INT maximum
#define MQ_INT_MAX      INT_MAX
/// \brief MQ_INT minimum
#define MQ_INT_MIN      INT_MIN

/// \brief MQ_WID maximum
#define MQ_WID_MAX      LLONG_MAX
/// \brief MQ_WID minimum
#define MQ_WID_MIN      LLONG_MIN

enum MqErrorE
pBufferGetA1 (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  union MqBufferAtomU * const ato,
  enum MqTypeE const type 
)
{
	buf->cur.B = buf->data;
  if (type == buf->type) {
    (*ato).Y = (*buf->cur.A).Y;
  } else if ( buf->type == MQ_STRT) {
    if (type == MQ_BYTT) {
			long ret;
      MQ_STR end;
      errno = 0;
      ret = str2int (buf->cur.C, &end, 0);
      (*ato).Y = (MQ_BYT) ret;
    // on 32bit: long=int
      if (unlikely (errno == ERANGE)) {
	if (unlikely (ret == LONG_MIN)) {
	  return RANGE_E (underflow, BYT);
	} else if (unlikely (ret == LONG_MAX)) {
	  return RANGE_E (overflow, BYT);
	}
      } else {
	if (unlikely (ret < (long) MQ_BYT_MIN)) {
	  return RANGE_E (underflow, BYT);
	} else if (unlikely (ret > (long) MQ_BYT_MAX)) {
	  return RANGE_E (overflow, BYT);
	} else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
	  return RANGE_E (size, BYT);
	}
      }
    } else {
      // MQ_BOLT
      (*ato).O = ((*buf->cur.C)-48);   // 48 is the ascii value from '0'
      if (unlikely ((*ato).O < 0)) {
	return RANGE_E (underflow, BOL);
      } else if (unlikely ((*ato).O > 1)) {
	return RANGE_E (overflow, BOL);
      } else if (unlikely (buf->cursize != 1)) {
	return RANGE_E (size, BOL);
      }
    }
  } else {
    return TYPE_E(type, buf->type);
  }
  return MQ_OK;
}

enum MqErrorE
pBufferGetA2 (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  union MqBufferAtomU * const ato,
  enum MqTypeE const type 
)
{
  buf->cur.B = buf->data;
  if (type == buf->type) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
    memcpy(ato,buf->cur.B,2);
#else
    (*ato).S = (*buf->cur.A).S;
#endif
  } else if ( buf->type == MQ_STRT ) {
    // MQ_SRTT
		MQ_STR end;
    long ret;
    errno = 0;
    ret = str2int (buf->cur.C, &end, 0);
    (*ato).S = (MQ_SRT) ret;
    if (unlikely (errno == ERANGE)) {
      if (unlikely (ret == LONG_MIN)) {
	return RANGE_E (underflow, SRT);
      } else if (unlikely (ret == LONG_MAX)) {
	return RANGE_E (overflow, SRT);
      }
    } else {
      if (unlikely (ret < (long) MQ_SRT_MIN)) {
	return RANGE_E (underflow, SRT);
      } else if (unlikely (ret > (long) MQ_SRT_MAX)) {
	return RANGE_E (overflow, SRT);
      } else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
	return RANGE_E (size, SRT);
      }
    }
  } else {
    return TYPE_E(type, buf->type);
  }
  return MQ_OK;
}

enum MqErrorE
pBufferGetA4 (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  union MqBufferAtomU * const ato,
  enum MqTypeE const type 
)
{
  buf->cur.B = buf->data;
  if (type == buf->type) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
    memcpy(ato,buf->cur.B,4);
#else
    (*ato).I = (*buf->cur.A).I;
#endif
  } else if ( buf->type == MQ_STRT ) {
    MQ_STR end;
    errno = 0;
    if (type == MQ_INTT) {
      // MQ_INTT
      long ret = str2int (buf->cur.C, &end, 0);
      (*ato).I = (MQ_INT) ret;
      // on 32bit: long=int
      if (unlikely (errno == ERANGE)) {
        if (unlikely (ret == LONG_MIN)) {
          return RANGE_E (underflow, INT);
        } else if (unlikely (ret == LONG_MAX)) {
          return RANGE_E (overflow, INT);
        }
      } else {
        if (unlikely (ret < (long) MQ_INT_MIN)) {
          return RANGE_E (underflow, INT);
        } else if (unlikely (ret > (long) MQ_INT_MAX)) {
          return RANGE_E (overflow, INT);
        } else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
          return RANGE_E (size, INT);
        }
      }
    } else {
#if defined(HAVE_STRTOF)
      // MQ_FLTT
      (*ato).F = str2flt (buf->cur.C, &end);
      if (unlikely (errno == ERANGE)) {
        if (unlikely ((*ato).F == 0.0)) {
          return RANGE_E (underflow, FLT);
        } else if (unlikely ((*ato).F == +MQ_HUGE_VALF || (*ato).F == -MQ_HUGE_VALF)) {
          return RANGE_E (overflow, FLT);
        }
      } else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
        return RANGE_E (size, FLT);
      }
#else /* ! HAVE_STRTOF */
      // MQ_DBLT
      double d = str2dbl (buf->cur.C, &end);
      if (unlikely (errno == ERANGE)) {
        if (unlikely (d == 0.0L)) {
          return RANGE_E (underflow, FLT);
        } else if (unlikely (d == +MQ_HUGE_VAL || d == -MQ_HUGE_VAL)) {
          return RANGE_E (overflow, FLT);
        }
      } else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
        return RANGE_E (size, FLT);
      }
      (*ato).F = (float) d;
#endif /* !HAVE_STRTOF */
    }
  } else {
    return TYPE_E(type, buf->type);
  }
  return MQ_OK;
}

enum MqErrorE
pBufferGetA8 (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  union MqBufferAtomU * const ato,
  enum MqTypeE const type 
)
{
  buf->cur.B = buf->data;
  if (type == buf->type) {
#if defined(HAVE_ALIGNED_ACCESS_REQUIRED)
    memcpy(ato,buf->cur.B,8);
#else
    (*ato).W = (*buf->cur.A).W;
#endif
  } else if ( buf->type == MQ_STRT ) {
    MQ_STR end;
    errno = 0;
    if (type == MQ_WIDT) {
      // MQ_WIDT
      (*ato).W = str2wid (buf->cur.C, &end, 0);
      if (unlikely (errno == ERANGE)) {
	if (unlikely ((*ato).W == MQ_WID_MIN)) {
	  return RANGE_E (underflow, WID);
	} else if (unlikely ((*ato).W == MQ_WID_MAX)) {
	  return RANGE_E (overflow, WID);
	}
      } else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
	return RANGE_E (size, WID);
      }
    } else {
      // MQ_DBLT
      (*ato).D = str2dbl (buf->cur.C, &end);
      if (unlikely (errno == ERANGE)) {
	if (unlikely ((*ato).D == 0.0L)) {
	  return RANGE_E (underflow, DBL);
	} else if (unlikely ((*ato).D == +MQ_HUGE_VAL || (*ato).D == -MQ_HUGE_VAL)) {
	  return RANGE_E (overflow, DBL);
	}
      } else if (unlikely ((end - buf->cur.C) != buf->cursize)) {
	return RANGE_E (size, DBL);
      }
    }
  } else {
    return TYPE_E(type, buf->type);
  }
  return MQ_OK;
}

MQ_INLINE enum MqErrorE
MqBufferGetY (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_BYT * const valP
)
{
  return pBufferGetA1(error,buf,(MQ_ATO*)valP,MQ_BYTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetO (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_BOL * const valP
)
{
  return pBufferGetA1(error,buf,(MQ_ATO*)valP,MQ_BOLT);
}

MQ_INLINE enum MqErrorE
MqBufferGetS (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_SRT * const valP
)
{
  return pBufferGetA2(error,buf,(MQ_ATO*)valP,MQ_SRTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetI (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_INT * const valP
)
{
  return pBufferGetA4(error,buf,(MQ_ATO*)valP,MQ_INTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetF (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_FLT * const valP
)
{
  return pBufferGetA4(error,buf,(MQ_ATO*)valP,MQ_FLTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetW (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_WID * const valP
)
{
  return pBufferGetA8(error,buf,(MQ_ATO*)valP,MQ_WIDT);
}

MQ_INLINE enum MqErrorE
MqBufferGetD (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_DBL * const valP
)
{
  return pBufferGetA8(error,buf,(MQ_ATO*)valP,MQ_DBLT);
}

MQ_INLINE enum MqErrorE
MqBufferGetB (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_BIN * const out,
  MQ_SIZE * const size
)
{
  *out = buf->data;
  *size = buf->cursize;
  return MQ_OK;
}

enum MqErrorE
MqBufferGetC (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  MQ_STR * const out
)
{
  /// the first MQ_ATO bytes of the \e bls are reserved for the native data
  MQ_STR stringP = (MQ_STR) (buf->bls+sizeof(MQ_ATO));

  buf->cur.B = buf->data;
  switch (buf->type) {
    case MQ_INTT:   sprintf (stringP, MQ_FORMAT_I, MqBufU2INT(buf->cur)); break;
    case MQ_WIDT:   sprintf (stringP, MQ_FORMAT_W, MqBufU2WID(buf->cur)); break;
    case MQ_DBLT:   sprintf (stringP, MQ_FORMAT_D, MqBufU2DBL(buf->cur)); break;
    case MQ_STRT:   stringP = buf->cur.C; break;
    default:	    return TYPE_E(MQ_STRT, buf->type);
  }
  *out = stringP;
  return MQ_OK;
}

char
MqBufferGetType (
  struct MqBufferS * const buf
)
{
  switch (buf->type) {
    case MQ_STRT: return 'C';
    case MQ_INTT: return 'I';
    case MQ_DBLT: return 'D';
    case MQ_WIDT: return 'W';
    case MQ_BINT: return 'B';
    case MQ_BYTT: return 'Y';
    case MQ_BOLT: return 'O';
    case MQ_SRTT: return 'S';
    case MQ_FLTT: return 'F';
    case MQ_LSTT: return 'L';
    case MQ_RETT: return 'R';
  }
  return '*';
}

/*****************************************************************************/
/*                                                                           */
/*                          buffer_getM functions                            */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
MqBufferGetYM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_BYT * const valP
)
{
  return pBufferGetA1(msgque->error,buf,(MQ_ATO*)valP,MQ_BYTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetOM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_BOL * const valP
)
{
  return pBufferGetA1(msgque->error,buf,(MQ_ATO*)valP,MQ_BOLT);
}

MQ_INLINE enum MqErrorE
MqBufferGetSM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_SRT * const valP
)
{
  return pBufferGetA2(msgque->error,buf,(MQ_ATO*)valP,MQ_SRTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetIM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_INT * const valP
)
{
  return pBufferGetA4(msgque->error,buf,(MQ_ATO*)valP,MQ_INTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetFM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_FLT * const valP
)
{
  return pBufferGetA4(msgque->error,buf,(MQ_ATO*)valP,MQ_FLTT);
}

MQ_INLINE enum MqErrorE
MqBufferGetWM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_WID * const valP
)
{
  return pBufferGetA8(msgque->error,buf,(MQ_ATO*)valP,MQ_WIDT);
}

MQ_INLINE enum MqErrorE
MqBufferGetDM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_DBL * const valP
)
{
  return pBufferGetA8(msgque->error,buf,(MQ_ATO*)valP,MQ_DBLT);
}

MQ_INLINE enum MqErrorE
MqBufferGetBM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_BIN * const out,
  MQ_SIZE * const size
)
{
  *out = buf->data;
  *size = buf->cursize;
  return MQ_OK;
}

MQ_INLINE enum MqErrorE
MqBufferGetCM (
  struct MqS const * const msgque,
  struct MqBufferS * const buf,
  MQ_STR * const out
)
{
  return MqBufferGetC(msgque->error, buf, out);
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_set                                   */
/*                                                                           */
/*****************************************************************************/

static struct MqBufferS *
sBufferSetA1 (
  struct MqBufferS * const buf,
  union MqBufferAtomU const in,
  enum MqTypeE const type
)
{
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  // the following line is !!not!! necessary because the initial
  // size of a new created buffer is bigger than sizeof(in)
  // pBufferNewSize (buf, sizeof(in));
  (*buf->cur.A).Y = in.Y;
  buf->cursize = 1;
  buf->type = type;
  return buf;
}

static struct MqBufferS *
sBufferSetA2 (
  struct MqBufferS * const buf,
  union MqBufferAtomU const in,
  enum MqTypeE const type
)
{
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  // the following line is !!not!! necessary because the initial
  // size of a new created buffer is bigger than sizeof(in)
  // pBufferNewSize (buf, sizeof(in));
  (*buf->cur.A).S = in.S;
  buf->cursize = 2;
  buf->type = type;
  return buf;
}

static struct MqBufferS *
sBufferSetA4 (
  struct MqBufferS * const buf,
  union MqBufferAtomU const in,
  enum MqTypeE const type
)
{
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  // the following line is !!not!! necessary because the initial
  // size of a new created buffer is bigger than sizeof(in)
  // pBufferNewSize (buf, sizeof(in));
  (*buf->cur.A).I = in.I;
  buf->cursize = 4;
  buf->type = type;
  return buf;
}

static struct MqBufferS *
sBufferSetA8 (
  struct MqBufferS * const buf,
  union MqBufferAtomU const in,
  enum MqTypeE const type
)
{
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  // the following line is !!not!! necessary because the initial
  // size of a new created buffer is bigger than sizeof(in)
  // pBufferNewSize (buf, sizeof(in));
  (*buf->cur.A).W = in.W;
  buf->cursize = 8;
  buf->type = type;
  return buf;
}

MQ_INLINE struct MqBufferS *
MqBufferSetY (
  struct MqBufferS * const buf,
  MQ_BYT const val
)
{
	union MqBufferAtomU ato;
	ato.Y = val;
  return sBufferSetA1(buf,ato,MQ_BYTT);
}

MQ_INLINE struct MqBufferS *
MqBufferSetO (
  struct MqBufferS * const buf,
  MQ_BOL const val
)
{
  union MqBufferAtomU ato;
	ato.O = val;
  return sBufferSetA1(buf,ato,MQ_BOLT);
}

MQ_INLINE struct MqBufferS *
MqBufferSetS (
  struct MqBufferS * const buf,
  MQ_SRT const val
)
{
  union MqBufferAtomU ato;
	ato.S = val;
  return sBufferSetA2(buf,ato,MQ_SRTT);
}

MQ_INLINE struct MqBufferS *
MqBufferSetI (
  struct MqBufferS * const buf,
  MQ_INT const val
)
{
  union MqBufferAtomU ato;
	ato.I = val;
  return sBufferSetA4(buf,ato,MQ_INTT);
}

MQ_INLINE struct MqBufferS *
MqBufferSetF (
  struct MqBufferS * const buf,
  MQ_FLT const val
)
{
  union MqBufferAtomU ato;
	ato.F = val;
  return sBufferSetA4(buf,ato,MQ_FLTT);
}

MQ_INLINE struct MqBufferS *
MqBufferSetW (
  struct MqBufferS * const buf,
  MQ_WID const val
)
{
  union MqBufferAtomU ato;
	ato.W = val;
  return sBufferSetA8(buf,ato,MQ_WIDT);
}

MQ_INLINE struct MqBufferS *
MqBufferSetD (
  struct MqBufferS * const buf,
  MQ_DBL const val
)
{
  union MqBufferAtomU ato;
	ato.D = val;
  return sBufferSetA8(buf,ato,MQ_DBLT);
}

struct MqBufferS *
MqBufferSetC (
  struct MqBufferS * const buf,
  MQ_STRB const * const in
)
{
  MQ_SIZE len = strlen(in);
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  pBufferNewSize (buf, len);
	strcpy (buf->cur.C, in);
  buf->cursize = len;
  buf->type = MQ_STRT;
  return buf;
}

struct MqBufferS *
MqBufferSetB (
  struct MqBufferS * const buf,
  MQ_BINB const * const in,
  MQ_SIZE const len
)
{
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  pBufferNewSize (buf, len);
  memcpy (buf->cur.B, in, len);
  buf->cursize = len;
  buf->type = MQ_BINT;
  return buf;
}

struct MqBufferS *
MqBufferSetU (
  struct MqBufferS * const buf,
  struct MqBufferS const * const in
)
{
  // reset the cur pointer to the start of the buffer
  buf->cur.B = buf->data;
  pBufferNewSize (buf, in->cursize);
  // the +1 is for the string \0 byte
  memcpy (buf->cur.B, in->data, (in->cursize+1));
  buf->cursize = in->cursize;
  buf->type = in->type;
  return buf;
}

struct MqBufferS *
MqBufferSetV (
  struct MqBufferS * const buf,
  MQ_STRB const * const fmt,
  ...
)
{
  va_list ap;
  MqBufferReset (buf);
  va_start (ap, fmt);
  MqBufferAppendVL (buf, fmt, ap);
  buf->cur.B = buf->data;
  va_end (ap);
  return buf;
}

enum MqErrorE
MqBufferCastTo (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  enum MqTypeE const type
)
{
  MQ_ATO val;

  if (buf->type == type)
    return MQ_OK;

  switch (type & MQ_TYPE_IS_NATIVE) {
    case 8:
      MqErrorCheck (pBufferGetA8 (error, buf, &val, type));
      sBufferSetA8 (buf, val, type);
      break;
    case 4:
      MqErrorCheck (pBufferGetA4 (error, buf, &val, type));
      sBufferSetA4 (buf, val, type);
      break;
    case 2:
      MqErrorCheck (pBufferGetA2 (error, buf, &val, type));
      sBufferSetA2 (buf, val, type);
      break;
    case 1:
      MqErrorCheck (pBufferGetA1 (error, buf, &val, type));
      sBufferSetA1 (buf, val, type);
      break;
    default:
      return MqErrorDb2(error,MQ_ERROR_CAST_TO_NON_ATOM);
  }

  return MQ_OK;

error:
  return MqErrorStack (error);
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_append                                */
/*                                                                           */
/*****************************************************************************/

struct MqBufferS *
MqBufferCopy (
  register struct MqBufferS * const dest,
  register struct MqBufferS const * const srce
)
{
  dest->cursize = srce->cursize;
  dest->numItems = srce->numItems;
  dest->type = srce->type;
  pBufferNewSize (dest, srce->cursize);
  memcpy (dest->data, srce->data, srce->cursize+1);
  dest->cur.B = dest->data + (srce->cur.B - srce->data);
  return dest;
}

MQ_INLINE struct MqBufferS *
MqBufferDup (
  struct MqBufferS const * const srce
)
{
  return MqBufferCopy (MqBufferCreate (srce->cursize), srce);
}

MQ_SIZE
MqBufferAppendC (
  register struct MqBufferS * const buf,
  MQ_STRB const * const str
)
{
  MQ_SIZE len = strlen (str);
  pBufferAddSize (buf, len);
  strcpy (buf->cur.C, str);
  buf->cursize += len;
  buf->cur.B += len;
  buf->type = MQ_STRT;
  return len;
}

MQ_SIZE
MqBufferAppendVL (
  register struct MqBufferS * const buf,
  MQ_STRB const * const fmt,
  va_list ap
)
{
  MQ_SIZE left;
  MQ_SIZE len;
  va_list ap_save;

  va_copy(ap_save,ap);
  //     MaxSize     CurrSize    
  left = buf->size - buf->cursize;
  // test the len
  len = vsnprintf (buf->cur.C, left, fmt, ap_save);
  // free arg-list
  va_end(ap_save);

  // glibc 2.0.6 return -1 when the output was truncated
  if (len == -1) {
    va_copy(ap_save,ap);
    len = vsnprintf (NULL, 0, fmt, ap_save);
    va_end(ap_save);
  }

  if (len >= left) {
    //                               CurSize      + len + \0
    _pBufferNewSize (buf, __func__, buf->cursize + len + 1);
    //       MaxSize     CurrSize     
    left = buf->size - buf->cursize;
    len = vsnprintf ((buf->cur.C), left, fmt, ap);
    if (len >= left)
      MqPanicV (MQ_ERROR_PANIC, __func__, -1, "length mismatch: len<" 
	MQ_FORMAT_Z "> > left<" MQ_FORMAT_Z ">", len, left);
  }
  buf->cursize += len;
  buf->cur.B += len;
  buf->type = MQ_STRT;
  return len;
}

MQ_SIZE
MqBufferAppendV (
  struct MqBufferS * const buf,
  MQ_STRB const * const fmt,
  ...
)
{
  MQ_SIZE ret;
  va_list ap;
  va_start (ap, fmt);
  ret = MqBufferAppendVL (buf, fmt, ap);
  va_end (ap);
  return ret;
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_push                                  */
/*                                                                           */
/*****************************************************************************/

MQ_SIZE
MqBufferPush (
  register struct MqBufferS * const buf,
  MQ_STRB const * const str
)
{
  MQ_INT const len = strlen (str);
  pBufferAddSize (buf, len);
  strcpy (buf->cur.C, str);
  buf->cursize += len;
  buf->cur.B += len;
  return len;
}

MQ_SIZE
MqBufferPop (
  register struct MqBufferS * const buf,
  MQ_STRB const * const str
)
{
  MQ_INT const len = strlen (str);
  buf->cursize -= len;
  buf->cur.B -= len;
  *buf->cur.B = '\0';
  return len;
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_log                                   */
/*                                                                           */
/*****************************************************************************/

void
MqBufferLogS (
  struct MqS const * const msgque,
  struct MqBufferS const * const buf,
  MQ_STRB const * const prefix
)
{
#define MY_MAX 300
  MQ_STR cur = (buf->cursize > MY_MAX ? (MQ_STR) "" : buf->cur.C);
  switch (buf->type) {
    case MQ_STRT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_C ">\n", cur);
      break;
    case MQ_BINT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_B ">\n", buf->cur.B);
      break;
    case MQ_BYTT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_Y ">\n", MqBufU2BYT(buf->cur));
      break;
    case MQ_BOLT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_O ">\n", MqBufU2BOL(buf->cur));
      break;
    case MQ_SRTT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_S ">\n", MqBufU2SRT(buf->cur));
      break;
    case MQ_INTT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_I ">\n", MqBufU2INT(buf->cur));
      break;
    case MQ_FLTT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_F ">\n", MqBufU2FLT(buf->cur));
      break;
    case MQ_WIDT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_W ">\n", MqBufU2WID(buf->cur));
      break;
    case MQ_DBLT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_D ">\n", MqBufU2DBL(buf->cur));
      break;
    case MQ_LSTT:
      MqDLogV (msgque, prefix, 0, "cur      = <" MQ_FORMAT_L ">\n", buf->cur.L);
      break;
    case MQ_RETT:
      MqDLogV (msgque, prefix, 0, "cur      = <%s>\n", MqLogTypeName(buf->type));
      break;
  }
#undef MY_MAX
}

/*****************************************************************************/
/*                                                                           */
/*                                debugging                                  */
/*                                                                           */
/*****************************************************************************/

#if defined _DEBUG
void
MqBufferLog (
  struct MqS const * const msgque,
  struct MqBufferS const * const buf,
  MQ_STRB const * const prefix
)
{
  MQ_STR tmp = (MQ_STR) MqSysMalloc(MQ_ERROR_IGNORE, (buf->cursize+1));
  // no memory -> no work
  if (!buf) return;

  MqDLogV (msgque, prefix, 0, ">>>> MqBufferS (%p)\n", buf);

  MqDLogV (msgque, prefix, 0, "data     = <%s>\n", MqLogC (tmp,(MQ_STR)buf->data, buf->cursize));
  MqDLogV (msgque, prefix, 0, "size     = <" MQ_FORMAT_Z ">\n", buf->size);
  MqDLogV (msgque, prefix, 0, "cursize  = <" MQ_FORMAT_Z ">\n", buf->cursize);
  MqDLogV (msgque, prefix, 0, "numItems = <" MQ_FORMAT_Z ">\n", buf->numItems);
  MqDLogV (msgque, prefix, 0, "curpos   = <" MQ_FORMAT_Z ">\n", (MQ_SIZE) (buf->cur.B - buf->data));
  MqDLogV (msgque, prefix, 0, "type     = <%s:%i>\n", MqLogTypeName(buf->type), buf->type);
  MqDLogV (msgque, prefix, 0, "pointer  = data<%p> buf->cur.B<%p>\n", buf->data, buf->cur.B);

  MqBufferLogS(msgque, buf, prefix);

  MqDLogV (msgque, prefix, 0, "alloc    = <%s>\n",
           (buf->alloc == MQ_DYNAMIC ? "MQ_DYNAMIC" : "STATIC"));
  MqDLogV (msgque, prefix, 0, "<<<< MqBufferS\n");

  MqSysFree(tmp);
}

#endif

