/**
 *  \file       src/error.c
 *  \brief      \$Id: error.c 1192 2009-05-19 12:53:58Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1192 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"

/*****************************************************************************/
/*                                                                           */
/*                              private                                      */
/*                                                                           */
/*****************************************************************************/

struct MqErrorS {
  struct MqS const * msgque;    ///<
  struct MqBufferS * text;      ///<
  enum MqErrorE code;		///<
  int num;	                ///<
  MQ_BOL append;		///< allow to append? MQ_YES or MQ_NO
};

#include "error.h"
#include "buffer.h"
#include "send.h"
#include "log.h"
#include "sys.h"
#include "mq_io.h"
#include "slave.h"

#define MSGQUE msgque

/*****************************************************************************/
/*                                                                           */
/*                              error_init                                   */
/*                                                                           */
/*****************************************************************************/

void
pErrorCreate (
  struct MqS const * const msgque,
  struct MqErrorS ** const out
)
{
  struct MqErrorS * error = (struct MqErrorS *) MqSysMalloc (MQ_ERROR_PANIC, sizeof (*error));
  if (likely (out != NULL))
    *out = error;

  error->msgque = msgque;
  error->code = MQ_OK;
  error->num = EXIT_SUCCESS;
  error->text = MqBufferCreate (1000);
  error->append = MQ_YES;
}

void
pErrorDelete (
  struct MqErrorS ** const errorP
)
{
  struct MqErrorS * const error = *errorP;
  if (unlikely (error == NULL))
    return;
  MqBufferDelete (&error->text);
  MqSysFree (*errorP);
}

/*****************************************************************************/
/*                                                                           */
/*                              error_api                                    */
/*                                                                           */
/*****************************************************************************/

void
MqPanicVL (
  struct MqErrorS * const error,
  MQ_STRB const * const prefix,
  MQ_INT const errnum,
  MQ_STRB const * const fmt,
  va_list ap
)
{
  if (error == MQ_ERROR_IGNORE) return;
  if (error == MQ_ERROR_PANIC) {
    struct MqBufferS * buf = MqBufferCreate (255);
    MqBufferSetV (buf, "PANIC: %s\n", fmt);
    MqDLogEVL (NULL, prefix, buf->cur.C, ap);
    MqBufferDelete (&buf);
  } else {
    struct MqS * const parent = pMqGetFirstParent(MqErrorGetMsgque(error));
    MqErrorSGenVL (error, "PANIC", MQ_ERROR, errnum, fmt, ap);
    if (prefix) MqErrorSAppendV (error, "found in function \"%s\"", prefix);
    MqErrorCopy(parent->error, error);
    pErrorReport(parent->error);
    MqExit ((struct MqS *) parent);
  }
  MqSysSleep (error, 2);
  SysAbort();
}

void
MqPanicV (
  struct MqErrorS * const error,
  MQ_STRB const * const prefix,
  MQ_INT const errnum,
  MQ_STRB const * const fmt,
  ...
)
{
  va_list ap;
  if (error == MQ_ERROR_IGNORE) return;
  va_start (ap, fmt);
  MqPanicVL (error, prefix, errnum, fmt, ap);
  va_end (ap);
}

/*****************************************************************************/
/*                                                                           */
/*                            error_create                                   */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
pErrorSet (
  struct MqErrorS * const error,
  MQ_INT const errnum,
  MQ_STRB const * const errtext,
  MQ_BOL const append
)
{
  struct MqS * master = error->msgque->config.master;
  if (!error) {
    MqPanicV(NULL,__func__,-1,"unable to set error");
  }
  error->code = MQ_ERROR;
  error->num = (int) (errnum > 0 ? errnum : EXIT_FAILURE);
  error->append = append;
  MqBufferReset (error->text);
  MqBufferAppendC (error->text, errtext);
  if (master != NULL)
    pErrorSync(master->error, error);
  return MQ_ERROR;
}

enum MqErrorE
MqErrorSGenVL (
  struct MqErrorS * const error,
  MQ_STRB const * const prefix,
  enum MqErrorE const errorcode,
  MQ_INT const errnum,
  MQ_STRB const * const fmt,
  va_list ap
)
{
  if (error == MQ_ERROR_IGNORE) {
    // do nothing
  } else if (error == MQ_ERROR_PANIC) {
    struct MqBufferS * buf = MqBufferCreate (255);
    MqBufferSetV (buf, "#(" MQ_FORMAT_I ") -> %s\n", errnum, fmt);
    if (errorcode == MQ_ERROR) {
      MqPanicVL (NULL, prefix, errnum, (MQ_STR) buf->data, ap);
    } else {
      MqDLogEVL (NULL, prefix, (MQ_STR) buf->data, ap);
    }
    MqBufferDelete (&buf);
  } else if (error == MQ_ERROR_PRINT) {
    struct MqBufferS * buf = MqBufferCreate (255);
    MqBufferSetV (buf, "#(" MQ_FORMAT_I ") -> %s\n", errnum, fmt);
    MqDLogEVL (NULL, prefix, (MQ_STR) buf->data, ap);
    MqBufferDelete (&buf);
  } else {
    struct MqS const * const msgque = error->msgque;
    struct MqS const * const master = msgque->config.master;
    struct MqBufferS * text = error->text;
    error->code = errorcode;
    error->num = (errnum > 0 ? errnum : EXIT_FAILURE);
    MqBufferReset (text);
    if (prefix) {
      if (msgque) {
	MqBufferAppendV (text, "%c> (%s) ", ( MQ_IS_SERVER (msgque) ?
	  (MQ_IS_CHILD (msgque) ? 's' : 'S') : (MQ_IS_CHILD (msgque) ? 'c' : 'C')),
	  msgque->config.name); 
      }
      MqBufferAppendV (text, "[%s] ", prefix);
    }
    MqBufferAppendVL (text, fmt, ap);
    if (master != NULL) {
      pErrorSync(master->error, error);
    }
  }
  return errorcode;
}

enum MqErrorE
MqErrorSGenV (
  struct MqErrorS * const error,
  MQ_STRB const * const prefix,
  enum MqErrorE const errorcode,
  MQ_INT const errnum,
  MQ_STRB const * const fmt,
  ...
)
{
  va_list ap;
  if (error != MQ_ERROR_IGNORE) {
    va_start (ap, fmt);
    MqErrorSGenVL (error, prefix, errorcode, errnum, fmt, ap);
    va_end (ap);
  }
  return errorcode;
}

MQ_INLINE enum MqErrorE
MqErrorC (
  struct MqErrorS * const error,
  MQ_STRB const * const prefix,
  MQ_INT const errnum,
  MQ_STRB const * const message
)
{
  return MqErrorSGenV(error, prefix, MQ_ERROR, errnum, "%s", message);
}

MQ_INLINE enum MqErrorE
MqErrorCM (
  struct MqS const * const msgque,
  MQ_STRB const * const prefix,
  MQ_INT const errnum,
  MQ_STRB const * const message
)
{
  return MqErrorSGenV(msgque->error, prefix, MQ_ERROR, errnum, "%s", message);
}

// no MQ_INLINE because "error" - code is allways slow
enum MqErrorE
MqErrorStackP(
  struct MqErrorS const * const error,
  char const * const func,
  char const * const file
)
{
  MQ_STRB * basename;
  if (error == MQ_ERROR_IGNORE) return MQ_ERROR;
  basename = MqSysBasename(file, MQ_YES);
  MqErrorSAppendV(error, "found in function \"%s\" at file \"%s\"", func, basename);
  free(basename);
  return MQ_ERROR;
}

MQ_INLINE enum MqErrorE
MqErrorStackPM(
  struct MqS const * const msgque,
  char const * const func,
  char const * const file
)
{
  return MqErrorStackP(msgque->error, func, file);
}

enum MqErrorE
MqErrorSAppendV (
  struct MqErrorS const * const error,
  MQ_STRB const * const fmt,
  ...
)
{
  va_list ap;
  if (error == MQ_ERROR_IGNORE) return MQ_ERROR;
  va_start (ap, fmt);
  if (error) {
    struct MqBufferS * text = error->text;
    struct MqS const * const msgque = error->msgque;
    struct MqS const * const master = msgque->config.master;
    if (!error->append) return error->code;
    MqBufferAppendV (text, "\n%c> (%s) ", ( MQ_IS_SERVER (msgque) ?
			(MQ_IS_CHILD (msgque) ? 's' : 'S') : (MQ_IS_CHILD (msgque) ? 'c' : 'C')),
					msgque->config.name); 
    if (master != NULL)
      MqBufferAppendV (master->error->text, "\n%c> (%s) ", ( MQ_IS_SERVER (msgque) ?
			  (MQ_IS_CHILD (msgque) ? 's' : 'S') : (MQ_IS_CHILD (msgque) ? 'c' : 'C')),
					  msgque->config.name); 
    MqBufferAppendVL (text, fmt, ap);
    if (master != NULL)
      MqBufferAppendVL (master->error->text, fmt, ap);
    return error->code;
  } else {
    MqDLogEVL (NULL, __func__, fmt, ap);
    MqLog (stderr, "\n");
  }
  va_end (ap);
  return MQ_ERROR;
}

MQ_INLINE void
pErrorAppendC (
  struct MqErrorS const * const error,
  MQ_STR msg
)
{
  struct MqS const * const master = error->msgque->config.master;
  MqBufferAppendC (error->text, msg);
  if (master != NULL)
    MqBufferAppendC (master->error->text, msg);
}

MQ_INLINE struct MqS *
MqErrorGetMsgque (
  struct MqErrorS const * const error
)
{
  return (error > MQ_ERROR_PRINT ? (struct MqS *) error->msgque : NULL);
}

/*****************************************************************************/
/*                                                                           */
/*                              error_misc                                   */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
MqErrorGetCode (
  struct MqErrorS const * const error
)
{
  return error->code;
}

MQ_INLINE enum MqErrorE
MqErrorGetCodeM (
  struct MqS const * const msgque
)
{
  return msgque->error->code;
}

MQ_INLINE enum MqErrorE
MqErrorSetCode (
  struct MqErrorS * const error,
  enum MqErrorE code
)
{
  return (error->code = code);
}

MQ_INLINE MQ_STR
MqErrorGetText (
  struct MqErrorS const * const error
)
{
  return (MQ_STR) error->text->data;
}

MQ_INLINE MQ_STR
MqErrorGetTextM (
  struct MqS const * const msgque
)
{
  return (MQ_STR) msgque->error->text->data;
}

MQ_INLINE void
MqErrorReset (
  struct MqErrorS * const error
)
{
  MqBufferReset (error->text);
  error->code = MQ_OK;
  error->append = MQ_YES;
  error->num = EXIT_SUCCESS;
}

MQ_INLINE void
MqErrorResetM (
  struct MqS const * const msgque
)
{
  MqErrorReset(msgque->error);
}

MQ_INLINE MQ_INT
MqErrorGetNum (
  struct MqErrorS const * const error
)
{
  return error->num;
}

void
pErrorSync (
  struct MqErrorS * const out,
  struct MqErrorS * const in
)
{
  MqBufferCopy (out->text, in->text);
  out->code = in->code;
  out->num = in->num;
  out->append = in->append;
}

enum MqErrorE
MqErrorCopy (
  struct MqErrorS * const out,
  struct MqErrorS * const in
)
{
  if (out != in) {
    pErrorSync (out, in);
    MqErrorReset(in);
  }

  return out->code;
}

// report parent error
void
pErrorReport(
  struct MqErrorS * const error
)
{
  struct MqS const * const msgque = error->msgque;
  if (error == NULL) return;
  if (msgque == NULL) return;
  if (MQ_IS_CHILD (msgque) || error->code != MQ_ERROR) return;
  if (MQ_IS_SERVER (msgque) && pIoCheck (msgque->io)) {
    // save the original error
    MQ_BUF err = MqBufferDup (error->text);
    // send the error to the client
    if (MqSendERROR (msgque->send) == MQ_ERROR) {
      if (!msgque->config.isSilent) {
	MqLog (stderr, "%s\n", (MQ_STR) err->data);
      }
    }
  } else if (msgque->config.isSilent == MQ_NO && msgque->config.master == NULL) {
    MqLog (stderr, "%s\n", (MQ_STR) error->text->data);
  }
  MqErrorReset(error);
}

/*****************************************************************************/
/*                                                                           */
/*                                debugging                                  */
/*                                                                           */
/*****************************************************************************/

#ifdef _DEBUG
void
MqErrorLog (
  struct MqErrorS const * const error,
  MQ_STRB const * const prefix
)
{
  struct MqS const * const msgque = MqErrorGetMsgque (error);
  MqDLogV (msgque, prefix, 0, ">>>> MqErrorS (%p)\n", error);
  MqDLogV (msgque, prefix, 0, "msgque   = <%p>\n", MqErrorGetMsgque (error));
  MqDLogV (msgque, prefix, 0, "status   = <%s>\n", MqLogStatus (error->code));
  MqDLogV (msgque, prefix, 0, "num      = <%i>\n", error->num);
  MqBufferLog (msgque, error->text, "error->text");
  MqDLogV (msgque, prefix, 0, "<<<< MqErrorS\n");
}
#endif /* _DEBUG */


