/**
 *  \file       src/event.c
 *  \brief      \$Id: event.c 1205 2009-05-20 10:51:09Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1205 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "event.h"
#include "sys.h"
#include "sys_com.h"
#include "io_private.h"

#define MSGQUE msgque

/// \brief everything \e io need for local storage
struct MqEventS {
  fd_set fdset;			///< used in 'select'
  int fdmax;			///< the maximum file-handle in \e fdset
  struct MqS ** DataL;		///< list on #MqS objects
  int DataLNum;			///< size of \e DataL
  int DataLCur;			///< first free position in \e DataL, <TT>DataLCur <= DataLNum</TT>
  int refCount;			///< how often is this data used?
};

#if defined(_MSC_VER)
#define PREFIX
#else
#define PREFIX static
PREFIX void sEventCreate (
) __attribute__ ((constructor));

PREFIX void sEventDelete (
) __attribute__ ((destructor));
#endif /* !_MSC_VER */

//######################################################################

static void
sEventDeleteAllClient (
  struct MqEventS * const event
)
{
  // process the results
  struct MqS *msgque;
  struct MqS **start, **end;

  // are sockets available? if not do nothing
  if (event == NULL || event->fdmax < 0) return;

  // attention, every delete on a parent-client change the event->DataL too (as side-effect)
  while (event->DataLCur) {
    start = event->DataL;
    end   = event->DataL + event->DataLCur;
    for (; start < end; start++) {
      msgque = *start;
      if (MQ_IS_CLIENT(msgque)) {
	MqDLogC(msgque,4,"FORCE delete\n");
	//MqLogMsgque(msgque,__func__);
	MqDelete(&msgque);
	// exit "for" loop to accept new "start" and "end"
	break;
      }
    }
    // list has no CLIENT ... exit "while" loop
    if (start == end) break;
  }
}

#if defined (MQ_HAS_THREAD)

/// \brief Thread-Specific-Data (TSD) key
#if defined(HAVE_PTHREAD)
static pthread_key_t event_key = 0L;
#else
static DWORD event_key = 0L;
#endif

/// \brief Check (and if not available) allocate TLS storage
/// \error
/// \param[out] eventP the new poiter to the \e event object
/// \retMqErrorE
static enum MqErrorE
sEventAlloc(
  struct MqErrorS * const error,
  struct MqEventS ** eventP
)
{
  struct MqEventS * sysevent = NULL;
  MqErrorReset(error);
#if defined(HAVE_PTHREAD)
  if ((sysevent = (struct MqEventS *) pthread_getspecific(event_key)) == NULL) {
#else
  if ((sysevent = (struct MqEventS *) TlsGetValue(event_key)) == NULL) {
#endif
    sysevent = (struct MqEventS *) MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (*sysevent));
    sysevent->fdmax = -1;
    sysevent->refCount = 0;
#if defined(HAVE_PTHREAD)
    if (pthread_setspecific(event_key,sysevent)) {
      return MqErrorDbV2 (error,MQ_ERROR_CAN_NOT, "pthread_setspecific");
    }
#else
    if (TlsSetValue(event_key,sysevent) == 0) {
      return MqErrorDbV2 (error,MQ_ERROR_CAN_NOT, "TlsSetValue");
    }
#endif
  }
  sysevent->refCount++;
  *eventP = sysevent;
  //MqDLog(MqErrorGetMsgque(error), 4, "create EVENT<%p>\n", sysevent);
  return MQ_OK;
}

/// \brief decrement refCounter and free storage if refCounter == 0
void
sEventFree(
  struct MqEventS ** eventP	    /// [out] the "event" object set to \c NULL
)
{
  struct MqEventS * const event = *eventP;

  if (unlikely(eventP == NULL || event == NULL)) 
    return;

  // cleanup eventP
  event->refCount--;
  if (unlikely(event->refCount == 0))  {
    MqSysFree(*eventP);
#if defined(HAVE_PTHREAD)
    pthread_setspecific(event_key,NULL);
#else
    TlsSetValue(event_key,NULL);
#endif
  } else {
    *eventP = NULL;
  }
  return;
}

/// \brief one-time initializer for the TSD key
PREFIX void
sEventCreate()
{
#if defined(HAVE_PTHREAD)
  if (pthread_key_create(&event_key, NULL) != 0) {
    MqPanicC(MQ_ERROR_PANIC,__func__,-1,"unable to 'pthread_key_create'");
  }
#else
  if ((event_key = TlsAlloc()) == TLS_OUT_OF_INDEXES) {
    MqPanicC(MQ_ERROR_PANIC,__func__,-1,"unable to 'TlsAlloc'");
  }
#endif
}

/// \brief one-time initializer for the TSD key
PREFIX void
sEventDelete()
{
  struct MqEventS * sysevent;
  if (event_key == 0L) return;
#if defined(HAVE_PTHREAD)
  sysevent = (struct MqEventS *) pthread_getspecific(event_key);
#else
  sysevent = (struct MqEventS *) TlsGetValue(event_key);
#endif

  if (sysevent != NULL) {
    sEventDeleteAllClient(sysevent);
    sEventFree(&sysevent);
  }

#if defined(HAVE_PTHREAD)
  pthread_key_delete(event_key);
  event_key = 0L;
#else
  TlsFree(event_key);
  event_key = 0L;
#endif
}
#undef PREFIX

#if defined(HAVE_PTHREAD)
#  define MqThreadSelf() pthread_self()
#else
#  define MqThreadSelf() GetCurrentThreadId()
#endif

//######################################################################
#else	/* ! MQ_HAS_THREAD */

struct MqEventS * sysevent = NULL;

static enum MqErrorE
sEventAlloc(
  struct MqErrorS * const error,    /// [in] the "error" object in duty
  struct MqEventS ** eventP	    /// [out] the "event" object to return
)
{
  MqErrorReset(error);
  if (sysevent == NULL) {
    sysevent = (struct MqEventS*) MqSysCalloc(MQ_ERROR_PANIC, 1, sizeof (*sysevent));
    sysevent->fdmax = -1;
    sysevent->refCount = 0;
  }
  sysevent->refCount++;
  *eventP = sysevent;
  return MQ_OK;
}

static void
sEventFree(
  struct MqEventS ** eventP	    /// [out] the "event" object set to \c NULL
)
{
  if (unlikely(eventP == NULL || (*eventP) == NULL)) return;
  struct MqEventS * const event = *eventP;
  event->refCount--;
  if (unlikely(event->refCount == 0))  {
    MqSysFree(*eventP);
    sysevent = NULL;
  } else {
    *eventP = NULL;
  }
  return;
}

static void
sEventCreate()
{
}

static void
sEventDelete()
{
  if (sysevent != NULL) {
    sEventDeleteAllClient(sysevent);
    sEventFree(&sysevent);
  }
}

#define MqThreadSelf() 0L

#endif   /* ! MQ_HAS_THREAD */

//######################################################################
/// \attention only call this function for a \e PARENT context
/// \attention this function set "msgque->io->sockP"
void
pEventAdd (
  struct MqS * const msgque,
  MQ_SOCK * const sockP
)
{
  const MQ_SOCK sock = *sockP;
  register struct MqEventS * event = msgque->io->event;

  // filter to check if work is available
  if (	    
        !msgque ||              // no msgque no data
        MQ_IS_CHILD(msgque)     // we only add PARENT to the event-list
     ) return;

  // check if enough space is available
  if (event->DataLCur >= event->DataLNum) {
    event->DataLNum += 10;
    event->DataL = (struct MqS **) MqSysRealloc (msgque->error, event->DataL, 
        sizeof (*event->DataL) * event->DataLNum);
  }
  // ATTENTION: we don't check on double insert !!
  event->DataL[event->DataLCur] = msgque;

  if (sock > event->fdmax)
    event->fdmax = sock;

  if (sock >= 0) {
    FD_SET (sock, &event->fdset);

    // add the link between "io" and the current socket handle
    msgque->io->sockP = sockP;
  }

  event->DataLCur++;

  MqDLog(msgque,4,"sock<%i> DataLCur<%i>\n", sock, event->DataLCur);

//pEventLog(msgque, event, __func__);
}

/// brief delete all events
void
sEventReset(
  struct MqEventS * event
)
{
  MqSysFree (event->DataL);
  event->DataLNum = 0;
  event->DataLCur = 0;
  event->fdmax = -1;
  FD_ZERO(&event->fdset);
}

enum MqErrorE
pEventCheck (
  struct MqErrorS * const error,
  struct MqEventS * event,
  enum MqIoSelectE const typ,
  struct timeval * const timeout
)
{
  // init
  fd_set fds, *readfds = NULL, *writefds = NULL;

  // are sockets available? if not do nothing
  if (event == NULL || event->fdmax < 0)
    return MQ_CONTINUE;

  // fill fd_set for use in 'select'
  memcpy (&fds, &event->fdset, sizeof (fd_set));

  // composit
  switch (typ) {
    case MQ_SELECT_RECV:
      readfds = &fds;
      break;
    case MQ_SELECT_SEND:
      writefds = &fds;
      break;
  }

/*
{
  struct MqS const * const msgque = MqErrorGetMsgque(error);
  MqLogData(msgque, "select");
  if (MQ_IS_SERVER(msgque)) {
    int i;
    for (i=0; i<=event->fdmax; i++) {
      if (FD_ISSET(i,&fds))
        MqDLogV(msgque,__func__,0,"sock<%i>, set<YES>\n", i);
    }
  }
}
*/

  // do the select
  return SysSelect (error, event->fdmax + 1, readfds, writefds, timeout);
}

enum MqErrorE
pEventStart (
  struct MqErrorS * const error,
  struct MqEventS * event,
  struct timeval const * const timeout,
  EventCreateF const proc
)
{
  // init
  fd_set fds;
  register struct MqS ** start;
  register struct MqS ** end;
  register MQ_SOCK sock;

  // are sockets available? if not do nothing
  if (!event || event->fdmax < 0)
    return MQ_CONTINUE;

  // fill fd_set for use in 'select'
  memcpy (&fds, &event->fdset, sizeof (fd_set));

  // do the select and process the results
  switch (SysSelect (error, event->fdmax + 1, &fds, NULL, timeout)) {
    case MQ_CONTINUE:
      return MQ_CONTINUE;
    case MQ_OK:
      break;
    case MQ_ERROR:
      goto error;
    case MQ_EXIT:
      return MQ_EXIT;
  }

  // process the results
  start = event->DataL;
  end = event->DataL + event->DataLCur;
  for (; start < end; start++) {
    sock = *((*start)->io->sockP);
    // is the context still in duty ? (sock >= 0)
    if (sock < 0 || !FD_ISSET (sock, &fds)) continue;
    // found valid socket -> call it
    switch ((*proc)(*start)) {
      case MQ_OK:
	return MQ_OK;
      case MQ_ERROR: {
	struct MqS * const msgque = MqErrorGetMsgque(error);
	struct MqS * const call_msgque = *start;
	struct MqErrorS * const call_error = call_msgque->error;
	// the following error's have to be handled:
	// 1. error happen in the same context as "pEventStart" was called
	// 2. "pEventStart" is called on a server context and get an error on a 
	//	client context
	// 3. error happen in the server context
	// 4. error happen in different client context	

	if (error == call_error) {
	  // case 1.
	  // do nothing just report the error
	  goto error;
	} else if (MQ_IS_SERVER(msgque) && MQ_IS_CLIENT(call_msgque)) {
	  // case 2.
	  // copy the error from the client to the server context
	  MqErrorCopy(error, call_error);
	  // send the error to the server-client and reset the server error
	  MqSendERROR(msgque->send);
	  break;
	} else if (MQ_IS_SERVER(call_msgque)) {
	  // case 3. -> a server error have to be reported
	  // we have an MQ error
	  // check to be in a transaction or not
	  if (MqIsTransaction(call_msgque) == MQ_NO) {
	    // no transaction
	    // return server error from a MqSendEND call asyncrone -> reset the error
	    MqSendERROR (call_msgque->send);
	  } else {
	    // in a transaction
	    // return a "server" error using a "_RET" package -> reset the error
	    MqSendSTART(call_msgque->send);
	    MqSendRETURN (call_msgque->send);
	  }
	  // error was "reset" and "pEventStart" context is clean
	  break;
	} else {
	  // copy the error to the "pEventStart" context
	  MqErrorCopy(error, call_error);
	  goto error;
	}
      }
      case MQ_EXIT: {
	// the MQ_EXIT is handeled here becasue this is the only place who
	// know the caller object (throught error) and the exit object (*start)
	struct MqS * delmsgque  = (*start)->exit_msgque;
	struct MqS * const callmsgque = MqErrorGetMsgque(error);
	(*start)->exit_msgque = NULL;
	// case 0. on Shutdown nothing should happen
	if (callmsgque->onShutdown == MQ_YES)
	  return MQ_CONTINUE;
	// security check (propably not used)
	if (delmsgque == NULL) MqPanicSYS(error);
	// case 1. delete myself
	if (callmsgque == delmsgque)  {
	  // can not delete myself -> upper code have to handle this
	  return MQ_EXIT;
	}
	if (MQ_IS_SERVER(delmsgque)) {
	  if (MQ_IS_CHILD(delmsgque)) {
	    // case 2. delmsgque is a server-child and not in use
	    MqDelete(&delmsgque);
	    return MQ_OK;
	  } else {
	    // case 3. delmsgque is a server-parent called by some other object
	    // problem if callmsgque is a child of delmsgque than MqDelete(&delmsgque)
	    // will delete callmsgque too
	    if (pMqGetFirstParent(callmsgque) == delmsgque) {
	      // problem: after return the address of delmsgque is lost -> this is an error
	      return MqErrorC(error,__func__,1,"unable to delete recursive object");
	    } else {
	      MqDelete(&delmsgque);
	      return MQ_OK;
	    }
	  }
	} else {
	  // case 4 delmsgque is a client -> the client is allways delete 
	  // at the end of the client-executable
	  return MQ_EXIT;
	}
      }
      case MQ_CONTINUE:	  // never used
	break;
    }
    // allways break after one was found, event->DataL could be invalid
    break;
  }

  return MQ_OK;

error:
  return MqErrorStack(error);
}

enum MqErrorE
pEventCreate (
  struct MqS const * const msgque,
  struct MqEventS ** const eventP
) {
  struct MqEventS * sysevent = NULL;

  // get/create the TLS data
  MqErrorCheck(sEventAlloc(msgque->error, &sysevent));

  // if the context was created by a "fork" cleanup events first
  if (MQ_IS_PARENT(msgque) && msgque->config.statusIs & MQ_STATUS_IS_FORK) {
    sEventReset(sysevent);
    sysevent->refCount = 1;
  }

  // report event back
  if (likely (eventP != NULL))
    *eventP = sysevent;

  MqDLog(msgque,4,"TSD-Id<%p>, Thread-Id<0x%lx>\n", sysevent, MqThreadSelf());
  return MQ_OK;

error:
  return MqErrorStack(msgque->error);
}

/// \attention only call this function for a \e PARENT context
/// \attention this function free "msgque->io->sockP"
void
pEventDelete (
  struct MqS const * const msgque
)
{
  struct MqEventS * event;
  struct MqS ** start;
  struct MqS ** end;
  int max = -1;  
  MQ_SOCK sock = -1, delsock = -1;

  // filter to check if work is available
  if (	    
        !msgque ||		// no "msgque" data
        !msgque->io ||		// no "io" data
        !msgque->io->event ||   // no "event" data
        MQ_IS_CHILD(msgque)	// we only delete PARENT from the event-list
     ) return;

  event = msgque->io->event;
  start = event->DataL;
  end = event->DataL + event->DataLCur;
  
//pEventLog(msgque, event, "pEventDelete-1");

  // find + clear
rescan:
  for (; start < end; start++) {
    sock = *((*start)->io->sockP);
    if (*start == msgque) {
      delsock = (sock < 0 ? -sock : sock);
      // free fdset
      FD_CLR (delsock, &event->fdset);
      // move following data 'one' to the left
      memmove(start,start+1,(end-start-1)*sizeof(*start));
      // decrease end
      end -= 1;
      event->DataLCur--;
      // we have shift the date one to the left -> get the new socket
      goto rescan;
    }
    // find the new max
    if (sock > max) {
      max = sock;
    }
  }

  // cleanup
  event->fdmax = max;


  // last one -> delete the infrastructure
  if (event->DataLCur == 0) {
    sEventReset(event);
  }

  // if SERVER delete all CLIENT too
  if (MQ_IS_SERVER_PARENT(msgque)) {
    //pEventLog(msgque, event, "pEventDelete-2");
    sEventDeleteAllClient(event);
  }

  // report status back
  MqDLog(msgque,4,"sock<%i> DataLCur<%i>\n", delsock, event->DataLCur);

//pEventLog(msgque, event, "pEventDelete-2");

  // free memory (if event->refCount == 0)
  sEventFree(&msgque->io->event);
}

/*****************************************************************************/
/*                                                                           */
/*                                debugging                                  */
/*                                                                           */
/*****************************************************************************/

#ifdef _DEBUG
void pEventLog (
  struct MqS const * const msgque,
  struct MqEventS * event,
  MQ_STRB const * const proc
)
{
  struct MqS ** start, ** end;
  if (!msgque || !event || msgque->config.isSilent == MQ_YES) return;
  MqDLogV(msgque, proc, 0, ">>>>> MqEventS<%p>\n", event);
  MqDLogV(msgque, proc, 0, "fdmax    = %i\n", event->fdmax);
  MqDLogV(msgque, proc, 0, "DataLNum = %i\n", event->DataLNum);
  MqDLogV(msgque, proc, 0, "DataLCur = %i\n", event->DataLCur);
  MqDLogV(msgque, proc, 0, "refCount = %i\n", event->refCount);
  start = event->DataL;
  end = event->DataL + event->DataLCur;
  for (; start < end; start++) {
    struct MqS * mq = *start;
    MqDLogV(msgque, proc, 0, " -> msgque<%p>, server<%s>, sock<%i>\n", mq, 
	(MQ_IS_SERVER(mq)?"yes":"no"), *(mq->io->sockP));
  }
  MqDLogV(msgque, proc, 0, "<<<<< MqEventS<%p>\n", event);
}
#endif /* _DEBUG */

