/**
 *  \file       src/slave.c
 *  \brief      \$Id: slave.c 140 2009-07-02 11:45:42Z dev1usr $
 *  
 *  (C) 2009 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 140 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "slave.h"
#include "error.h"

#define MSGQUE slave->msgque
#define MQ_ERROR_S msgque->error

struct MqSlaveS {
  struct MqS * msgque;		    ///< link to the msgque object

  // slaves link
  struct MqS ** slaves;		    ///< SLAVE: \c NULL, MASTER a list of pointers of \e slave object 
  MQ_SIZE used;			    ///< number of slave objects in slaves[]
  MQ_SIZE size;			    ///< size of slaves[]
};

/*****************************************************************************/
/*                                                                           */
/*                                C R E A T E                                */
/*                                                                           */
/*****************************************************************************/

static enum MqErrorE
pSlaveChildCreate (
  struct MqSlaveS * const   slave,
  MQ_SIZE		    id,
  struct MqS * const	    slave_msgque
)
{
  struct MqS * const  msgque = slave->msgque;

  // ERROR, ID is no valid identifer
  if (id < 0 || id >= MQ_SLAVE_MAX)
    return MqErrorDbV2(msgque->error, MQ_ERROR_ID_OUT_OF_RANGE, "slave", 0, id, MQ_SLAVE_MAX-1);

  // ERROR, if ID is allready used
  if (slave->slaves && id >= 0 && id < slave->used && slave->slaves[id] != NULL)
    return MqErrorDbV2(msgque->error, MQ_ERROR_ID_IN_USE, "slave", id);

  // get the new max
  if (id >= slave->used)
    slave->used = (id+1);

  // everything is OK, 1. check if slave array is available and has enough space
  if (slave->slaves == NULL) {
    slave->size = 10;
    slave->slaves = (struct MqS **) MqSysCalloc(MQ_ERROR_PANIC, 10, sizeof(struct MqS*));
  }

  // array to small ?
  if (slave->size < slave->used) {
    MQ_SIZE oldsize = slave->size;
    slave->size = (2*slave->used > MQ_SLAVE_MAX ? MQ_SLAVE_MAX : 2*slave->used);
    slave->slaves = (struct MqS **) MqSysRealloc(MQ_ERROR_PANIC, slave->slaves, slave->size * sizeof(struct MqS*));
    memset(slave->slaves+oldsize, 0, slave->size-oldsize);
  }

  slave->slaves[id] = slave_msgque;
  slave_msgque->config.master = msgque;
  slave_msgque->config.master_id = id;
  return MQ_OK;
}

static enum MqErrorE
pSlaveParentCreate (
  struct MqSlaveS * const   slave,
  MQ_SIZE		    id,
  struct MqS * const	    slave_msgque
)
{
  struct MqS * const  msgque = slave->msgque;

  // ERROR, if MASTER-MSGQUE is a CHILD or if it is a PARENT with CHILDs available
  if (MQ_IS_CHILD(msgque) || msgque->childs != NULL)
    return MqErrorDbV(MQ_ERROR_MASTER_SLAVE, "master");

  // ERROR, if SLAVE-MSGQUE is a SERVER or is a CHILD or if it is a PARENT with CHILDs available
  if (MQ_IS_SERVER(slave_msgque) || MQ_IS_CHILD(slave_msgque) || slave_msgque->childs != NULL)
    return MqErrorDbV(MQ_ERROR_MASTER_SLAVE_CLIENT, "slave");

  return pSlaveChildCreate(slave, id, slave_msgque);
}

static void
pSlaveItemDelete (
  struct MqSlaveS const * const slave,
  MQ_SIZE id
)
{
  struct MqS * slave_msgque;
  if (slave->slaves == NULL || id < 0 || id >= slave->used || slave->slaves[id] == NULL) return;
  // attention !! SLAVE set "slave->slaves[id]" use a teporary variable as work-around
  slave_msgque = slave->slaves[id];
  MqDLog(slave->msgque, 5, "delete SLAVE <%p>\n", slave_msgque);
  MqDelete (&slave_msgque);
  slave->slaves[id] = NULL;
}

enum MqErrorE
pSlaveCreate (
  struct MqS * const msgque,
  struct MqBufferLS * const argv,
  struct MqBufferLS ** const alfaP,
  struct MqSlaveS ** const out
)
{
  struct MqErrorS * error = msgque->error;
  struct MqSlaveS * const slave = (struct MqSlaveS *) MqSysCalloc (error, 1, sizeof (*slave));
  if (unlikely(slave == NULL)) goto error;
  if (likely (out != NULL)) *out = slave;
  slave->msgque = msgque;

  // if child ... create all slave object-childs of parent too
  if (MQ_IS_CHILD(msgque)) {

    struct MqS * const master = msgque;
    struct MqS const * const parent = msgque->config.parent;
    struct MqSlaveS * parent_slave = parent->slave;
    MQ_SIZE id;

    // I'm a CHILD of a MASTER -> add a CHILD to all of my SLAVES too
    for (id=0; id<parent_slave->used; id++) {
      MqCreateF fChildCreate;
      struct MqConfigS * config;
      struct MqS * slave_parent = parent_slave->slaves[id];
      struct MqS * slave_msgque;
      // check id PARENT has a SLAVE with ID
      if (slave_parent == NULL) continue;
      // create the CHILD of a SLAVE
      // a SLAVE is allways a CLIENT
      config = MqConfigDup(&slave_parent->config);
      config->parent = slave_parent;
      config->master = master;
      config->myCtx = NULL;
      fChildCreate = (config->fChildCreate == NULL ? MqCreate : config->fChildCreate);
      // create the child
      if ((*fChildCreate) (config, NULL, NULL, &slave_msgque) == MQ_ERROR) {
	MqBufferLDelete(alfaP);
	MqErrorCopy(msgque->error, slave_msgque->error);
	MqDelete(&slave_msgque);
	goto error;
      }
      // add me as slave
      if (pSlaveChildCreate(slave, id, slave_msgque) == MQ_ERROR) {
	MqBufferLDelete(alfaP);
	MqDelete(&slave_msgque);
	goto error;
      }
    }

  } else {

    // add my link to the master
    if (msgque->config.master != NULL) {
      struct MqS * const master = msgque->config.master;

      // ERROR, if MASTER is not a SERVER
      if (MQ_IS_CLIENT(master)) {
	MqBufferLDelete(alfaP);
	return MqErrorDb(MQ_ERROR_WORKER_IS_NO_SERVER);
      }

      // ERROR, if MASTER is NOT a PARENT or is a PARENT and CHILDS are available
      if (MQ_IS_CHILD(master) || master->childs != NULL) {
	MqBufferLDelete(alfaP);
	return MqErrorDbV(MQ_ERROR_MASTER_SLAVE, "master");
      }

      // if alfaP == NULL than create and initialize it
      if (*alfaP == NULL || (*alfaP)->cursize == 0) {
	if (*alfaP == NULL) *alfaP = MqBufferLCreate(1);
	MqBufferLAppendC(*alfaP, "WORKER");
	msgque->isWORKER = MQ_YES;
      }
  
      // ERROR, if MASTER is set but alfa[0] is not a "WORKER"
      if (msgque->isWORKER == MQ_NO) {
	MqErrorDbV(MQ_ERROR_INVALID_ARGUMENT, "WORKER", 
	  ((*alfaP) != NULL && (*alfaP)->cursize > 0 ? (*alfaP)->data[0]->cur.C : "empty"));
	MqBufferLDelete(alfaP);
	return MQ_ERROR;
      }

      // create a LINK between MYSELF and the MASTER
      MqErrorCheck (pSlaveChildCreate(master->slave, msgque->config.master_id, msgque));
    }
  }

error:
  return MqErrorGetCode (error);
}

void
pSlaveShutdown (
  struct MqSlaveS * const slave
)
{
  MQ_SIZE id;
  if (unlikely(slave == NULL)) return;
  for (id=0; id<slave->used; id++) {
    pSlaveItemDelete(slave, id);
  }
}

void
pSlaveDelete (
  struct MqSlaveS ** const slaveP
)
{
  struct MqSlaveS * const slave = *slaveP;
  struct MqS * msgque;

  if (unlikely(slave == NULL)) return;               // nothing to do
  msgque = slave->msgque;

  // delete my link(s) to the slaves
  MqSysFree(slave->slaves);

  // delete my link to the master
  if (msgque->config.master != NULL) {
    if (msgque->config.master_id < msgque->config.master->slave->used)
      msgque->config.master->slave->slaves[msgque->config.master_id] = NULL;
    msgque->config.master = NULL;
    msgque->config.master_id = 0;
  }

  // delete the slave object
  MqSysFree(*slaveP);
}

/*****************************************************************************/
/*                                                                           */
/*                                P U B L I C                                */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
MqSlaveCreate (
  struct MqS * const  msgque,
  MQ_SIZE	      id,
  struct MqS * const  slave_msgque
)
{
  return pSlaveParentCreate(msgque->slave, id, slave_msgque);
}

enum MqErrorE
MqSlaveDelete (
  struct MqS * const  msgque,
  MQ_SIZE   id
)
{
  // ERROR, if MASTER-MSGQUE is a CHILD
  if (MQ_IS_CHILD(msgque))
    return MqErrorDb(MQ_ERROR_SLAVE_CHILD_DELETE);
  pSlaveItemDelete(msgque->slave, id);
  MqErrorReset(msgque->error);
  return MQ_OK;
}

struct MqS *
MqSlaveGet (
  struct MqS * const  msgque,
  MQ_SIZE   id
)
{
  return (id < 0 || id >= msgque->slave->used ? NULL : msgque->slave->slaves[id]);
}

