/**
 *  \file       src/sys.c
 *  \brief      \$Id: sys.c 1194 2009-05-19 15:44:34Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1194 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"

#include "sys.h"
#include "error.h"
#include "mq_io.h"
#include "log.h"

#include <errno.h>
#include <fcntl.h>

#if !defined(_MSC_VER)
#  include <libgen.h>
#endif

#ifdef MQ_IS_WIN32
# if !defined(_MSC_VER)
//	we require WindowsXP or higher -> used for "getaddrinfo"
#   define WINVER WindowsXP
#   include <w32api.h>
# endif
#   include <windows.h>
#   include <process.h>
#else
#   include <sys/wait.h>
#  if defined(HAVE_POSIX_SPAWN)
#   include <spawn.h>
#  endif
#  if defined(HAVE_SIGNAL_H)
#   include <signal.h>
#  endif
#endif

#if defined(MQ_HAS_THREAD)
# if defined(HAVE_PTHREAD)
#  define mqthread_ret_t void*
#  define mqthread_stdcall
# elif defined(MQ_IS_WIN32)
#  define mqthread_ret_t unsigned
#  if defined(_MANAGED)
//   MS awitch "/clr" is active
#    define mqthread_stdcall
#  else
#    define mqthread_stdcall __stdcall
#  endif
# endif
#endif /* MQ_HAS_THREAD */

#define MqErrorSys(cmd) \
  MqErrorV (error, __func__, errno, \
    "can not '%s' -> ERR<%s>", MQ_CPPXSTR(cmd), strerror (errno))

#define MqErrorSysAppend(cmd) \
  MqErrorSAppendV (error,\
    "can not '%s' -> ERR<%s>", MQ_CPPXSTR(cmd), strerror (errno))

#define MQ_ERROR_S error

#define MSGQUE MqErrorGetMsgque(error)

/*****************************************************************************/
/*                                                                           */
/*                               sys_memory                                  */
/*                                                                           */
/*****************************************************************************/

MQ_PTR 
MqSysCalloc (
  struct MqErrorS * const error,
  MQ_SIZE nmemb,
  MQ_SIZE size
)
{
  MQ_PTR ptr = calloc (nmemb, size);
  if (unlikely (ptr == NULL))
    MqErrorC (error, __func__, errno, strerror (errno));
  return ptr;
}

MQ_PTR 
MqSysMalloc (
  struct MqErrorS * const error,
  MQ_SIZE size
)
{
  MQ_PTR ptr = malloc (size);
  if (unlikely (ptr == NULL))
    MqErrorC (error, __func__, errno, strerror (errno));
  return ptr;
}

MQ_PTR
MqSysRealloc (
  struct MqErrorS * const error,
  MQ_PTR buf,
  MQ_SIZE size
)
{
  MQ_PTR ptr = realloc (buf, size);
  if (unlikely (ptr == NULL))
    MqErrorC (error, __func__, errno, strerror (errno));
  return ptr;
}

/*****************************************************************************/
/*                                                                           */
/*                                sys_env                                    */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
SysGetEnv (
  struct MqErrorS * const error,
  const char *str,
  MQ_STR *ret
)
{
  if ((*ret = (MQ_STR) getenv (str)) == NULL) {
    return MqErrorV (error, __func__, errno, "can not read the environment variable \"%s\"", str);
  }
  return MQ_OK;
}

/*
MQ_INLINE enum MqErrorE
SysSetEnv (
  struct MqErrorS * const error,
  const char *name,
  const char *str
)
{
  if (unlikely (setenv (name, str, 1))) {
    return MqErrorSys(setenv);
  }
  return MQ_OK;
}

MQ_INLINE enum MqErrorE
SysUnSetEnv (
  struct MqErrorS * const error,
  const char *name
)
{
#ifdef HAVE_UNSETENV_RETURN
  if (unlikely (unsetenv (name))) {
    return MqErrorSys(unsetenv);
  }
#else
  unsetenv (name);
#endif
  return MQ_OK;
}
*/

/*****************************************************************************/
/*                                                                           */
/*                                sys_file                                   */
/*                                                                           */
/*****************************************************************************/

/*
MQ_INLINE enum MqErrorE
SysOpen (
  struct MqErrorS * const error,
  MQ_STR pathname,
  int flags,
  mode_t mode,
  MQ_INT *fh
)
{
  if (unlikely ((*fh = open (pathname, flags, mode)) == -1)) {
    return MqErrorV (error, __func__, errno,
                     "can not open file <%s> -> ERR<%s>", pathname, strerror (errno));
  }
  return MQ_OK;
}

MQ_INLINE enum MqErrorE
SysStat (
  struct MqErrorS * const error,
  const MQ_STR file,
  struct stat * st
)
{
  if (unlikely (stat (file, st) == -1)) {
    if (!error || errno == ENOENT) {
      return MQ_CONTINUE;
    } else {
      return MqErrorSys(stat);
    }
  }
  return MQ_OK;
}
*/

/*****************************************************************************/
/*                                                                           */
/*                                sys_io                                     */
/*                                                                           */
/*****************************************************************************/

/*
enum MqErrorE
SysRead (
  struct MqErrorS * const error,
  MQ_INT hdl,
  MQ_BIN buf,
  const MQ_SIZE numBytes
)
{
  register MQ_BIN lbuf = buf;
  register MQ_SIZE lsize = numBytes;
  register MQ_SIZE lread = 0;

  while (lread != lsize) {
    lbuf += lread;
    lsize -= lread;

    // read data in buf
    lread = read (hdl, lbuf, lsize);

    // check for errors
    if (unlikely (lread == -1)) {
      return MqErrorSys (read);
    } else if (unlikely (lread == 0)) {
      return MqErrorDb2 (error,MQ_ERROR_EOF);
    }
  }

  return MQ_OK;
}

enum MqErrorE
SysWrite (
  struct MqErrorS * const error,
  MQ_INT hdl,
  MQ_BIN buf,
  const MQ_SIZE numBytes
)
{
  register MQ_BIN  lbuf = buf;
  register MQ_SIZE lsize = numBytes;
  register MQ_SIZE lwrite = 0;

  while (lwrite != lsize) {

    lbuf += lwrite;
    lsize -= lwrite;

    // write data from buf
    lwrite = write (hdl, lbuf, lsize);

    // check for errors
    if (unlikely (lwrite == -1)) {
      return MqErrorSys (write);
    }
  }

  return MQ_OK;
}

enum MqErrorE
SysClose (
  struct MqErrorS * const error,
  MQ_INT *hdl
)
{
  if (*hdl == -1)
    return MQ_OK;
	if (unlikely (close (*hdl) == -1)) {
		return MqErrorV (error, __func__, errno, "can not close fileHdl ERR<%s>", strerror (errno));
  }
  *hdl = -1;
  return MQ_OK;
}
*/

#if defined(MQ_IS_POSIX)
MQ_INLINE enum MqErrorE
SysUnlink (
  struct MqErrorS * const error,
  const MQ_STR fileName
)
{
  struct stat buf;
  if (unlikely (stat (fileName, &buf) == -1)) {
    switch (errno) {
      case ENOENT:
	return MQ_OK;
      default:
	return MqErrorV (error, __func__, errno,
	  "can not access file <%s> ERR<%s>", fileName, strerror (errno));
    }
  }
  if (unlikely (unlink (fileName) == -1 && errno != ENOENT)) {
    return MqErrorV (error, __func__, errno,
                     "can not delete file <%s> ERR<%s>", fileName, strerror (errno));
  }
  return MQ_OK;
}
#endif

#if defined(_MSC_VER)

#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
#else
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
#endif
 
int gettimeofday(
    struct mq_timeval *tv, 
    struct mq_timezone *tz
)
{
  FILETIME ft;
  unsigned __int64 tmpres = 0;
  static int tzflag;
 
  if (NULL != tv)
  {
    GetSystemTimeAsFileTime(&ft);
 
    tmpres |= ft.dwHighDateTime;
    tmpres <<= 32;
    tmpres |= ft.dwLowDateTime;
 
    /*converting file time to unix epoch*/
    tmpres /= 10;  /*convert into microseconds*/
    tmpres -= DELTA_EPOCH_IN_MICROSECS; 
    tv->tv_sec = (long)(tmpres / 1000000UL);
    tv->tv_usec = (long)(tmpres % 1000000UL);
  }
 
  if (NULL != tz)
  {
    if (!tzflag)
    {
      _tzset();
      tzflag++;
    }
    tz->tz_minuteswest = _timezone / 60;
    tz->tz_dsttime = _daylight;
  }
 
  return 0;
}

#endif /* ! _MSC_VER */

MQ_INLINE enum MqErrorE
SysGetTimeOfDay (
  struct MqErrorS * const error,
  struct mq_timeval * tv,
  struct mq_timezone * tz
)
{
  if (unlikely (gettimeofday (tv, tz) == -1)) {
    return MqErrorSys (gettimeofday);
  }
  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                                sys_misc                                   */
/*                                                                           */
/*****************************************************************************/

#if defined(HAVE_FORK)
enum MqErrorE
SysServerFork (
  struct MqErrorS * const error,	///< [ini,out] error handler
  struct MqConfigS * config,		///< [in,out] server configuration (memroy will be freed)
  struct MqBufferLS * const argv,	///< [in] command-line arguments befor #MQ_ALFA
  struct MqBufferLS * const alfa,	///< [in] command-line arguments after #MQ_ALFA
  MqCreateF fMain,			///< [in] entry-point using \c main syntax
  const char * name,			///< [in] the name of the process
  struct MqIdS * idP			///< [out] the process identifer
)
{
  // this is used for a filter pipeline like "| atool split .. @ cut ... @ join ..."
  // argv[0] is set by the tool (cut or join) and need !not! be set by name
  MqErrorCheck(SysFork(error,idP));
  if ((*idP).val.process == 0) {
    struct MqS * msgque;
    config->statusIs = (enum MqStatusIsE) (config->statusIs | MQ_STATUS_IS_FORK);
    config->isServer = MQ_YES;
    MqErrorCheck1(fMain(config,argv,alfa,&msgque));
    // the MqProcessEvent is necessary becasue "ParentCreate" have to come back
    // ifNot: java create object will fail
    if (msgque != NULL && msgque->config.isServer == MQ_YES)
      MqProcessEvent(msgque, MQ_TIMEOUT, MQ_WAIT_FOREVER);
error1:
    MqExit(msgque);
  }
  // free the config created above
  MqSysFree(config->name);
  MqSysFree(config);
  // do not create a "defunc" process
  MqErrorCheck (SysIgnorSIGCHLD(error));
  return MQ_OK;

error:
  return MqErrorDbV (MQ_ERROR_CAN_NOT_START_SERVER, name);
}
#endif   /* HAVE_FORK */

#if defined(MQ_HAS_THREAD)

struct SysServerThreadCreateS {
  struct MqConfigS * config;
  MqCreateF fMain;
  struct MqBufferLS * argv;
  struct MqBufferLS * alfa;
};

static mqthread_ret_t mqthread_stdcall sSysServerThreadCreate (
  void * data
) 
{
  struct SysServerThreadCreateS * argP = (struct SysServerThreadCreateS *) data;
  // save data local
  struct MqConfigS * config = argP->config;
  struct MqBufferLS * const argv = argP->argv;
  struct MqBufferLS * const alfa = argP->alfa;
  MqCreateF fMain = argP->fMain;
  struct MqS * msgque = NULL;
  // cleanup
  free(argP);
  // add my configuration
  config->statusIs = (enum MqStatusIsE) (config->statusIs | MQ_STATUS_IS_THREAD);
  // do some cleanup ...
  MqErrorCheck(fMain(config, argv, alfa, &msgque));
  //MqDLogC(msgque,0,"THREAD exit");
  // the MqProcessEvent is necessary becasue "ParentCreate" have to come back
  // ifNot: java create object will fail
  if (msgque != NULL && msgque->config.isServer == MQ_YES) {
    MqProcessEvent(msgque, MQ_TIMEOUT, MQ_WAIT_FOREVER);
  }
  // a extension (like tcl) does use an embedded interpreter
  // and does not return a "msgque"
error:
  if (msgque == NULL) {
    SysExit(1,0);
  } else {
    MqExit(msgque);
  }
  return (mqthread_ret_t) NULL;
}

/// \brief start a new thread
enum MqErrorE
SysServerThread (
  struct MqErrorS * const error,    ///< [in,out] error handler
  struct MqConfigS * config,	    ///< [in,out] server configuration (memroy will be freed)
  struct MqBufferLS * const argv,   ///< [in] command-line arguments befor #MQ_ALFA
  struct MqBufferLS * const alfa,   ///< [in] command-line arguments after #MQ_ALFA
  const MqCreateF fMain,	    ///< [in] entry-point using \c main syntax
  const char * name,		    ///< [in] the name of the process
  int state,			    ///< [in] detachstate of the thread \c PTHREAD_CREATE_DETACHED or 
                                    ///< \c PTHREAD_CREATE_JOINABLE
  struct MqIdS * idP		    ///< [out] the thread identifer
)
{
  // without "fMain" the thread has no entry-point
  // this is used for a filter pipeline like "| atool split .. @ cut ... @ join ..."
  // argv[0] is set by the tool (cut or join) and need !not! be set by name
  mqthread_t threadId;
#if defined(HAVE_PTHREAD)
  int ret;
#endif

  // fill thread data
  struct SysServerThreadCreateS * argP = (struct SysServerThreadCreateS *) MqSysMalloc(error,sizeof(*argP));
  argP->config = config;
  argP->fMain = fMain;
  argP->argv = argv;
  argP->alfa = alfa;

  // after a "thread" no "fork" is possible
  config->ignoreFork = MQ_YES;

#if defined(HAVE_PTHREAD)

  // thread attributes
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, state);

  // start thread
  do {
    ret = pthread_create(&threadId,&attr,sSysServerThreadCreate,argP);
  } while (ret == EAGAIN);

  if (unlikely(ret != 0)) {
    MqErrorDbV (MQ_ERROR_CAN_NOT_START_SERVER, name);
    return MqErrorSysAppend (pthread_create);
  }

  // cleanup attribute
  pthread_attr_destroy(&attr);

#elif defined(MQ_IS_WIN32)

  if (unlikely ( (threadId = _beginthreadex(NULL, 0, sSysServerThreadCreate, argP, 0, NULL)) == 0)) {
    MqErrorDbV (MQ_ERROR_CAN_NOT_START_SERVER, name);
    return MqErrorSysAppend (_beginthreadex);
  }

#endif

  // save tid
  (*idP).val.thread = threadId;
  (*idP).type = MQ_ID_THREAD;

  return MQ_OK;
}
#endif   /* MQ_HAS_THREAD */

/// \brief spawn a new process
enum MqErrorE
SysServerSpawn (
  struct MqErrorS * const error,	///< [ini,out] error handler
  char * * argv,			///< [in] command-line arguments
  const char * name,			///< [in] the name of the process
  struct MqIdS * idP			///< [out] the process identifer
)
{
  pid_t pid;

#if defined(HAVE_POSIX_SPAWN)
  if (unlikely (posix_spawnp(&pid, name, NULL, NULL, (char *const *) argv, __environ) != 0)) {
    goto error;
  }
  goto ok;

#elif (defined(HAVE_FORK) || defined(HAVE_VFORK))  && defined(HAVE_EXECVP)
  // fork to create the child
#if defined(HAVE_VFORK)
  if (unlikely ((pid = vfork()) == -1)) {
#else /* fork */
  if (unlikely ((pid = fork()) == -1)) {
#endif
    goto error;
  }
  if (pid == 0) {
    // this is the child
    if (execvp(name, argv) == -1) {
      _exit (EXIT_FAILURE);
    }
    // the child will never reach this
  }
  goto ok;

#elif defined(MQ_IS_WIN32)
  char buf[2048];
  char *nbuf=buf;

  // build command-line
  for (;*argv != '\0'; argv++) {
    nbuf += sprintf(nbuf, "\"%s\" ", *argv);
  }

//printC(buf)
//system("set PATH");

  // start process
  if (unlikely ((pid = _spawnlp (_P_NOWAIT, name, buf, NULL)) == -1)) {
    //printC(strerror(errno))
    goto error;
  }
  goto ok;

#else
# error unable to SysServerSpawn
#endif

ok:
  (*idP).val.process = pid;
  (*idP).type = MQ_ID_PROCESS;
  return MQ_OK;

error:
  return MqErrorDbV (MQ_ERROR_CAN_NOT_START_SERVER, name);
}

MQ_INLINE enum MqErrorE
MqSysUSleep (
  struct MqErrorS * const error,
  unsigned int const usec
)
{
#if defined(MQ_IS_POSIX)
  if (unlikely (usleep (usec))) {
    return MqErrorSys (usleep);
  }
#elif defined (MQ_IS_WIN32)
  Sleep ((usec + 500) / 1000);
#else
#   error not supported
#endif
  return MQ_OK;
}

MQ_INLINE enum MqErrorE
MqSysSleep (
  struct MqErrorS * const error,
  unsigned int const sec
)
{
#if defined(MQ_IS_POSIX)
  if (unlikely (sleep (sec))) {
    return MqErrorSys (sleep);
  }
#elif defined (MQ_IS_WIN32)
  Sleep (sec * 1000);
#else
#   error not supported
#endif
  return MQ_OK;
}

#if defined(_MSC_VER)
MQ_STR
MqSysBasename (
  char const * const in,
  MQ_BOL includeExtension
)
{
  char drive[10];
  char dir[512];
  char * fname = (char *) MqSysMalloc(MQ_ERROR_PANIC, 138);
  char ext[10];
  if (_splitpath_s(in, drive, 10, dir, 512, fname, 128, ext, 10) != 0)
    MqPanicV(MQ_ERROR_PANIC, __func__, 1, "unable to extract basename from '%s'", in);
  if (includeExtension)
    return strcat(fname,ext);
  else
    return fname;
}
#else
MQ_STR
MqSysBasename (
  char const * const in,
  MQ_BOL includeExtension
)
{
  if (in == NULL) return mq_strdup("");
  // double mq_strdup is necessary because LINUX may modify the input string
  // and FREEBSD modify the output string in future call's to 'basename'
  // ... Is This POSIX ?
  MQ_STR t1 = mq_strdup((MQ_STR) in);
  const MQ_STR orig = mq_strdup(basename(t1));
  free (t1);
  if (includeExtension == MQ_NO) {
    // delete the extension like '.EXE'
    MQ_STR end;
    for (end = orig + strlen(orig); end > orig; end--) {
      if (*end != '.') continue;
      *end = '\0';
      break;
    }
  }
  return orig;
}
#endif

#ifdef HAVE_FORK
enum MqErrorE
SysFork (
  struct MqErrorS * const error,
  struct MqIdS * idP
)
{
  if (unlikely (((*idP).val.process = fork()) == -1)) {
    return MqErrorSys (fork);
  }
  (*idP).type = MQ_ID_PROCESS;
  return MQ_OK;
}
#endif

#ifdef MQ_IS_POSIX
enum MqErrorE
SysIgnorSIGCHLD (
  struct MqErrorS * const error
)
{
  if (unlikely ((signal(SIGCHLD, SIG_IGN)) == SIG_ERR)) {
    return MqErrorSys (signal);
  }
  return MQ_OK;
}

enum MqErrorE
SysAllowSIGCHLD (
  struct MqErrorS * const error
)
{
  if (unlikely ((signal(SIGCHLD, SIG_DFL)) == SIG_ERR)) {
    return MqErrorSys (signal);
  }
  return MQ_OK;
}
#endif /* MQ_IS_POSIX */

#if ! defined(SysDaemonize)
/// \brief daemonize the current process and save the resulting pid into \e pidfile
enum MqErrorE
SysDaemonize (
  struct MqErrorS * const error,	///< [in,out] error handler
  const MQ_STR pidfile			///< [in] file to save the \e pid
) 
{
  int fd, fpid;
  struct MqIdS id;

  // step 0 -> open the pidfile
  if ((fpid = open(pidfile, O_WRONLY|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1) 
    return MqErrorSys (open);

  // step 1 -> delete initial process
  MqErrorCheck(SysFork(error, &id));
  if (id.val.process != 0)
    _exit(0);

  // step 2 -> create a new session
  if (setsid() == -1)
    return MqErrorSys (setsid);

  // step 3 -> fork the new process, parent write the pid
  MqErrorCheck(SysFork(error, &id));
  if (id.val.process != 0) {
    char buf[20];
    snprintf(buf,20,"%i",id.val.process);
    write(fpid,buf,strlen(buf));
    _exit(0);
  }

  // step 4 -> close the fpid after writing
  if (close(fpid) == -1)
    return MqErrorSys (close);

  // step 5 -> close default io
  fd = open("/dev/null", O_RDONLY);
  if (fd != 0) {
    dup2(fd, 0);
    close(fd);
  }
  fd = open("/dev/null", O_WRONLY);
  if (fd != 1) {
    dup2(fd, 1);
    close(fd);
  }
  fd = open("/dev/null", O_WRONLY);
  if (fd != 2) {
    dup2(fd, 2);
    close(fd);
  }

  return MQ_OK;

error:
  return MqErrorStack(MQ_ERROR_S);
}

#endif /* ! defined(SysDaemonize) */

/// \brief delete file from filesystem (used to delete the pidfile)
MQ_INLINE enum MqErrorE
SysUnLink (
  struct MqErrorS * const error,    ///< [in,out] error handler
  const MQ_STR filename		    ///< [in] filename to unlink
) {
  if (mq_unlink (filename) == -1) {
    return MqErrorV (error, __func__, errno,
      "can not unlink file <%s> -> ERR<%s>", filename, strerror (errno));
  }
  return MQ_OK;
}

/// \brief deep free of an argv array
MQ_INLINE void SysFreeArgvArray (
  char *** argvP		    ///< [in,out] a pointer to the argv array
) {
  char ** arg = *argvP;
  
  if (arg == NULL) return;

  // 1. free the arguments
  for (;*arg != NULL;arg++)
    free(*arg);

  // 2. free the array itself
  free (*argvP);
  *argvP = NULL;
}

/// \brief thread save exit
void SysExit (
  int isThread,			    ///< [in] is this a thread
  int num			    ///< [in] exit number
) {
  if (isThread) {
#if defined(MQ_HAS_THREAD)
# if defined(HAVE_PTHREAD)
    pthread_exit(NULL);
# elif defined(MQ_IS_WIN32)
   _endthreadex(num);
# endif
#endif   /* MQ_HAS_THREAD */
  }
  exit(num);
}

/// \brief thread save abort
void SysAbort (
) {
  abort();
}


