/**
 *  \file       src/tcp_io.c
 *  \brief      \$Id: tcp_io.c 1183 2009-05-18 09:58:23Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1183 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "tcp_io.h"
#include "generic_io.h"
#include "io_private.h"
#include "error.h"
#include "log.h"
#include "sys.h"
#include "buffer.h"
#include "bufferL.h"

#define MQ_ERROR_S tcp->io->msgque->error
#define MSGQUE	 tcp->io->msgque

/*****************************************************************************/
/*                                                                           */
/*                                private                                    */
/*                                                                           */
/*****************************************************************************/

/** -brief everything needed for a \e Tcp msgque link
 */
struct TcpS {

  struct MqIoS const * io;	    ///< link to 'io' object
  struct GenericS * generiC;	    ///< generic part of the connection

  struct addrinfo * remoteaddr;	    ///< remote address-info
  struct addrinfo * localaddr;	    ///< local address-info

  struct MqBufferS * host;	    ///< host name
  struct MqBufferS * port;	    ///< port name

  struct MqBufferS * myhost;	    ///< CLIENT: my host name
  struct MqBufferS * myport;	    ///< CLIENT: my port name

};

static enum MqErrorE
sTcpCreateAddrInfo (
  struct TcpS const * const tcp,    ///< [in] the tcp object in duty
  struct MqBufferS * const host,    ///< [in] the host name
  struct MqBufferS * const port,    ///< [in] the port name
  struct addrinfo ** adressinfo	    ///< [out] the linked list of addrinfo
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  struct addrinfo hints;
  MQ_STR hoststr = NULL , portstr = NULL;

  // 1. get the host (NULL if not set)
  if (host && host->cursize) {
    MqErrorCheck(MqBufferGetC (error, host, &hoststr));
  }

  // 2. get the port (NULL if not set)
  if (port && port->cursize) {
    MqErrorCheck(MqBufferGetC (error, port, &portstr));
  }

  // 3. fill the hints
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_flags = AI_PASSIVE;
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_protocol = 0;
  hints.ai_canonname = NULL;
  hints.ai_addr = NULL;
  hints.ai_next = NULL;

  // 4. Call the getaddrinfo
  MqErrorCheck(SysGetAddrInfo(MQ_ERROR_S,hoststr,portstr,&hints,adressinfo));

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

static enum MqErrorE
sTcpCheckArg (
  struct TcpS * const tcp,
  struct MqBufferLS * const argv
)
{
  struct MqErrorS * const error = MQ_ERROR_S;

  MqErrorCheck (MqBufferLCheckOptionU (error, argv, "--host", &tcp->host));
  MqErrorCheck (MqBufferLCheckOptionU (error, argv, "--port", &tcp->port));
  MqErrorCheck (MqBufferLCheckOptionU (error, argv, "--myhost", &tcp->myhost));
  MqErrorCheck (MqBufferLCheckOptionU (error, argv, "--myport", &tcp->myport));

  if (!tcp->host) {
    if (MQ_IS_SERVER(tcp->io->msgque)) {
      // the SERVER listen by default to ALL interfaces
      tcp->host = MqBufferCreateC ("0.0.0.0");
    } else {
      // the CLIENT connect by default to localhost
      tcp->host = MqBufferCreateC ("localhost");
    }
  }

  if (!tcp->port)
    return MqErrorDbV2(error, MQ_ERROR_OPTION_REQUIRED,"TCP","--port");

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

void
TcpCleanup (
  struct MqErrorS * const error,
  struct MqBufferLS * const argv
)
{
  struct MqBufferS * dummy = NULL;
  MqBufferLCheckOptionU (error, argv, "--host", &dummy);
  MqBufferDelete(&dummy);
  MqBufferLCheckOptionU (error, argv, "--port", &dummy);
  MqBufferDelete(&dummy);
  MqBufferLCheckOptionU (error, argv, "--myhost", &dummy);
  MqBufferDelete(&dummy);
  MqBufferLCheckOptionU (error, argv, "--myport", &dummy);
  MqBufferDelete(&dummy);
}

/*****************************************************************************/
/*                                                                           */
/*                                public                                     */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
TcpCreate (
  struct MqIoS * const io,
  struct MqBufferLS * const argv,
  struct TcpS ** const tcpPtr
)
{
  struct MqS const * const msgque = io->msgque;
  struct MqErrorS * const error = msgque->error;

  // ok we have an TCP style communication
  register struct TcpS * const tcp = (struct TcpS * const) MqSysCalloc (error, 1, sizeof (*tcp));
  if (likely (tcpPtr != NULL))
    *tcpPtr = tcp;

  tcp->io = io;

  // get TCP options
  MqErrorCheck (sTcpCheckArg (tcp, argv));

  // init the generic part
  MqErrorCheck (GenericCreate (io, argv, &tcp->generiC));

  // fill remote address
  MqErrorCheck (sTcpCreateAddrInfo (tcp, tcp->host, tcp->port, &tcp->remoteaddr));

  // fill my address
  if (tcp->myhost || tcp->myport) {
    if (MQ_IS_CLIENT (msgque)) {
      MqErrorCheck (sTcpCreateAddrInfo (tcp, tcp->myhost, tcp->myport, &tcp->localaddr));
    } else if(tcp->myhost) {
      return MqErrorDbV (MQ_ERROR_OPTION_WRONG,"server","--myhost");
    } else if(tcp->myport) {
      return MqErrorDbV (MQ_ERROR_OPTION_WRONG,"server","--myport");
    }
  }

  // set the socket
  MqErrorCheck (GenericCreateSocket (tcp->generiC, tcp->remoteaddr->ai_family, 
		    tcp->remoteaddr->ai_socktype, tcp->remoteaddr->ai_protocol));

  // inform the user
  if (unlikely(msgque->config.debug >= 4)) {
    const char *host;
    MQ_INT port;
    MqErrorCheck(SysGetTcpInfo(error,tcp->remoteaddr,&host,&port));
    MqDLogV (msgque, __func__, 4, "remote socket host<%s> and port<%u>\n", host, port);
  }

  return MQ_OK;

error:
  return MqErrorStack (error);
}

void 
TcpDelete (
  struct TcpS ** const tcpP
)
{
  struct TcpS * const tcp = *tcpP;
  if (unlikely(!tcp))	return;		// nothing to do

  GenericDelete (&tcp->generiC);

  SysFreeAddrInfo(&tcp->remoteaddr);
  SysFreeAddrInfo(&tcp->localaddr);

  MqBufferDelete (&tcp->host);
  MqBufferDelete (&tcp->port);
  MqBufferDelete (&tcp->myhost);
  MqBufferDelete (&tcp->myport);

  MqSysFree (*tcpP);
}

/*****************************************************************************/
/*                                                                           */
/*                                public                                     */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
TcpServer (
  register struct TcpS * const tcp
)
{
  MqErrorCheck (GenericServer
                (tcp->generiC, tcp->remoteaddr->ai_addr, tcp->remoteaddr->ai_addrlen));
  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

enum MqErrorE
TcpConnect (
  register struct TcpS * const tcp
) {
  struct MqS const * const msgque = MSGQUE;
  const char *host;
  MQ_INT port;

  // create my socket adress
  if (tcp->localaddr) {
    MqErrorCheck (GenericBind
		  (tcp->generiC, tcp->localaddr->ai_addr, tcp->localaddr->ai_addrlen));
  }

  // do the connect
  MqErrorCheck (GenericConnect
                (tcp->generiC, tcp->remoteaddr->ai_addr, tcp->remoteaddr->ai_addrlen,
		    tcp->io->timeout));

  // now get sockaddr !after! successful connect
  MqErrorCheck (GenericGetSockName
                (tcp->generiC, tcp->remoteaddr->ai_addr, (socklen_t*) &tcp->remoteaddr->ai_addrlen));

  // finaly get tcp->host and tcp->port
  MqErrorCheck(SysGetTcpInfo(msgque->error,tcp->remoteaddr,&host,&port));
  MqBufferSetC (tcp->host, host);
  MqBufferSetI (tcp->port, port);
  MqDLog(msgque, 4, "local socket host<%s> and port<%u>\n", host, port);

  return MQ_OK;

error:
  return MqErrorStack (msgque->error);
}

/*****************************************************************************/
/*                                                                           */
/*                                debugging                                  */
/*                                                                           */
/*****************************************************************************/

#ifdef _DEBUG
void
TcpLogAddr (
  struct MqS const * const msgque,
  struct addrinfo * addrinfo
)
{
  struct sockaddr_in * sockaddr_in = (struct sockaddr_in *) addrinfo->ai_addr;
  const char *str;
  MQ_INT port;
  MqDLogV (msgque, __func__, 0, ">>>> sockaddr_in (%p)\n", sockaddr_in);
  str = (sockaddr_in->sin_family == AF_INET ? "AF_INET" : "WRONG!!");
  MqDLogV (msgque, __func__, 0, " sin_family = %s(%i)\n", str, sockaddr_in->sin_family);
  SysGetTcpInfo(msgque->error,addrinfo,&str,&port);
  MqDLogV (msgque, __func__, 0, " sin_port   = %u\n", port);
  MqDLogV (msgque, __func__, 0, " sin_addr   = %s\n", str);
  MqDLogV (msgque, __func__, 0, "<<<< sockaddr_in\n");
}
#endif

