/**
 *  \file       src/token.c
 *  \brief      \$Id: token.c 1176 2009-05-14 13:57:36Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1176 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "token.h"
#include "sys.h"
#include "error.h"
#include "log.h"
#include "send.h"
#include "buffer.h"
#include "bufferL.h"
#include "read.h"
#include "mq_io.h"
#include "log.h"
#include "trans.h"
#include "cache.h"

#define TOKEN_LEN (HDR_TOK_LEN+1)
#define MQ_ERROR_S token->msgque->error
#define MSGQUE token->msgque

#if defined(WORDS_BIGENDIAN)
#   define pTokenCmpD(s1,s2) (pByte2INT(s1) - pByte2INT(s2))
#else
#   define pTokenCmpD(s1,s2) (pByte2INT(s2) - pByte2INT(s1))
#endif

#if defined(WORDS_BIGENDIAN)
#   define pTokenCmpI(i1,i2) (i1 - i2)
#else
#   define pTokenCmpI(i1,i2) (i2 - i1)
#endif

/*****************************************************************************/
/*                                                                           */
/*                               token_define                                */
/*                                                                           */
/*****************************************************************************/

struct MqTokenS;

struct pTokenItemS {
  MQ_INT name;			///< the pToken name (cast into an integer)
  MqTokenF proc;		///< the pTokenHandler for calling
  MQ_PTR data;                  ///< additional data for pToken
  MqTokenDataFreeF datafreeF;	///< function to free additional data
};

struct pTokenSpaceS {
  struct MqErrorS * error;              ///< only for errors
  struct pTokenItemS *items;    ///< array of the items
  struct pTokenItemS *lastItem; ///< save the last item
  MQ_SIZE size;                 ///< max possible number of items
  MQ_SIZE used;                 ///< number of used entries in the items
  MQ_INT sorted;		///< is the array sorted ?
};

struct MqTokenS {
  struct MqS * msgque;		///< link to msgque object
  MQ_STRB current[TOKEN_LEN];	///< kind the the current action
  struct pTokenSpaceS *loc;     ///< the local handler
};

/*****************************************************************************/
/*                                                                           */
/*                          token_static                                     */
/*                                                                           */
/*****************************************************************************/

static enum MqErrorE sTokenSpaceCreate (
  struct MqErrorS * const error,
  MQ_SIZE const add,
  struct pTokenSpaceS ** const out
);

static enum MqErrorE sTokenSpaceFree (
  struct pTokenSpaceS ** const spaceP
);

static enum MqErrorE sTokenSpaceAdd (
  register struct pTokenSpaceS * const space,
  MQ_SIZE const add
);

static enum MqErrorE sTokenSpaceDelItem (
  struct pTokenSpaceS * const space,
  MQ_STRB const * const name
);

/*****************************************************************************/
/*                                                                           */
/*                               token_init                                  */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
pTokenCreate (
  struct MqS * const msgque,
  struct MqTokenS ** const out
)
{
  register struct MqTokenS * const token = (struct MqTokenS *const)
        MqSysCalloc (msgque->error, 1, sizeof (*token));

  if (out) *out = token;

  token->msgque = msgque;

  MqErrorCheck (sTokenSpaceCreate (MQ_ERROR_S, 10, &token->loc));

  return MQ_OK;

error:
  return MqErrorStack (msgque->error);
}

enum MqErrorE
pTokenDelete (
  register struct MqTokenS ** const tokenP
)
{
  struct MqTokenS * const token = *tokenP;
  if (unlikely(!token))
    return MQ_OK;

  // 1. free the item's
  MqErrorCheck (sTokenSpaceFree (&token->loc));

  // 2. free my memory
  MqSysFree (*tokenP);

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

/*****************************************************************************/
/*                                                                           */
/*                              token_memory                                 */
/*                                                                           */
/*****************************************************************************/

static enum MqErrorE
sTokenSpaceCreate (
  struct MqErrorS * const error,
  MQ_SIZE const add,
  struct pTokenSpaceS ** const out
)
{
  // create the structure
  register struct pTokenSpaceS *space = (struct pTokenSpaceS *) 
					    MqSysCalloc (error, 1, sizeof (*space));

  // filling the items
  space->items = (struct pTokenItemS *) MqSysCalloc (error, add, sizeof (*space->items));
  space->size = add;
  space->used = 0;
  space->sorted = 0;
  space->lastItem = space->items;
  space->error = error;

  // space back
  *out = space;

  return MQ_OK;
}

static enum MqErrorE
sTokenSpaceAdd (
  register struct pTokenSpaceS * const space,
  MQ_SIZE const add
)
{
  register MQ_SIZE newsize = space->used + add;

  if (newsize <= space->size)
    return MQ_OK;

  // add extra space
  newsize *= 2;

  // alloc new space
  space->items = (struct pTokenItemS *)
	MqSysRealloc (space->error, space->items, (newsize * sizeof (*space->items)));
  memset(space->items+space->used, '\0', newsize-space->used);
  space->size = newsize;
  space->lastItem = space->items;

  return MQ_OK;
}

static enum MqErrorE
sTokenSpaceFree (
  struct pTokenSpaceS ** const spaceP
)
{
  struct pTokenSpaceS *space = *spaceP;

  if (!space)
    return MQ_OK;

  sTokenSpaceDelItem (space, "_ALL");

  MqSysFree (space->items);
  MqSysFree (*spaceP);

  return MQ_OK;
}

MQ_INLINE static struct pTokenItemS *
sTokenSpaceFindItem (
  register struct pTokenItemS * start,
  register struct pTokenItemS const * const end,
  register MQ_INT const nameI
)
{
  while (start < end && start->name != nameI)
    start++;
  return (start >= end ? NULL : start);
}

static enum MqErrorE
sTokenSpaceDelItem (
  struct pTokenSpaceS * const space,
  MQ_STRB const * const name
)
{
  register struct pTokenItemS * start = space->items;
           struct pTokenItemS * end = start + space->used;

  if (strncmp (name, "_ALL", HDR_TOK_LEN)) {
    start = sTokenSpaceFindItem (start, end, pByte2INT(name));

    if (unlikely (start == NULL))
      return MqErrorV (space->error, __func__, -1, "item '%s' not found", name);

    // call the data item cleanup
    if (start->data && start->datafreeF) {
      (*start->datafreeF) (MqErrorGetMsgque(space->error), start->data);
    }
    // one item less
    space->used -= 1;
    // move all ites !after! start one to the left
    memmove(start, start+1, end-start-1);
    // set the last item to zero
    memset(start+space->used, '\0', sizeof(struct pTokenItemS));
  } else {
    while (start < end--) {
      if (end->data && end->datafreeF) {
	(*end->datafreeF) (MqErrorGetMsgque(space->error), end->data);
      }
    }
    // set all items to zero
    memset(start, '\0', space->used * sizeof(struct pTokenItemS));
    space->used = 0;
  }

  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                              token_misc                                   */
/*                                                                           */
/*****************************************************************************/

//static void _pTokenPrint(struct MqS * msgque, const char *prefix);

MQ_INLINE static int
MQ_CDECL sTokenCompare1 (
  const void *one,
  const void *two
)
{
  return pTokenCmpI (((struct pTokenItemS *) one)->name, ((struct pTokenItemS *) two)->name);
}

MQ_INLINE static int
MQ_CDECL sTokenCompare2 (
  const void *key,
  const void *two
)
{
  return pTokenCmpI (*((MQ_INT*)key), ((struct pTokenItemS *) two)->name);
}

/*
MQ_INLINE MQ_SIZE 
pTokenGetUsed (
  register struct MqTokenS const * const token
)
{
  return token->loc->used;
}
*/

// "MQ_INLINE" because the proc in only used once
MQ_INLINE enum MqErrorE
pTokenInvoke (
  register struct MqTokenS const * const token
)
{
  register struct pTokenItemS *item = NULL;
  register struct pTokenSpaceS * const space = token->loc;
  const MQ_INT icurrent = pByte2INT(token->current);

  // clear error buffer
  MqErrorReset (MQ_ERROR_S);

  if (unlikely (!space->sorted)) {
    qsort (space->items, space->used, sizeof (struct pTokenItemS), sTokenCompare1);
    space->sorted = 1;
  }
  // ATTENTION: token->lastItem was filled in pTokenCreate
  if (sTokenCompare2 (&icurrent, space->lastItem) == 0) {
    item = space->lastItem;
  }

  if (unlikely (item == NULL)) {
    item = (struct pTokenItemS *) bsearch (&icurrent, space->items,
	space->used, sizeof (struct pTokenItemS), sTokenCompare2);
  }

  if (likely (item != NULL)) {
    space->lastItem = item;
  } else {
    return MqErrorV (MQ_ERROR_S, __func__, -1, "token <%s> not found", token->current);
  }

/*
    MqDLog(MSGQUE,1,"current<%s>, currentPtr->proc<%p>, currentPtr->data<%p>\n", 
	token->current, currentPtr->proc, currentPtr->data);
*/

  MqErrorCheck (item->proc (token->msgque, item->data));

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

enum MqErrorE
pTokenAddHdl (
  struct MqTokenS const * const token,
  MQ_STRB const * const name,
  MqTokenF proc,
  MQ_PTR data,
  MqTokenDataFreeF datafreeF
)
{
  struct MqS const * const msgque = token->msgque;
  register struct pTokenSpaceS * const space = token->loc;
  register struct pTokenItemS *free;

  MQ_INT nameI = pByte2INT(name); 

  MqDLog (msgque, 4, "HANDEL %s ADD %s: proc<%p>, data<%p>\n",
          (token == msgque->srvT ? "SERVICE" : "RETURN"), name, proc, data);

  if (sTokenSpaceFindItem (space->items, space->items + space->used, nameI))
    return MqErrorV (space->error, __func__, -1, "item '%s' already available", name);

  MqErrorCheck (sTokenSpaceAdd (space, 1));

  free = space->items + space->used;

  space->sorted = 0;
  space->used += 1;

  free->name = nameI;
  free->proc = proc;
  free->data = data;
  free->datafreeF = datafreeF;

  return MQ_OK;

error:
  return MQ_ERROR;
}

enum MqErrorE
pTokenDelHdl (
  struct MqTokenS const * const token,
  MQ_STRB const * const name
)
{
  register struct MqS const * const msgque = token->msgque;
  MqDLog (msgque, 6, "HANDEL DEL %s: token<%p>\n", name, token);
  return sTokenSpaceDelItem (token->loc, name);
}

enum MqErrorE
pTokenCheckSystem (
  register struct MqTokenS const * const token
)
{
  register struct MqS * const msgque = token->msgque;
  register struct MqErrorS * const error = msgque->error;
  register MQ_STRB const * curr = token->current;

  // 1. check on USER defined token
  MqDLog (msgque, 5, "check token<%s>\n", curr);
  if (likely (*curr != '_'))
    return MQ_OK;

  // 3. work on msgque-system-token
  curr++;
  switch (*curr) {
    case 'R': {                // _RET: return from a service
      MqErrorCheck(pTransSetResult (msgque->trans, MQ_TRANS_END, msgque->read));
      break;
    }
    case 'F': {                // _FTR: SERVER get filter data
      struct MqSendS * send = msgque->send;
      if (msgque->config.fFTR) {
	enum MqErrorE ret;
	// if a filer follow, use this filter to send data
	if (msgque->myFilter != NULL) {
	  msgque->send = msgque->myFilter->send;
	}
	ret = (*msgque->config.fFTR) (msgque, msgque->config.FTRdata);
	msgque->send = send;
	if (ret == MQ_ERROR) {
	  // if the error is in the "myFilter" copy back
	  if (msgque->myFilter != NULL && MqErrorGetCode(msgque->myFilter->error) == MQ_ERROR) {
	    MqErrorCopy(msgque->error, msgque->myFilter->error);
	  }
	}
      } else {
	MqErrorDb (MQ_ERROR_TOKEN_FTR);
      }
      // default answer
      MqSendSTART (send);
      MqErrorCheck (MqSendRETURN (send));
      break;
    }
    case 'E':
    {   
	curr++;
	switch (*curr) {
	    case 'O': {		// _EOF: SERVER EndOfFile for input stream
	      struct MqSendS * send = msgque->send;
	      enum MqErrorE ret = MQ_OK;
	      if (msgque->config.fEOF) {
		// if a filer follow, use this filter to send data
		if (msgque->myFilter != NULL) {
		  msgque->send = msgque->myFilter->send;
		}
		ret = (*(MqTokenF)msgque->config.fEOF) (msgque, msgque->config.EOFdata); 
		msgque->send = send;
		if (ret == MQ_ERROR) {
		  // if the error is in the "myFilter" copy back
		  if (msgque->myFilter != NULL && MqErrorGetCode(msgque->myFilter->error) == MQ_ERROR) {
		    MqErrorCopy(msgque->error, msgque->myFilter->error);
		  }
		}
	      }
	      if (ret == MQ_OK && msgque->myFilter != NULL) {
		if (MqSendEOF(msgque->myFilter->send, MQ_TIMEOUT_USER) == MQ_ERROR) {
		  MqErrorCopy (msgque->error, msgque->myFilter->error);
		}
	      }
	      // default answer
	      MqSendSTART (send);
	      MqErrorCheck (MqSendRETURN (send));
	      break;
	    }
	    case 'R': {		// _ERR: CLIENT on error in SERVER
	      MQ_STR errtext;
	      MQ_INT errnum;
	      MqErrorCheck (MqReadI (msgque->read, &errnum));
	      MqErrorCheck (MqReadC (msgque->read, &errtext));
	      return pErrorSet (error, errnum, errtext, MQ_NO);
	    }
	}
      break;
    }
    case 'I':                  // _IAA:: SERVER, (I) (A)m (A)vailable
      // -> client need the rmtCtx pointer from the server
      // -> server need 'binary mode' and 'endian mode' from the client
      if (MQ_IS_SERVER_PARENT (msgque)) {
	struct MqSendS * const send = msgque->send;
	MQ_BOL myendian, mystring;
	struct MqBufferS * name = NULL;

        msgque->config.isString = MQ_YES;

	// read the binary mode
	MqReadO(msgque->read, &mystring);

	// read the other endian and set my msgque->endian
	MqReadO(msgque->read, &myendian);
# if defined(WORDS_BIGENDIAN)
	msgque->endian = (myendian ? MQ_NO : MQ_YES);
# else
	msgque->endian = (myendian ? MQ_YES : MQ_NO);
# endif

	// read server name
	MqReadU(msgque->read, &name);
	if (name->cursize > 0) {
	    MqSysFree(msgque->config.name);
	    msgque->config.name = mq_strdup(name->cur.C);
	}

	// send my endian back
        MqSendSTART (send);
# if defined(WORDS_BIGENDIAN)
	MqSendO (send, MQ_YES);
# else
        MqSendO (send, MQ_NO);
# endif
	MqErrorCheck (pSendSYSTEM_RETR (send));

	// set the binary mode
	msgque->config.isString = mystring;
	pReadSetType(msgque->read, mystring);

	break;
      }
      break;
    case 'O':                  // _OKS: SERVER , CHILD statup OK
      if (MQ_IS_SERVER (msgque)) {
	struct MqS * fmsgque;
	struct MqReadS * const read = msgque->read;
        struct MqSendS * const send = msgque->send;
	struct MqConfigS * configP;
	MQ_INT i;
        // if "--fCreate" is available then "--fDelete" have to be available too
	if (unlikely (msgque->config.fChildCreate == NULL)) {
	    return MqErrorDbV(MQ_ERROR_OPTION_REQUIRED, "SERVER", "--fChildCreate");
	} else if (unlikely (msgque->config.fChildDelete == NULL)) {
	    return MqErrorDbV(MQ_ERROR_OPTION_REQUIRED, "SERVER", "--fChildDelete");
	}
	// need client-code to handel ContextCreate request
	configP = MqConfigDup(&msgque->config);
	configP->parent = msgque;
	configP->myCtx = NULL;
        MqErrorCheck (MqReadI (read, &i));
	if (i != -1) configP->debug = i;
        MqErrorCheck (MqReadI (read, &i));
	if (i) configP->isSilent = MQ_YES;
        if ((*configP->fChildCreate) (configP, NULL, NULL, &fmsgque) == MQ_ERROR) {
          //return MqErrorDb (MQ_ERROR_CREATE_CHILD);
	  MqErrorCopy(msgque->error, fmsgque->error);
	  return MQ_ERROR;
        }
	MqSendSTART (send);
        MqErrorCheck (pSendSYSTEM_RETR (send));
	// no MqBufferLDelete(&argv) because the CHILD will cleanup argv;
      }
      break;
    case 'P':                   // _PEO: the (P)rocess (E)vent (O)k
      break;
    case 'S':
      curr++;
      switch (*curr) {
        case 'H':		// _SHD: SERVER shutdown request from remote Mq
	  // shutdown all dependenig objects
	  pMqShutdown(msgque);
	  MqSendSTART (msgque->send);
	  pSendSYSTEM_RETR (msgque->send);
	  MqErrorReset(msgque->error);
	  // the deletion will be handled in "pEventStart"
	  pMqGetFirstParent(msgque)->exit_msgque = msgque;
	  return MQ_EXIT;
        case 'R':              // _SRT: System return
          pTransSetResult (msgque->trans, MQ_TRANS_END, NULL);
          break;
      }
      break;
    default:                   // got a wrong message
      goto wrong_token;
  }

  return MQ_CONTINUE;

wrong_token:
  MqErrorV (error, __func__, -1,
	    "unknown Mq-System-Token \"%s\" found", msgque->srvT->current);

error:
  return MqErrorStack (error);
}

/*****************************************************************************/
/*                                                                           */
/*                            token_set/get                                  */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE MQ_STRB const *
pTokenGetCurrent (
  struct MqTokenS const * const token
)
{
  return token->current;
}

MQ_INLINE void
pTokenSetCurrent (
  struct MqTokenS * const token,
  MQ_STRB const * const str
)
{
  memcpy (token->current, str, HDR_TOK_LEN);
}

MQ_BOL
pTokenCheck (
  struct MqTokenS const * const token,
  MQ_STRB const * const str
)
{
  return (!pTokenCmpD(token->current,str));
}



