/**
 *  \file       src/trans.c
 *  \brief      \$Id: trans.c 1210 2009-05-21 06:33:49Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1210 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "trans.h"
#include "sys.h"
#include "error.h"
#include "log.h"
#include "read.h"
#include "cache.h"

#define TOKEN_LEN (HDR_TOK_LEN+1)
#define MSGQUE msgque

/*****************************************************************************/
/*                                                                           */
/*                               trans_init                                  */
/*                                                                           */
/*****************************************************************************/

struct MqTransS {
  struct MqS * msgque;		    ///< link to the msgque object
  MQ_SIZE   transIdR;		    ///< the largest currently used tranId number
  MQ_SIZE   transIdZ;		    ///< the size of the tranIdA array
  struct MqTransItemS ** transIdA;  ///< array of struct MqTransItemS * pointer's
  struct MqCacheS * transCache;	    ///< cache for MqTransItemSP
};

struct MqTransItemS {
  enum MqTransE status;		    ///< START, END, TIMEOUT or CANCEL
  struct MqReadS * result;	    ///< the result of the transaction (the answer)
  struct MqTransS const * trans;    ///< link to the \e MqTransS object
  MQ_TIME_T start;		    ///< time transaction start
  MQ_HDL last;			    ///< temporary storage of last transaction
  MQ_HDL transId;		    ///< unique transaction ID

  // the next line are used to define a callback
  MqTokenF proc;		    ///< callback function
  MQ_PTR data;			    ///< USER defined data for the callback function
  MqTokenDataFreeF datafreeF;	    ///< free USER defined data
};

/*****************************************************************************/
/*                                                                           */
/*                               C R E A T E                                 */
/*                                                                           */
/*****************************************************************************/

static struct MqTransItemS *
pTransItemCreate (
  struct MqTransS * const trans
)
{
  struct MqTransItemS * const item = (struct MqTransItemS *) MqSysMalloc (MQ_ERROR_PANIC, sizeof (*item));
  trans->transIdR++;
  item->transId = trans->transIdR;
  if (trans->transIdR >= trans->transIdZ) {
    trans->transIdA = (struct MqTransItemS **) MqSysRealloc (MQ_ERROR_PANIC, 
          trans->transIdA, (trans->transIdZ + 100) * sizeof(item));
    memset (trans->transIdA + trans->transIdZ, 0, 100 * sizeof(item));
    trans->transIdZ += 100;
  }
  return trans->transIdA[item->transId] = item;
}

static void
pTransItemDelete (
  struct MqTransItemS ** const itemP
)
{
  struct MqTransItemS * const item = *itemP;
  struct MqTransS const * const trans = item->trans;
  if (unlikely(item == NULL))
    return;
  if (item->result)
    pReadDelete (&item->result);
  trans->transIdA[item->transId] = NULL;
  MqSysFree(*itemP);
}

enum MqErrorE
pTransCreate (
  struct MqS * const msgque,
  struct MqTransS ** const out
)
{
  struct MqErrorS * error = msgque->error;
  struct MqTransS * const trans = (struct MqTransS *) MqSysCalloc (error, 1, sizeof (*trans));
  if (likely (out != NULL))
    *out = trans;
  trans->msgque = msgque;

  pCacheCreate (error, (CacheCF) pTransItemCreate, (CacheDF) pTransItemDelete, trans, &trans->transCache);

  return MqErrorGetCode(error);
}

void
pTransDelete (
  struct MqTransS ** const transP
)
{
  struct MqTransS * trans;
  struct MqS const * msgque;

  if (unlikely(transP == NULL || *transP == NULL)) return;               // nothing to do
  trans = *transP;
  msgque = trans->msgque;

  pCacheDelete (&trans->transCache);

  MqSysFree (trans->transIdA);
  MqSysFree (*transP);

  // free the
}

MQ_HDL
pTransPop (
  struct MqTransS const * const trans,
  MqTokenF const proc,
  MQ_PTR data,
  MqTokenDataFreeF datafreeF
)
{
  struct MqTransItemS * const item = (struct MqTransItemS *) pCachePop (trans->transCache);

  item->status = MQ_TRANS_START;
  item->result = NULL;
  item->trans = trans;
  item->start = time (NULL);
  item->proc = proc;
  item->data = data;
  item->datafreeF = datafreeF;
  item->last = trans->msgque->_trans;

  trans->transIdA[item->transId] = item;

  return item->transId;
}

void
pTransPush (
  struct MqTransS const * const trans,
  MQ_HDL transId
)
{
  struct MqTransItemS * const item = trans->transIdA[transId];
  if (unlikely(item == NULL)) return;
  if (item->data && item->datafreeF)
    (*item->datafreeF)(trans->msgque, item->data);
  pCachePush (trans->transCache, item);
  trans->transIdA[transId] = NULL;
}

/*****************************************************************************/
/*                                                                           */
/*                              trans_misc                                   */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE int
pTransCheckStart (
  struct MqTransS const * const trans,
  const MQ_HDL transId
)
{
  return (trans->transIdA[transId]->status == MQ_TRANS_START);
}

/*****************************************************************************/
/*                                                                           */
/*                              trans_get                                    */
/*                                                                           */
/*****************************************************************************/

MQ_HDL
pTransGetLast (
  struct MqTransS const * const trans,
  const MQ_HDL transId
)
{
  return trans->transIdA[transId]->last;
}

struct MqReadS *
pTransGetResult (
  struct MqTransS const * const trans,
  const MQ_HDL transId
)
{
  struct MqTransItemS * const item = trans->transIdA[transId];
  struct MqReadS * const ret = item->result;
  item->result = NULL;
  return ret;
}

MQ_INLINE enum MqTransE
pTransGetStatus (
  struct MqTransS const * const trans,
  const MQ_HDL transId
)
{
  return trans->transIdA[transId]->status;
}

/*****************************************************************************/
/*                                                                           */
/*                              trans_set                                    */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
pTransSetResult (
  struct MqTransS const * const trans,
  enum MqTransE const status,
  struct MqReadS * const result
)
{
  struct MqS * msgque = trans->msgque;
  MQ_HDL _trans = trans->msgque->_trans;
  struct MqTransItemS * item;
  if (unlikely(_trans < 0 || _trans > trans->transIdR)) 
      return MqErrorV (msgque->error, __func__, -1, "invalid transaction-id '%i'", _trans);
  item = trans->transIdA[_trans];
  if (item == NULL) return MQ_CONTINUE;
  item->status = status;
  if (item->proc) {
    switch (pReadGetReturnCode (msgque->read)) {
      case MQ_RETURN_OK: {
	enum MqErrorE ret;
	// from: MqSendEND_AND_CALLBACK
	ret = (*item->proc)(msgque, item->data);
	pTransPush(trans, _trans);
	MqErrorCheck(ret);
	return MQ_OK;
      }
      case MQ_RETURN_ERROR: {
	struct MqReadS * const read = msgque->read;
	MQ_STR msg;
	// (*proc) is never called
	pTransPush(trans, _trans);
	MqDLog(msgque,5,"%s\n","got ERROR return code");
	MqErrorCheck(pRead_RET_START (read));
	// write HEADER
	MqErrorV (msgque->error, "callback-error", pReadGetReturnNum (read),
		"<Num|%i>\n", pReadGetReturnNum (read));
	// write ERROR-STACK
	while (MqReadGetNumItems (read)) {
	  MqErrorCheck (MqReadC (read, &msg));
	  pErrorAppendC (msgque->error, msg);
	}
	pRead_RET_END (read);
	return MQ_ERROR;
      }
    }
  } else if (result) {
    item->result = result;
    // from: MqSendEND_AND_WAIT
    msgque->read = (struct MqReadS * const) pCachePop (msgque->readCache);
  }
  return MQ_OK;

error:
  return MqErrorStack(msgque->error);
}



