/**
 *  \file       src/uds_io.c
 *  \brief      \$Id: uds_io.c 1183 2009-05-18 09:58:23Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1183 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"

#if defined(MQ_IS_POSIX)

#include "uds_io.h"
#include "io_private.h"
#include "error.h"

/*###########################################################################*/
/*###                                                                     ###*/
/*###                              posix                                  ###*/
/*###                                                                     ###*/
/*###########################################################################*/

#include "generic_io.h"
#include "log.h"
#include "sys.h"
#include "buffer.h"
#include "bufferL.h"

#include <sys/un.h>

#define MQ_ERROR_S  uds->io->msgque->error
#define MSGQUE uds->io->msgque

/*****************************************************************************/
/*                                                                           */
/*                                private                                    */
/*                                                                           */
/*****************************************************************************/

/** -brief ...
 */
struct UdsS {
  struct MqIoS const * io;	///< link to 'io' object
  struct GenericS * generic;	///< generic part of the connection
  struct sockaddr_un sockaddr;  ///< socket address
  struct MqBufferS * file;      ///< file-name
};

void
UdsCleanup (
  struct MqErrorS * const error,
  struct MqBufferLS * const argv
)
{
  struct MqBufferS * dummy = NULL;
  MqBufferLCheckOptionU (error, argv, "--file", &dummy);
  MqBufferDelete(&dummy);
}

static enum MqErrorE
sUdsCheckArg (
  struct UdsS * const uds,
  struct MqBufferLS * const argv
)
{
  MqErrorCheck (MqBufferLCheckOptionU (MQ_ERROR_S, argv, "--file", &uds->file));
  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

static void
sUdsFillAddress (
  struct UdsS * const uds
)
{
  strcpy (uds->sockaddr.sun_path, uds->file->cur.C);
  uds->sockaddr.sun_family = AF_UNIX;
}

/*****************************************************************************/
/*                                                                           */
/*                                public                                     */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
UdsCreate (
  struct MqIoS * const io,
  struct MqBufferLS * const argv,
  struct UdsS ** const udsPtr
)
{
  struct MqS const * const msgque = io->msgque;
  struct MqErrorS * const error = msgque->error;

  // ok we have an UDS style communication
  register struct UdsS * const uds = (struct UdsS *) MqSysCalloc (error, 1, sizeof (*uds));
  if (likely (udsPtr != NULL))
    *udsPtr = uds;

  uds->io   = io;
  uds->file = MqBufferCreateC ("");

  // get the UDS options
  MqErrorCheck (sUdsCheckArg (uds, argv));

  // check UDS options
  if (!uds->file->cursize) {
    return MqErrorDbV(MQ_ERROR_OPTION_REQUIRED, "UDS", "--file");
  }

  // intit the generic part
  MqErrorCheck (GenericCreate (io, argv, &uds->generic));

  // set the socket
  MqErrorCheck (GenericCreateSocket (uds->generic, AF_UNIX, SOCK_STREAM, 0));

  sUdsFillAddress (uds);

  return MQ_OK;

error:
  return MqErrorStack (error);
}

enum MqErrorE
UdsDelete (
  struct UdsS ** const udsP
)
{
  struct UdsS * const uds = *udsP;
  if (unlikely(!uds)) return MQ_OK;   // nothing to do

  GenericDelete (&uds->generic);
  MqBufferDelete (&uds->file);
  MqSysFree (*udsP);

  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                                public                                     */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
UdsServer (
  register struct UdsS * const uds
)
{
  struct MqS const * const msgque = MSGQUE;
  struct MqErrorS * const error = MQ_ERROR_S;

  // delete the UDS file (only if it exists)
  MqErrorCheck (SysUnlink(error,uds->file->cur.C));

  MqDLog (msgque, 2, "create UNIX uds socket file<%s>\n", uds->file->cur.C);
  MqErrorCheck (GenericServer
                (uds->generic, (struct sockaddr *) &uds->sockaddr, sizeof (struct sockaddr_un)));

  // if it is the !only! server context then cleanup the uds socket file 
  if (msgque->config.startAs == MQ_START_DEFAULT) {
    MqDLog (msgque, 2, "delete UNIX uds socket file<%s>\n", uds->file->cur.C);
    MqErrorCheck (SysUnlink (error, uds->file->cur.C));
  }
  return MQ_OK;

error:
  return MqErrorStack (error);
}

enum MqErrorE
UdsConnect (
  register struct UdsS * const uds
)
{
  struct MqS const * const msgque = MSGQUE;

  // do the connect
  MqErrorCheck (GenericConnect (uds->generic, (struct sockaddr *const ) &uds->sockaddr, 
		    sizeof (struct sockaddr_un), uds->io->timeout));

/**
 * now get sockaddr !after! successful connect
 * calling GenericGetSockName on UNIX domain sockets is NOT portable
 * BSD style UNIX's does not fully support this
 *
 *   MqErrorCheck (GenericGetSockName
 *              (uds->generic, (struct sockaddr *) &uds->sockaddr, sockaddrlen));
 */

  // finaly get uds->file
  MqBufferSetC (uds->file, uds->sockaddr.sun_path);

  MqDLog (msgque, 1, "local socket file<%s>\n", (MQ_STR) uds->file->data);

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

/*****************************************************************************/
/*                                                                           */
/*                                set/get                                    */
/*                                                                           */
/*****************************************************************************/

#endif

