/**
 *  \file       javamsgque/misc_java.c
 *  \brief      \$Id: misc_java.c 1192 2009-05-19 12:53:58Z dev1usr $
 *  
 *  (C) 2009 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1192 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "context_java.h"

extern jmethodID    NS(MID_Throwable_getMessage);
extern jmethodID    NS(MID_Class_getName);
extern jmethodID    NS(MID_StringWriter_init);
extern jmethodID    NS(MID_PrintWriter_init);
extern jmethodID    NS(MID_StringWriter_toString);
extern jmethodID    NS(MID_Throwable_printStackTrace);
extern jclass	    NS(Class_RuntimeException);
extern jclass	    NS(Class_StringWriter);
extern jclass	    NS(Class_PrintWriter);

/// \brief call a \c void method like a \e service-function od a \e config-function
/// \msgque
/// \param[in] object the object to call the method for
/// \param[in] class  the class of \e object or \c NULL for calling a virtual function
/// \param[in] method the \c void method to call
/// \param[in] data one additional \e jobject argument
/// \retMqErrorE \n
///   - \c MQ_OK - no error happen
///   - \c MQ_ERROR - \e JAVA error happen, \libmsgque error object is set
///   - \c MQ_CONTINUE - \e libmsgque error happen and error object is unchanged, 
///                      let upper do additional error handling steps
///   .
enum MqErrorE 
NS(ProcCall) (
  struct MqS *	msgque,
  jobject	object,
  jclass	class,
  jmethodID	method,
  jobject	data
)
{
  struct JavaContextS * context;
  JNIEnv * env;

  SETUP_context
  env = context->env;

  // clean JAVA and libmsgque error
  (*env)->ExceptionClear(env);
  MqErrorReset(msgque->error);

  // call the function
  context->callLevel++;
  if (class == NULL) {
    if (data == NULL) {
      (*env)->CallVoidMethod(env, object, method);
    } else {
      (*env)->CallVoidMethod(env, object, method, data);
    }
  } else {
    if (data == NULL) {
      (*env)->CallNonvirtualVoidMethod(env, object, class, method);
    } else {
      (*env)->CallNonvirtualVoidMethod(env, object, class, method, data);
    }
  }
  context->callLevel--;

  // no error return OK
  if((*env)->ExceptionCheck(env) == JNI_FALSE) 
    return MQ_OK;

  // is the EROOR from java or MQ
  if (MqErrorGetCode (msgque->error) == MQ_ERROR) {
    // error is from MQ let upper code decide to handle this
    (*env)->ExceptionClear(env);
    return MQ_CONTINUE;
  }
  return NS(pErrorJavaToMq)(env, msgque);
}

enum MqErrorE
NS(pErrorJavaToMq)(
  JNIEnv *	  env,
  struct MqS	  *msgque
) {
  jthrowable exception;
  jstring message;

  // ok error is from JAVA
  exception = (*env)->ExceptionOccurred(env);
  // 1. get detail Message
  message = (jstring) (*env)->CallObjectMethod(env, exception, NS(MID_Throwable_getMessage));
  if (message == NULL) {
  // 2. get StackTrace
    /*
      final Writer result = new StringWriter();
      final PrintWriter printWriter = new PrintWriter(result);
      t.printStackTrace(printWriter);
      result.toString();
    */
    jobject stringWriter = (*env)->NewObject(env, NS(Class_StringWriter), NS(MID_StringWriter_init));
    jobject printWriter = (*env)->NewObject(env, NS(Class_PrintWriter), NS(MID_PrintWriter_init), stringWriter);
    (*env)->CallVoidMethod(env, exception, NS(MID_Throwable_printStackTrace), printWriter);
    message = (jstring) (*env)->CallObjectMethod(env, stringWriter, NS(MID_StringWriter_toString));
  }
  // 3. Add java-error-message to MQ error
  if (message != NULL) {
    const char *str = (*env)->GetStringUTFChars(env,message,NULL);
    MqErrorC(msgque->error, __func__, -1, str);
    (*env)->ReleaseStringUTFChars(env,message,str);
  } else {
    MqErrorC(msgque->error, __func__, -1, "unknown Exception");
  }
  // clean error and return
  (*env)->ExceptionClear(env);
  return MQ_ERROR;
}

void NS(pErrorFromString) (
  JNIEnv *	      env,
  const char *	      func,
  const char *	      file,
  const int	      line,
  const char *	      err
)
{
  const int len = (strlen(err)+100);
  char * buf = MqSysMalloc(MQ_ERROR_PANIC, len);
  char * fn = MqSysBasename(file, MQ_YES);
  mq_snprintf(buf,len,"\n        at %s(%s:%i)\n\n%s\n", func, fn , line, err);
  (*env)->ThrowNew(env, NS(Class_RuntimeException), buf);
  MqSysFree(fn);
  MqSysFree(buf);
}

void NS(pErrorFromMq) (
  JNIEnv *	      env,
  const char *	      func,
  const char *	      file,
  const int	      line,
  struct MqS *const   msgque
)
{
  struct JavaContextS * context;
  SETUP_context;
  if (context->callLevel != 0) {
    (*env)->ThrowNew(env, NS(Class_RuntimeException), "MQ-ERROR");
  } else {
    NS(pErrorFromString)(env, func, file, line, MqErrorGetText(msgque->error));
    MqErrorReset(msgque->error);
  }
}

MQ_STR
NS(pGetStringFromStaticJavaField)(
  JNIEnv *	env, 
  jclass	class,
  const char	*name,
  const char	*sig
)
{
  const jfieldID fid = (*env)->GetStaticFieldID(env, class, name, sig);
  const jobject val = (*env)->GetStaticObjectField(env, class, fid);
  const char *str = (*env)->GetStringUTFChars(env, val, NULL);
  MQ_STR ret = mq_strdup(str);
  (*env)->ReleaseStringUTFChars(env, val, str);
  return ret;
}

MQ_BOL
NS(pGetBoolFromStaticJavaField)(
  JNIEnv *	env, 
  jclass	class,
  const char	*name,
  const char	*sig
)
{
  const jfieldID fid = (*env)->GetStaticFieldID(env, class, name, sig);
  const jboolean val = (*env)->GetStaticBooleanField(env, class, fid);
  return (val == JNI_TRUE ? MQ_YES : MQ_NO);
}

MQ_INT
NS(pGetIntFromStaticJavaField)(
  JNIEnv *	env, 
  jclass	class,
  const char	*name,
  const char	*sig
)
{
  const jfieldID fid = (*env)->GetStaticFieldID(env, class, name, sig);
  return (*env)->GetStaticIntField(env, class, fid);
}

jmethodID
NS(pGetMethodIdFromObjectField)(
  JNIEnv *      env,
  jobject       obj,
  const char *  name
)
{
  const char *str;
  const jclass class = (*env)->GetObjectClass(env, obj);
  const jfieldID id = (*env)->GetFieldID(env, class, name, "Ljava/lang/String;");
  const jobject fieldO = (*env)->GetObjectField(env, obj, id);
  jmethodID method;
  if (fieldO == NULL) return NULL;
  str = (*env)->GetStringUTFChars(env,fieldO,NULL);
  method = (*env)->GetMethodID(env, class, str, "()V");
  (*env)->ReleaseStringUTFChars(env,fieldO,str);
  return method;
}

const char*
NS(pGetClassName)(
  JNIEnv *	env,
  jclass	class
)
{
  const char *msg, *ret;

  // call the "getName" on "class"
  jstring getNameO = (*env)->CallObjectMethod(env, class, NS(MID_Class_getName));
  if (getNameO == NULL) return mq_strdup("");
  
  // get the "string" out of "getNameO"
  msg = (*env)->GetStringUTFChars(env, getNameO, NULL);
  ret = mq_strdup(msg);
  (*env)->ReleaseStringUTFChars(env, getNameO, msg);

  return ret;
}

