/**
 *  \file       pymsgque/context_python.c
 *  \brief      \$Id: context_python.c 129 2009-06-30 12:10:31Z dev1usr $
 *  
 *  (C) 2009 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 129 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "msgque_python.h"

#define MSGQUE msgque

//////////////////////////////////////////////////////////////////////////////////
///
///				    Start
///
//////////////////////////////////////////////////////////////////////////////////

/** \brief create a new child-context in server mode
 *
 * \param[in,out] config configuration data
 * \param[in] argv server specific configuration data
 * \param[in] alfa this should allways be NULL
 * \param[out] msgqueP the new created #MqS object
 * \retMqErrorE
 **/
enum MqErrorE
NS(ChildCreate) (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS ** msgqueP
) {
  struct MqS * msgque;

  // this is the default for:
  //  1. MQ_STATUS_IS_INITIAL
  //  2. MQ_STATUS_IS_DUP & MQ_STATUS_IS_FORK
  struct PythonContextS * context = (struct PythonContextS *) config->myCtx;

  // 1. setup the SERVER+CHILD context using PARENT data
  if (config->parent != NULL && context == NULL) {
    struct PythonContextS * parentContext;
    context = MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (*context));
    config->myCtx = (struct MqContextS *) context;
    parentContext = (struct PythonContextS *) config->parent->config.myCtx;
    memcpy(context,parentContext,sizeof(*context));
    Py_INCREF(context->class);	// this is the class from the parent
    context->self = NULL;	// the 'self' object belongs to the PARENT
  }

  // 2. create the Message-Queue
  MqErrorCheck (MqCreate (config, argv, alfa, msgqueP));
  msgque = *msgqueP;

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(msgque, &argv));

  if ((msgque->config.statusIs & MQ_STATUS_IS_FORK) && msgque->config.parent == NULL) {
    // after a "fork" the internal state has to be cleaned
    PyOS_AfterFork();
  }

  // 3. create the Python Object
  if (context->self == NULL) {
    context->self = PyType_GenericNew((PyTypeObject*)context->class, NULL, NULL);
    if (context->self == NULL) return MQ_ERROR;

    if (context->config != NULL) {
      context->config = PyMethod_New(PyMethod_GET_FUNCTION(context->config), context->self);
    }
  }

  // 4. set the 'hdl'
  ((PyMqs_Obj*)context->self)->msgque = msgque;

  // 5. init the new object
  if (msgque->config.isServer == MQ_YES && context->config != NULL) {
    switch (NS(pCallVoidMethod)(msgque, context->config, context->self)) {
      case MQ_OK:           break;
      case MQ_CONTINUE:
      case MQ_ERROR:        goto error;
      case MQ_EXIT:	    return MQ_EXIT;
    }
  }

  if (unlikely(msgque->config.debug >= 2) && unlikely(msgque->config.parent != NULL))
    MqDLogV(msgque,__func__,2,"CREATE context<%p>\n",context);

  // everything is OK
  return MQ_OK;

error:
  return MqErrorStack ((*msgqueP)->error);
}


/// \brief delete a child-context
/// \param[in,out] msgqueP the context to delete
////
void
NS(ChildDelete) (
  struct MqS ** msgqueP
)
{
  struct PythonContextS *context;

  // return if the object was allread deleted
  if (!msgqueP && !(*msgqueP)) return;

  context = (struct PythonContextS *) (*msgqueP)->config.myCtx;

  if (context != NULL) {
    // the onDelete is necessary to avoid the double freeing of memory.
    // if a parent is deleted all childs are deleted as well.
    // during child delete the command-object is deleted (Tcl_DeleteCommandFromToken).
    // if the command-object is deleted the "ContextDelete" is called again -> double free
    if (context->onDelete == MQ_YES) return;
    context->onDelete = MQ_YES;

    if (context->self != NULL) {
      // delet my object from the GC
      PyObject_GC_UnTrack(context->self);
      // unset the 'hdl'
      ((PyMqs_Obj*)context->self)->msgque = 0;
      // unlock my global references
      Py_XDECREF(context->self); context->self = NULL;
    }
    if (context->class != NULL) {
      Py_XDECREF(context->class); context->class = NULL;
    }
    if (context->config != NULL) {
      Py_XDECREF(context->config); context->config = NULL;
    }
  }

  MqDelete (msgqueP);
  MqSysFree (context);
}


/// \brief create a TCL parent context
///
///
enum MqErrorE
NS(ParentCreate) (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS       ** msgqueP
)
{
  struct MqS * msgque;

  // this is the default for:
  //  1. MQ_STATUS_IS_INITIAL
  //  2. MQ_STATUS_IS_DUP & MQ_STATUS_IS_FORK
  struct PythonContextS * context = (struct PythonContextS *) config->myCtx;
  //struct MqErrorS * error = MQ_ERROR_PANIC;

  // The fork get a new context
  // not needed

  // create the server-context
  MqErrorCheck (NS(ChildCreate)(config, argv, alfa, msgqueP));
  msgque = *msgqueP;

  MqDLog (msgque, 2, "CREATE context<%p>\n", context);

  // everything is OK
  return MQ_OK;

error:
  return MqErrorStack((*msgqueP)->error);
}

/// \brief delete a parent-context in server mode
///
/// \param[in,out] msgqueP the context to delete
///
void
NS(ParentDelete) (
  struct MqS ** msgqueP
)
{
  struct PythonContextS *context;

  // return if the object was allread deleted
  if (!msgqueP && !(*msgqueP)) return;

  context = (struct PythonContextS *) (*msgqueP)->config.myCtx;

  // the onDelete is necessary to avoid the double freeing of memory.
  // if a parent is deleted all childs are deleted as well.
  if (context->onDelete == MQ_YES) return;

  // !! attention, no thread support in python !!
  // save data !befor! NS(ChildDelete) will delete it
  // PyThreadState * thread = context->thread;
  // int state_is_dup_thread = (
  //   thread != NULL &&
  //     ((*msgqueP)->config.statusIs & MQ_STATUS_IS_DUP) &&
  // 	((*msgqueP)->config.statusIs & MQ_STATUS_IS_THREAD)
  // );

  // finally use child-delete to clean the context
  NS(ChildDelete)(msgqueP);

  // !! attention, no thread support in python !!
  // cleanup the env
  //if (state_is_dup_thread ) Py_EndInterpreter(thread);
}

/// \brief delete a parent process or thread
///
/// \param[in] isThread bool value set to \c 1 (yes) if it is a thread and \c 0 (no) if not
/// \param[in] num the exit number
///
/// \e libmsgque has the \e thread created -> \e libmsgque
/// have to delete the thread -> no \e thread action in this handler
///
///
void
NS(ParentExit) (
  int isThread,
  int num
)
{
  if (isThread) {
  } else {
    Py_Exit(num);
  }
}

PyObject* NS(ProcessEvent) (
  PyObject    *self,
  PyObject    *args,
  PyObject    *keywds
)
{
  // 1. parse command-line
  int timeout=-1;
  const char *wait="NO";
  enum MqWaitOnEventE flags = MQ_WAIT_NO;
  static char *kwlist[] = {"timeout", "wait", NULL};

  SETUP_msgque;

  if (!PyArg_ParseTupleAndKeywords(args, keywds, "|is:ProcessEvent", kwlist,
	&timeout, &wait)) { 
    return NULL;
  }

  if (!strncmp(wait,"NO",2)) {
    flags = MQ_WAIT_NO;
  } else if (!strncmp(wait,"ONCE",4)) {
    flags = MQ_WAIT_ONCE;
  } else if (!strncmp(wait,"FOREVER",7)) {
    flags = MQ_WAIT_FOREVER;
  } else {
    PyErr_SetString(PyExc_TypeError, "wait=(NO|ONCE|FOREVER)");
    goto error;
  }
  ErrorMqToPythonWithCheck(MqProcessEvent(msgque, timeout, flags));
  Py_RETURN_NONE;
error:
  return NULL;
}

PyObject* NS(Delete) (
  PyObject    *self
)
{
  struct MqS * msgque = ((PyMqs_Obj*)self)->msgque;
  MqDelete(&msgque);
  Py_RETURN_NONE;
}

PyObject* NS(Init) (
  PyObject    *self,
  PyObject    *list
)
{
  Py_ssize_t len;
  struct MqBufferLS * initB = MqInitCreate();
  PyObject *item, *utf8;
  Py_ssize_t i;
  if (!PyList_Check(list)) return NULL;
  len = PyList_GET_SIZE(list);
  for (i=0; i<len; i++) {
    if ((item = PyList_GetItem(list, i)) == NULL) return NULL;
    if ((utf8 = PyUnicode_AsUTF8String(item)) == NULL) return NULL;
    MqBufferLAppendC(initB, PyBytes_AsString(utf8));
    Py_DECREF(utf8);
    Py_DECREF(list);
  }
  Py_RETURN_NONE;
}

