/**
 *  \file       pymsgque/misc_python.c
 *  \brief      \$Id: misc_python.c 1193 2009-05-19 14:08:18Z dev1usr $
 *  
 *  (C) 2009 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1193 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "msgque_python.h"

#define MSGQUE context->msgque

//////////////////////////////////////////////////////////////////////////////////
///
///				    Helper
///
//////////////////////////////////////////////////////////////////////////////////


/// \brief call a \c void method like a \e service-function od a \e config-function
/// \msgque
/// \param[in] callable the callable object to call
/// \param[in] arg the argument to call the method for
/// \retMqErrorE \n
///   - \c MQ_OK - no error happen
///   - \c MQ_ERROR - \e PYTHON error happen, \libmsgque error object is set
///   - \c MQ_CONTINUE - \e libmsgque error happen and error object is unchanged, 
///                      let upper code do additional error handling steps
///   .
enum MqErrorE 
NS(pCallVoidMethod) (
  struct MqS *	msgque,
  PyObject   *	callable,
  PyObject   *	arg
)
{
  PyObject *pType, *pValue, *pTraceback, *utf8, *str;
  char *message = NULL;
  struct PythonContextS *context;
  PyObject *result;
  SETUP_context

  // clean Python and libmsgque error
  PyErr_Clear();
  MqErrorReset(msgque->error);

  // call the function
  context->callLevel++;
  if (arg == NULL || ( PyMethod_Check(callable) && PyMethod_Self(callable) == arg) )
    result = PyObject_CallObject(callable, NULL);
  else
    result = PyObject_CallFunctionObjArgs(callable, arg, NULL);
  context->callLevel--;

  // no error return OK
  if(PyErr_Occurred() == NULL)
    return MQ_OK;

  // is the EROOR from java or MQ
  if (MqErrorGetCode (msgque->error) == MQ_ERROR) {
    // error is from MQ let upper code decide to handle this
    PyErr_Clear();
    return MQ_CONTINUE;
  }

  // ok error is from Python
  // 1. get detail Message
  PyErr_Fetch( &pType, &pValue, &pTraceback);
  // 3. get the error-message
  if (pValue != NULL) {
    str = PyObject_Str(pValue);
    if (str != NULL) {
      utf8 = PyUnicode_AsUTF8String(str);
      if (utf8 != NULL) {
	message = PyBytes_AsString(utf8);
	Py_DECREF(utf8);
      }
      Py_DECREF(str);
    }
  }
  if (message == NULL) {
    message = "Exception";
  }
  MqErrorC(msgque->error, __func__, -1, message);

  // 4. get the trace
  if (pTraceback != NULL) {
    PyObject *tb = PyImport_ImportModule("traceback");
    if (tb != NULL) {
      PyObject *traceO = PyObject_CallMethod(tb,"format_tb","O",pTraceback);
      if (traceO != NULL) {
	Py_ssize_t i, max = PyList_GET_SIZE(traceO);
	for (i=0; i<max; i++) {
	  utf8 = PyUnicode_AsUTF8String(PyList_GET_ITEM(traceO,i));
	  if (utf8 != NULL) {
	    MqErrorSAppendC(msgque->error, PyBytes_AsString(utf8));
	    Py_DECREF(utf8);
	  }
	}
	Py_DECREF(traceO);
      }
      Py_DECREF(tb);
    }
  }

  // clean error and return
  PyErr_Clear();
  return MQ_ERROR;
}

void NS(pErrorFromMq) (
  const char *	      func,
  const char *	      file,
  const int	      line,
  struct MqS * const  msgque
)
{
  struct PythonContextS *context;
  SETUP_context
  if (context->callLevel != 0) {
    PyErr_SetString(PyExc_RuntimeError, "MQ-ERROR");
  } else {
    char *err = MqErrorGetText(msgque->error);
    const int len = (strlen(err)+100);
    char * buf = MqSysMalloc(MQ_ERROR_PANIC, len);
    char * fn = MqSysBasename(file, MQ_YES);
    mq_snprintf(buf,len,"at %s(%s:%i)\n\n%s\n", func, fn, line, err);
    PyErr_SetString(PyExc_RuntimeError, buf);
    MqSysFree(fn);
    MqSysFree(buf);
    MqErrorReset(msgque->error);
  }
}

