/**
 *  \file       pymsgque/msgque_python.c
 *  \brief      \$Id: msgque_python.c 1182 2009-05-15 13:59:20Z dev1usr $
 *  
 *  (C) 2009 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1182 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "msgque_python.h"

#define MSGQUE msgque

PyTypeObject NS(MqS_Type);

static enum MqErrorE
NS(sFilter) (
  struct MqS * msgque,
  MQ_PTR dataP
)
{
  return NS(pCallVoidMethod)(msgque, (PyObject *)dataP, CONTEXT->self);
}

static void
NS(sFilterFree) (
  struct MqS const * const msgque,
  MQ_PTR data
)
{
  Py_XDECREF(data);
}

static PyObject *
NS(MqS_new)(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyMqs_Obj *self = (PyMqs_Obj *)type->tp_alloc(type, 0);
  self->msgque = NULL;
  return (PyObject *)self;
}

static int
NS(MqS_traverse)(PyMqs_Obj *self, visitproc visit, void *arg)
{
  struct PythonContextS *context;
  struct MqS * msgque = ((PyMqs_Obj*)self)->msgque;
  if (msgque == NULL) return 0;
  SETUP_context
  Py_VISIT(context->config);
  Py_VISIT(context->class);
  return 0;
}

static int
NS(MqS_clear)(PyMqs_Obj *self)
{
  struct MqS * msgque = self->msgque;
  // the attributes: "config,create,class" are freed in "NS(ChildDelete)" called by "MqDelete"
  MqDelete(&msgque);
  self->msgque = NULL;

  return 0;
}

static void
NS(MqS_dealloc)(PyMqs_Obj* self)
{
  NS(MqS_clear)(self);
  Py_TYPE(self)->tp_free((PyObject*)self);
}

static int
NS(MqS_init)(PyMqs_Obj *self, PyObject *args, PyObject *kwds)
{
  register struct PythonContextS * context = MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (*context));
  struct MqConfigS * config;
  struct MqS * msgque;

  struct MqBufferLS * const largv = MqBufferLCreate (10);
  struct MqBufferLS * const lalfa = MqBufferLCreate (10);
  struct MqBufferLS * tgt = largv;

  // 1. parse command-line
  PyObject  *isSilentO=NULL, *isStringO=NULL, *isServerO=NULL, *isFilterO=NULL, *argvO=NULL,
	    *parentO=NULL, *FTRO=NULL, *EOFO=NULL, *configO=NULL, *masterO=NULL;
  const char *name=NULL;

  static char *kwlist[] = {"argv", "silent", "string", "server", "filter",
      "debug", "buffersize", "timeout", "name", "parent", 
      "FTR", "EOF", "config", "master", "master_id", NULL};

  int master_id = 0, debug = 0, buffersize = 4096, timeout = MQ_TIMEOUT90;

  // fill the context
  context->self = (PyObject*)self;
  Py_INCREF(context->self);
  context->class = (PyObject*)context->self->ob_type;
  Py_INCREF(context->class);

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O!O!O!O!O!iiisO!OOOO!i:MqS_init", kwlist,
	&PyList_Type,		&argvO,
	&PyBool_Type,		&isSilentO,
	&PyBool_Type,		&isStringO,
	&PyBool_Type,		&isServerO,
	&PyBool_Type,		&isFilterO,
	&debug,
	&buffersize,
	&timeout,
	&name,
	&NS(MqS_Type),		&parentO,
	&FTRO,
	&EOFO,
	&configO,
	&NS(MqS_Type),		&masterO,
	&master_id
    )) {
    return -1;
  }

  // init config
  if (parentO != NULL)	{
    struct MqS * const parent = ((PyMqs_Obj*)parentO)->msgque;
    config = MqConfigDup (&parent->config);
    MqConfigInitParent (config, parent);
  } else if (masterO != NULL)	{
    struct MqS * const master = ((PyMqs_Obj*)masterO)->msgque;
    config = MqConfigDup (&master->config);
    MqConfigInitMaster (config, master, master_id);
  } else {
    config = MqConfigInit();
  }
  config->debug = debug;
  config->buffersize = buffersize;
  config->timeout = timeout;
  config->myCtx = (struct MqContextS *) context;

#define checkO(name) if (name ## O != NULL) config->name = (name ## O == Py_True ? MQ_YES : MQ_NO)
  checkO(isSilent);
  checkO(isString);
  checkO(isServer);
  checkO(isFilter);
#undef checkO

  if (name != NULL)
    MqConfigInitName (config, name);

  if (FTRO != NULL) {
    if (!PyCallable_Check(FTRO)) {
	PyErr_SetString(PyExc_TypeError, "parameter 'FTR' must be callable");
	return -1;
    }
    Py_INCREF(FTRO);
    config->fFTR	= NS(sFilter);
    config->FTRdata	= FTRO;
    config->FTRfree	= NS(sFilterFree);
  }

  if (EOFO != NULL) {
    if (!PyCallable_Check(EOFO)) {
	PyErr_SetString(PyExc_TypeError, "parameter 'EOF' must be callable");
	return -1;
    }
    Py_INCREF(EOFO);
    config->fEOF	= NS(sFilter);
    config->EOFdata	= EOFO;
    config->EOFfree	= NS(sFilterFree);
  }

  if (configO != NULL) {
    if (!PyCallable_Check(configO)) {
	PyErr_SetString(PyExc_TypeError, "parameter 'config' must be callable");
	return -1;
    }
    Py_INCREF(configO);
    context->config	= configO;
  }

  config->fChildCreate	= NS(ChildCreate);
  config->fChildDelete	= NS(ChildDelete);
  config->fParentCreate = NS(ParentCreate);
  config->fParentDelete = NS(ParentDelete);
  config->fParentExit	= NS(ParentExit);

  // threads and python well, not really supported
  // read more at: http://docs.python.org/3.0/c-api/init.html#thread-state-and-the-global-interpreter-lock
  config->ignoreThread    = MQ_YES;

  // the python sys.argv allways has the first argument set to the script name
  //   -> we don't set it

  // fill the argv/alfa
  if (argvO != NULL) {
    PyObject *iterator = PyObject_GetIter(argvO);
    PyObject *item, *utf8;
    char *str;

    while ((item = PyIter_Next(iterator))) {
      utf8 = PyUnicode_AsUTF8String(item);
      if (utf8 == NULL) {
	PyErr_SetString(PyExc_TypeError, "expect 'unicode' type for 'argv' argument");
	break;
      }
      str = PyBytes_AsString(utf8);
      if (str[0] == MQ_ALFA) {
	tgt = lalfa;
      } else {
	MqBufferLAppendC(tgt,str);
      }

      /* release reference when done */
      Py_DECREF(item);
      Py_DECREF(utf8);
    }
    Py_DECREF(iterator);
  }
  if (PyErr_Occurred()) return -1;

//MA
//MqBufferLLogS(NULL, largv, "largv");
//MqBufferLLogS(NULL, lalfa, "lalfa");

  // 7. create Context
  if (config->parent == NULL) {
    MqErrorCheck(NS(ParentCreate)(config, largv, lalfa, &msgque));
  } else {
    MqErrorCheck(NS(ChildCreate)(config, largv, lalfa, &msgque));
  }

  return 0;

error:
  if (msgque->config.isServer == MQ_NO) {
    // set error-message
    ErrorMqToPython();
    // allways delete the msgque object because this is the last chance to do it
    // !!attention!! never use "?Parentdelete" ot "?ChildeDelete" directly
    // "MqDelete" will call the functions from above
    MqDelete(&msgque);
  } else {
    MqExit(msgque);
  }
  return -1;
}

//////////////////////////////////////////////////////////////////////////////////
///
///				 Method definition
///
//////////////////////////////////////////////////////////////////////////////////

// from msgque_python.c

PyObject* NS(ProcessEvent)	    ( PyObject*, PyObject*, PyObject* );
PyObject* NS(Delete)		    ( PyObject*		    );
PyObject* NS(Init)		    ( PyObject*, PyObject* );

#define ProcessEvent_DOC	    "[[timeout(sec),wait(bool)[,forever(bool)]]]\nStart the eventloop and wait for incomming messages."
#define Delete_DOC		    "[noARG] Delete the LibMsgque object, but keep the PyMqS object alive."
#define Init_DOC		    "[list] initialize the arguments prefix"

// from send_python.c

PyObject* NS(SendSTART)		    ( PyObject*		    );
PyObject* NS(SendEND)		    ( PyObject*, PyObject*  );
PyObject* NS(SendEND_AND_WAIT)	    ( PyObject*, PyObject*  );
PyObject* NS(SendEND_AND_CALLBACK)  ( PyObject*, PyObject*  );
PyObject* NS(SendRETURN)	    ( PyObject*		    );
PyObject* NS(SendFTR)		    ( PyObject*, PyObject*  );
PyObject* NS(SendEOF)		    ( PyObject*, PyObject*  );
PyObject* NS(SendL_START)	    ( PyObject*		    );
PyObject* NS(SendL_END)	    ( PyObject*		    );
PyObject* NS(SendY)		    ( PyObject*, PyObject*  );
PyObject* NS(SendO)		    ( PyObject*, PyObject*  );
PyObject* NS(SendS)		    ( PyObject*, PyObject*  );
PyObject* NS(SendI)		    ( PyObject*, PyObject*  );
PyObject* NS(SendF)		    ( PyObject*, PyObject*  );
PyObject* NS(SendW)		    ( PyObject*, PyObject*  );
PyObject* NS(SendD)		    ( PyObject*, PyObject*  );
PyObject* NS(SendC)		    ( PyObject*, PyObject*  );
PyObject* NS(SendB)		    ( PyObject*, PyObject*  );
PyObject* NS(SendU)		    ( PyObject*, PyObject*  );

#define SendSTART_DOC		  "[noARG] Start to build a SEND package."
#define	SendEND_DOC		  "[token] Send package, no answer expected."
#define	SendEND_AND_CALLBACK_DOC  "[token, callback] Send package and expect answer using a callback."
#define	SendEND_AND_WAIT_DOC	  "[token, timeout] Send package and wait for an answer."
#define	SendRETURN_DOC		  "[noARG] Answer a service call."
#define	SendFTR_DOC		  "[[timeout]] Send a line in Filter-Mode."
#define	SendEOF_DOC		  "[[timeout]] Send a EndOfFile in Filter-Mode."
#define SendY_DOC		  "[value] Send a int as (1 byte signed char)."
#define SendO_DOC		  "[value] Send a boolean"
#define SendS_DOC		  "[value] Send a int as (2 byte short integer)."
#define SendI_DOC		  "[value] Send a int as (4 byte integer)."
#define SendF_DOC		  "[value] Send a float as (4 byte float)."
#define SendW_DOC		  "[value] Send a int as (8 byte wide integer)."
#define SendD_DOC		  "[value] Send a double as (8 byte double)."
#define SendC_DOC		  "[value] Send a utf8 string as (char array with \0 at the end)."
#define SendB_DOC		  "[value] Send a byte array as (unsigned char array, \0 is allowed)."
#define SendU_DOC		  "[value] Send a PyMqS buffer object."
#define	SendL_START_DOC		  "[noARG] Start to Send an embedded List-Item."
#define	SendL_END_DOC		  "[noARG] End to Send an embedded List-Item."

// from service_python.c

PyObject* NS(ServiceCreate)	    ( PyObject*, PyObject*  );
PyObject* NS(ServiceDelete)	    ( PyObject*, PyObject*  );

#define ServiceCreate_DOC	  "[token, callable] Create a new service"
#define ServiceDelete_DOC	  "[token] Delete a service"

// from slave_python.c

PyObject* NS(SlaveCreate)	    ( PyObject*, PyObject*  );
PyObject* NS(SlaveDelete)	    ( PyObject*, PyObject*  );
PyObject* NS(SlaveGet)		    ( PyObject*, PyObject*  );

#define SlaveCreate_DOC	    "[id, slave-context] Create a new master/slave link"
#define SlaveDelete_DOC	    "[id] Delete slave of id"
#define SlaveGet_DOC	    "[id] Return slave-context of id"

// from service_python.c

PyObject* NS(ReadY)		  ( PyObject* );
PyObject* NS(ReadO)		  ( PyObject* );
PyObject* NS(ReadS)		  ( PyObject* );
PyObject* NS(ReadI)		  ( PyObject* );
PyObject* NS(ReadF)		  ( PyObject* );
PyObject* NS(ReadW)		  ( PyObject* );
PyObject* NS(ReadD)		  ( PyObject* );
PyObject* NS(ReadC)		  ( PyObject* );
PyObject* NS(ReadB)		  ( PyObject* );
PyObject* NS(ReadU)		  ( PyObject* );
PyObject* NS(ReadL_START)	  ( PyObject* );
PyObject* NS(ReadL_END)		  ( PyObject* );
PyObject* NS(ReadGetNumItems)	  ( PyObject* );
PyObject* NS(ReadItemExists)	  ( PyObject* );
PyObject* NS(ReadUndo)		  ( PyObject* );

#define ReadY_DOC		  "[noARG] Read a int (1 byte signed char) from the current PyMqS package."
#define ReadO_DOC		  "[noARG] Read a boolean (1 byte boolean) from the current PyMqS package."
#define ReadS_DOC		  "[noARG] Read a int (2 byte signed short) from the current PyMqS package."
#define ReadI_DOC		  "[noARG] Read a int (4 byte singed integer) from the current PyMqS package."
#define ReadF_DOC		  "[noARG] Read a float (4 byte float) from the current PyMqS package."
#define ReadW_DOC		  "[noARG] Read a int (8 byte singed long long integer) from the current PyMqS package."
#define ReadD_DOC		  "[noARG] Read a float (8 byte doubel) from the current PyMqS package."
#define ReadC_DOC		  "[noARG] Read a a UTF8 string from the current PyMqS package."
#define ReadB_DOC		  "[noARG] Read a byte array from the current PyMqS package."
#define ReadU_DOC		  "[noARG] Read a PyMqS buffer from the current PyMqS package."
#define ReadL_START_DOC		  "[noARG] Start to read an embedded list item from the current PyMqS package."
#define ReadL_END_DOC		  "[noARG] End to read an embedded list item from the current PyMqS package."
#define ReadGetNumItems_DOC	  "[noARG] Get the number of items left in the PyMqS data package."
#define ReadItemExists_DOC	  "[noARG] Is an additional item available in the PyMqS data package."
#define ReadUndo_DOC		  "[noARG] Undo the last Read? operation."

// from config_python.c

PyObject* NS(ConfigGetDebug)	( PyObject* );
PyObject* NS(ConfigGetIsString)	( PyObject* );
PyObject* NS(ConfigGetIsSilent)	( PyObject* );
PyObject* NS(ConfigGetIsServer)	( PyObject* );
PyObject* NS(ConfigGetIsParent)	( PyObject* );
PyObject* NS(ConfigGetIsSlave)	( PyObject* );
PyObject* NS(ConfigGetParent)	( PyObject* );
PyObject* NS(ConfigGetName)	( PyObject* );
PyObject* NS(ConfigGetCtxId)	( PyObject* );
PyObject* NS(ConfigGetToken)	( PyObject* );
PyObject* NS(ConfigGetMaster)	( PyObject* );

#define ConfigGetIsString_DOC	"[noARG] boolean, 'True' if the object is using the 'string' configuration."
#define ConfigGetIsSilent_DOC	"[noARG] boolean, 'True' if the object is using the 'silent' configuration."
#define ConfigGetIsServer_DOC	"[noARG] boolean, 'True' if the object belongs to a 'server'."
#define ConfigGetIsParent_DOC	"[noARG] boolean, 'True' if the object belongs to a 'parent'."
#define ConfigGetIsSlave_DOC	"[noARG] boolean, 'True' if the object belongs to a 'slave'."
#define ConfigGetDebug_DOC	"[noARG] Return the debug level (int from 0 up to 9)."
#define ConfigGetParent_DOC	"[noARG] return the PyMqS 'parent' object of the 'current' object."
#define ConfigGetName_DOC	"[noARG] return the PyMqS 'name' (str) of the object."
#define ConfigGetCtxId_DOC	"[noARG] return the PyMqS 'identifer' (int) of the object."
#define ConfigGetToken_DOC	"[noARG] return the PyMqS 'token' (str) of the current transaction."
#define ConfigGetMaster_DOC     "[noARG] return the PyMqS 'master' object of the 'current' object."

// from pybuffer.c

PyObject* NS(GetType)		( PyObject*, PyObject* );
PyObject* NS(GetY)		( PyObject*, PyObject* );
PyObject* NS(GetO)		( PyObject*, PyObject* );
PyObject* NS(GetS)		( PyObject*, PyObject* );
PyObject* NS(GetI)		( PyObject*, PyObject* );
PyObject* NS(GetF)		( PyObject*, PyObject* );
PyObject* NS(GetW)		( PyObject*, PyObject* );
PyObject* NS(GetD)		( PyObject*, PyObject* );
PyObject* NS(GetC)		( PyObject*, PyObject* );
PyObject* NS(GetB)		( PyObject*, PyObject* );
PyObject* NS(GetL_START)	( PyObject*, PyObject* );
PyObject* NS(GetL_END)		( PyObject*	       );

#define GetType_DOC		"[Object] Return the type (STR|BIN|BYT|BOL|SRT|INT|FLT|WID|DBL|LST)\n from the PyMqS buffer object"
#define GetY_DOC		"[Object] Get and return a MQ_BYT as int (1 byte signed char) from the PyMqS buffer object."
#define GetO_DOC		"[Object] Get a bool (1 byte boolean) from the PyMqS buffer object."
#define GetS_DOC		"[Object] Get a int (2 byte signed short) from the PyMqS buffer object."
#define GetI_DOC		"[Object] Get a int (4 byte singed integer) from the PyMqS buffer object."
#define GetF_DOC		"[Object] Get a double (4 byte float) from the PyMqS buffer object."
#define GetW_DOC		"[Object] Get a int (8 byte singed long long integer) from the PyMqS buffer object."
#define GetD_DOC		"[Object] Get a double (8 byte doubel) from the PyMqS buffer object."
#define GetC_DOC		"[Object] Get a the UTF8 string from the PyMqS buffer object."
#define GetB_DOC		"[Object] Get a byte array from the PyMqS buffer object."
#define GetL_START_DOC		"[Object] Start an embedded list from the PyMqS buffer object."
#define GetL_END_DOC		"[Object] End an embedded list from the PyMqS buffer object."

// from test_python.c

PyObject* NS(ErrorC)		( PyObject*, PyObject* );

#define ErrorC_DOC		"[function-name, error-number, error-message] (testing only) create a PyMqS error"

// Fill the Struct-Array

#define ARG(N,M) { #N , (PyCFunction) NS(N), M, N ## _DOC}
static PyMethodDef NS(Methods)[] = {

    ARG(ProcessEvent,	      METH_VARARGS | METH_KEYWORDS),
    ARG(Delete,		      METH_NOARGS),
    ARG(Init,		      METH_STATIC | METH_O),

    ARG(SendSTART,	      METH_NOARGS),
    ARG(SendEND,	      METH_VARARGS),
    ARG(SendEND_AND_WAIT,     METH_VARARGS),
    ARG(SendEND_AND_CALLBACK, METH_VARARGS),
    ARG(SendRETURN,	      METH_NOARGS),
    ARG(SendFTR,	      METH_VARARGS),
    ARG(SendEOF,	      METH_VARARGS),
    ARG(SendL_START,	      METH_NOARGS),
    ARG(SendL_END,	      METH_NOARGS),
    ARG(SendY,		      METH_VARARGS),
    ARG(SendO,		      METH_VARARGS),
    ARG(SendS,		      METH_VARARGS),
    ARG(SendI,		      METH_VARARGS),
    ARG(SendF,		      METH_VARARGS),
    ARG(SendW,		      METH_VARARGS),
    ARG(SendD,		      METH_VARARGS),
    ARG(SendC,		      METH_VARARGS),
    ARG(SendB,		      METH_VARARGS),
    ARG(SendU,		      METH_O),

    ARG(ReadL_START,	      METH_NOARGS),
    ARG(ReadL_END,	      METH_NOARGS),
    ARG(ReadY,		      METH_NOARGS),
    ARG(ReadO,		      METH_NOARGS),
    ARG(ReadS,		      METH_NOARGS),
    ARG(ReadI,		      METH_NOARGS),
    ARG(ReadF,		      METH_NOARGS),
    ARG(ReadW,		      METH_NOARGS),
    ARG(ReadD,		      METH_NOARGS),
    ARG(ReadC,		      METH_NOARGS),
    ARG(ReadB,		      METH_NOARGS),
    ARG(ReadU,		      METH_NOARGS),
    ARG(ReadGetNumItems,      METH_NOARGS),
    ARG(ReadItemExists,	      METH_NOARGS),
    ARG(ReadUndo,	      METH_NOARGS),

    ARG(ServiceCreate,	      METH_VARARGS),
    ARG(ServiceDelete,	      METH_VARARGS),

    ARG(SlaveCreate,	      METH_VARARGS),
    ARG(SlaveDelete,	      METH_VARARGS),
    ARG(SlaveGet,	      METH_VARARGS),

    ARG(ConfigGetIsString,    METH_NOARGS),
    ARG(ConfigGetIsSilent,    METH_NOARGS),
    ARG(ConfigGetIsServer,    METH_NOARGS),
    ARG(ConfigGetIsParent,    METH_NOARGS),
    ARG(ConfigGetIsSlave,     METH_NOARGS),
    ARG(ConfigGetDebug,	      METH_NOARGS),
    ARG(ConfigGetParent,      METH_NOARGS),
    ARG(ConfigGetName,	      METH_NOARGS),
    ARG(ConfigGetCtxId,	      METH_NOARGS),
    ARG(ConfigGetToken,	      METH_NOARGS),
    ARG(ConfigGetMaster,      METH_NOARGS),

    ARG(GetType,	      METH_O),
    ARG(GetY,		      METH_O),
    ARG(GetO,		      METH_O),
    ARG(GetS,		      METH_O),
    ARG(GetI,		      METH_O),
    ARG(GetF,		      METH_O),
    ARG(GetW,		      METH_O),
    ARG(GetD,		      METH_O),
    ARG(GetC,		      METH_O),
    ARG(GetB,		      METH_O),
    ARG(GetL_START,	      METH_O),
    ARG(GetL_END,	      METH_NOARGS),

    ARG(ErrorC,		      METH_VARARGS),
    
    {NULL, NULL, 0, NULL}        /* Sentinel */
};
#undef ARG

static PyModuleDef NS(Module) = {
  PyModuleDef_HEAD_INIT,
  "pymsgque",
  "python language binding for: http://libmsgque.sourceforge.net/",
  -1,
  NULL, NULL, NULL, NULL, NULL
};

#define PyMqS_Type_DOC	"The msgque project is an infrastructure to link software together to\n\
act like a single software. To link mean distributing work from one\n\
software to an other software an wait or not wait for an answer. The\n\
linking is done using unix or inet domain sockets and is based on\n\
packages send from one software to an other software and back. The\n\
tclmsgque project is used to handle all the different aspects for setup\n\
and maintain the link and is responsible for:\n\
\n\
o starting and stopping the server application\n\
o starting and stopping the communication interface\n\
o sending and receiving package data\n\
o reading and writing data from or into a package\n\
o setup and maintain the event-handling for an asynchronous transfer\n\
o propagate the error messages from the server to the client"


PyTypeObject NS(MqS_Type) = {
  {PyObject_HEAD_INIT(NULL)},
  "pymsgque.PyMqS",		  /* tp_name */
  sizeof(PyMqs_Obj),		  /* tp_basicsize */
  0,				  /* tp_itemsize */
  (destructor)NS(MqS_dealloc),	  /* tp_dealloc */
  0,				  /* tp_print */
  0,				  /* tp_getattr */
  0,				  /* tp_setattr */
  0,				  /* tp_reserved */
  0,				  /* tp_repr */
  0,				  /* tp_as_number */
  0,				  /* tp_as_sequence */
  0,				  /* tp_as_mapping */
  0,				  /* tp_hash  */
  0,				  /* tp_call */
  0,				  /* tp_str */
  0,				  /* tp_getattro */
  0,				  /* tp_setattro */
  0,				  /* tp_as_buffer */
  Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC,
				  /* tp_flags */
  PyMqS_Type_DOC,		  /* tp_doc */
  (traverseproc)NS(MqS_traverse), /* tp_traverse */
  (inquiry)NS(MqS_clear),	  /* tp_clear */
  0,				  /* tp_richcompare */
  0,				  /* tp_weaklistoffset */
  0,				  /* tp_iter */
  0,				  /* tp_iternext */
  NS(Methods),			  /* tp_methods */
  0,				  /* tp_members */
  0,				  /* tp_getset */
  0,				  /* tp_base */
  0,				  /* tp_dict */
  0,				  /* tp_descr_get */
  0,				  /* tp_descr_set */
  0,				  /* tp_dictoffset */
  (initproc)NS(MqS_init),	  /* tp_init */
  0,				  /* tp_alloc */
  NS(MqS_new),			  /* tp_new */
};


PyMODINIT_FUNC
PyInit_pymsgque(void) 
{
  PyObject* m;

  NS(MqS_Type).ob_base.ob_base.ob_type = &PyType_Type;
  if (PyType_Ready(&NS(MqS_Type)) < 0)
    return NULL;

  m = PyModule_Create(&NS(Module));
  if (m == NULL)
    return NULL;

  Py_INCREF(&NS(MqS_Type));
  PyModule_AddObject(m, "MqS", (PyObject *)&NS(MqS_Type));

  // init libmsgque global data
  if (MqInitBuf == NULL && Py_GetProgramName() != NULL) {
    struct MqBufferLS * initB = MqInitCreate();
    char * buf1 = MqSysMalloc(MQ_ERROR_IGNORE, 250);
    PyObject *sys, *list, *item, *utf8;
    if (buf1 == NULL) return PyErr_NoMemory();
    MqMainSelector = NULL;
    if (wcstombs(buf1, Py_GetProgramName(), 250) != -1)
      MqBufferLAppendC(initB, buf1);
    if ((sys = PyImport_ImportModule("sys")) == NULL) return NULL;
    if ((list = PyObject_GetAttrString(sys, "argv")) == NULL) return NULL;
    if ((item = PyList_GetItem(list, 0)) == NULL) return NULL;
    if ((utf8 = PyUnicode_AsUTF8String(item)) == NULL) return NULL;
    MqBufferLAppendC(initB, PyBytes_AsString(utf8));
    Py_DECREF(utf8);
    Py_DECREF(list);
    Py_DECREF(sys);
  }

  return m;
}

