/**
 *  \file       src/bufferL.c
 *  \brief      \$Id: bufferL.c 1186 2009-05-19 06:41:03Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1186 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "bufferL.h"
#include "buffer.h"
#include "error.h"
#include "sys.h"
#include "log.h"

#define MQ_ERROR_S error
#define MSGQUE NULL

/*****************************************************************************/
/*                                                                           */
/*                              buffer_init                                  */
/*                                                                           */
/*****************************************************************************/

struct MqBufferLS *
MqBufferLCreate (
  MQ_SIZE const size
)
{
  register struct MqBufferLS * const bufL = (struct MqBufferLS * const) 
      MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (*bufL));

  bufL->data = (struct MqBufferS **) MqSysCalloc (MQ_ERROR_PANIC, size, sizeof (*bufL->data));
  bufL->size = size;
  bufL->cursize = 0;
  bufL->cur = bufL->data;

  return bufL;
}

void
MqBufferLCreateArgv (
  int const argc,
  char ** const argv,
  struct MqBufferLS ** const args,
  struct MqBufferLS ** const alfa
)
{
  char ** start = argv;
  char ** const end   = argv+argc;
  char const *  str;

	*args = MqBufferLCreate (argc);
  *alfa = MqBufferLCreate (argc);

  for (; start < end; start++) {
    str = *start;
    if (*str == MQ_ALFA && *(str+1) == '\0') break;
    MqBufferLAppendC (*args, str);
  }
  for (start++; start < end; start++) {
    MqBufferLAppendC (*alfa, *start);
  }
}

struct MqBufferLS * 
MqBufferLDup (
  struct MqBufferLS const * const in
)
{
  struct MqBufferLS * ret;
  struct MqBufferS ** retP, ** inP;
  int n;

	if (unlikely(in == NULL)) return NULL;

	ret = MqBufferLCreate(in->cursize);
  retP = ret->data;
  inP  = in->data;
  
  for (n=0; n<in->cursize; n++,inP++,retP++) {
    *retP = MqBufferDup(*inP);
  }  
  ret->cursize = in->cursize;
  return ret;
}

enum MqErrorE
pBufferLSplitAlfa (
  struct MqBufferLS ** alfa1P,		///< [in,out] source of split (all arguments after MQ_ALFA will go to alfa2P
  struct MqBufferLS ** alfa2P 		///< [out] the new alfa1 object
)
{
  register struct MqBufferS ** argP, ** endP;
  register struct MqBufferLS * in;
  MQ_SIZE idx;
  struct MqBufferLS * alfa2;
  struct MqBufferS ** inP;

	// check if "work" is available
  if (!alfa1P || !(*alfa1P)) {
    *alfa2P = NULL;
    return MQ_OK;
  }

  if ((*alfa1P)->cursize == 0) {
    *alfa2P = NULL;
    MqBufferLDelete (alfa1P);
    return MQ_OK;
  }

  in = *alfa1P;

  // 1. find the MQ_ALFA
  idx = -1;
  {
    argP = in->data;
    endP = in->data + in->cursize;
    while (argP < endP) {
      if ((*argP)->type != MQ_STRT) continue;
      if (*((*argP)->cur.C) == MQ_ALFA) {
        struct MqBufferS ** tmpP = argP;
	idx = in->cursize - (endP - argP);
	MqBufferDelete(tmpP);
	break;
      }
      argP++;
    }
  }

  if (idx == -1) return MQ_OK;		// nothing to do *alfa1P == in

  // 2. create alfa1/2
  alfa2 = *alfa2P = MqBufferLCreate (in->cursize-idx-1);
  inP = in->data+idx+1;

  // fill alfa2
  argP = alfa2->data;
  endP = alfa2->data + alfa2->size;
  for (;argP < endP; argP++, inP++) {
    *argP = *inP;
  }  
  alfa2->cursize = alfa2->size;

  // clean not used array data from "in" (e.g. *alfa1P)
  in->cur = in->data+idx;
  memset(in->cur, '\0', in->cursize-idx);
  in->cursize = idx;

  return MQ_OK;
}

void
pBufferLDeleteStatic (
  struct MqBufferLS * const bufL
)
{
  register struct MqBufferS * const *arg, * const *end;

  if (unlikely (bufL == NULL))
    return;

  arg = bufL->data;
  end = bufL->data + bufL->cursize;
  
  for (; arg < end; arg++) {
    MqBufferDelete ((struct MqBufferS * *) arg);
  }
  MqSysFree (bufL->data);
}

void
MqBufferLDelete (
  struct MqBufferLS ** bufP
)
{
  if (likely(bufP != NULL && *bufP != NULL)) {
    pBufferLDeleteStatic (*bufP);
    MqSysFree (*bufP);
  }
}

/*****************************************************************************/
/*                                                                           */
/*                              buffer_atom                                  */
/*                                                                           */
/*****************************************************************************/

/// \brief extend \e bufL to \e newSize size
/// \bufL
/// \param newSize expand \e bufL able to hold \e newsize data
static void
sBufferLNewSize (
  register struct MqBufferLS * const bufL,
  MQ_SIZE const newSize
)
{
  bufL->size = newSize + (newSize / 3);  // + 1/3 reserve
  bufL->data = (struct MqBufferS **) MqSysRealloc (MQ_ERROR_PANIC, (MQ_PTR) bufL->data, bufL->size * sizeof (*bufL->data));
  bufL->cur = bufL->data + bufL->cursize;
}

/// \brief wrapper for \e sBufferLNewSize
#define sBufferLAddSize(bufL, addSize) \
    if (unlikely((bufL->cursize + addSize) > bufL->size)) \
        sBufferLNewSize(bufL, (bufL->cursize + addSize))

/*****************************************************************************/
/*                                                                           */
/*                              buffer_append                                */
/*                                                                           */
/*****************************************************************************/

/// \attention after append the \e in belongs to \e bufL
void
MqBufferLAppend (
  register struct MqBufferLS * const bufL,
  struct MqBufferS * const in,
  MQ_SIZE const position
)
{
  sBufferLAddSize (bufL, 1);
  if (position == -1 || position >= bufL->cursize) {
    // append to the end
    *bufL->cur = in;
  } else {
    // append before the end
    struct MqBufferS ** end = bufL->data + bufL->cursize;
    struct MqBufferS ** const arg = bufL->data + position;
    for ( ; arg < end ; end--) {
	*end = *(end-1);
    }
    *arg = in;
  }
  bufL->cursize++;
  bufL->cur++;
}

void
MqBufferLAppendL (
  register struct MqBufferLS * const bufL,
  struct MqBufferLS * const in,
  MQ_SIZE position
)
{
  int shift = in->cursize;
  struct MqBufferS ** end;
  struct MqBufferS ** arg, **argin;
  sBufferLAddSize (bufL, shift);
  if (position == -1 || position >= bufL->cursize) {
    // append to the end
    position = bufL->cursize;
  } else {
    // shift everythin after position to the right
    end = bufL->data + bufL->cursize + shift;
    arg = bufL->data + position;
    for ( ; arg < end ; end--) {
	*end = *(end-shift);
    }
  }
  // copy in data starting at position
  arg = bufL->data + position;
  end = arg + shift;
  argin = in->data;
  for ( ; arg < end ; arg++, argin++) {
      *(arg) = MqBufferDup(*argin);
  }
  bufL->cursize+=shift;
  bufL->cur = bufL->data + bufL->cursize;
}

MQ_INLINE void
MqBufferLAppendY (
  struct MqBufferLS * const bufL,
  MQ_BYT const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateY (val), -1);
}

MQ_INLINE void
MqBufferLAppendO (
  struct MqBufferLS * const bufL,
  MQ_BOL const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateO (val), -1);
}

MQ_INLINE void
MqBufferLAppendS (
  struct MqBufferLS * const bufL,
  MQ_SRT const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateS (val), -1);
}

MQ_INLINE void
MqBufferLAppendI (
  struct MqBufferLS * const bufL,
  MQ_INT const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateI (val), -1);
}

MQ_INLINE void
MqBufferLAppendF (
  struct MqBufferLS * const bufL,
  MQ_FLT const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateF (val), -1);
}

MQ_INLINE void
MqBufferLAppendW (
  struct MqBufferLS * const bufL,
  MQ_WID const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateW (val), -1);
}

MQ_INLINE void
MqBufferLAppendD (
  struct MqBufferLS * const bufL,
  MQ_DBL const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateD (val), -1);
}

MQ_INLINE void
MqBufferLAppendC (
  struct MqBufferLS * const bufL,
  MQ_STRB const * const val
)
{
  MqBufferLAppend (bufL, MqBufferCreateC (val), -1);
}

MQ_INLINE void
MqBufferLAppendU (
  struct MqBufferLS * const bufL,
  MQ_BUF const val
)
{
  MqBufferLAppend (bufL, val, -1);
}

/*****************************************************************************/
/*                                                                           */
/*                            bufferL_option                                 */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
MqBufferLCheckOptionO (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_BOL * const var
)
{
  MQ_SIZE index;
  MQ_SIZE len;
  *var = MQ_NO;
  if (unlikely(bufL == NULL || bufL->cursize == 0)) return MQ_OK;
  index = 0;
  len = strlen(opt);
  while ((index = MqBufferLSearchC (bufL, opt, len, index)) != -1) {
    *var = MQ_YES;
    MqBufferLDeleteItem (error, bufL, index, 1);
  }
  return MQ_OK;
}

/// \brief search for an \e option with an #MqBufferAtomU argument in an #MqBufferLS object
/// \error
/// \bufL0
/// \optionL
/// \atom_out
/// \atom_type
/// \param[in] func value extraction function pointer
/// \retMqErrorE
/// \attL
static enum MqErrorE
sBufferLCheckOptionA (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  union MqBufferAtomU * const ato,
  enum MqTypeE const type,
  const pBufferGetAF func
)
{
  MQ_SIZE index;
  MQ_SIZE len;
  if (unlikely(bufL == NULL || bufL->cursize == 0)) return MQ_OK;
  index = 0;
  len = strlen(opt);
  while ((index = MqBufferLSearchC (bufL, opt, len, index)) != -1) {
    MqBufferLDeleteItem (error, bufL, index, 1);
    if (likely (bufL->cursize) && likely (index < bufL->cursize)) {
      MqErrorCheck ((*func) (error, bufL->data[index], ato, type));
      MqBufferLDeleteItem (error, bufL, index, 1);
    } else {
      return MqErrorDbV (MQ_ERROR_OPTION_ARG, opt);
    }
  }

  return MQ_OK;

error:
  return MqErrorStack (error);
}

MQ_INLINE enum MqErrorE 
MqBufferLCheckOptionY (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_BYT * const valP
)
{
  return sBufferLCheckOptionA (error, bufL, opt, (union MqBufferAtomU * const)valP , MQ_BYTT, pBufferGetA1);
}

MQ_INLINE enum MqErrorE 
MqBufferLCheckOptionS (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_SRT * const valP
)
{
  return sBufferLCheckOptionA (error, bufL, opt, (union MqBufferAtomU * const)valP , MQ_SRTT, pBufferGetA2);
}

MQ_INLINE enum MqErrorE 
MqBufferLCheckOptionI (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_INT * const valP
)
{
  return sBufferLCheckOptionA (error, bufL, opt, (union MqBufferAtomU * const)valP , MQ_INTT, pBufferGetA4);
}

MQ_INLINE enum MqErrorE 
MqBufferLCheckOptionF (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_FLT * const valP
)
{
  return sBufferLCheckOptionA (error, bufL, opt, (union MqBufferAtomU * const)valP , MQ_FLTT, pBufferGetA4);
}

MQ_INLINE enum MqErrorE 
MqBufferLCheckOptionW (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_WID * const valP
)
{
  return sBufferLCheckOptionA (error, bufL, opt, (union MqBufferAtomU * const)valP , MQ_WIDT, pBufferGetA8);
}

MQ_INLINE enum MqErrorE 
MqBufferLCheckOptionD (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_DBL * const valP
)
{
  return sBufferLCheckOptionA (error, bufL, opt, (union MqBufferAtomU * const)valP , MQ_DBLT, pBufferGetA8);
}

enum MqErrorE
MqBufferLCheckOptionC (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_STR * const var
)
{
  MQ_SIZE index;
  MQ_SIZE len;

  if (!bufL) return MQ_OK;

	index = 0;
  len = strlen(opt);

	while ((index = MqBufferLSearchC (bufL, opt, len, index)) != -1) {
    MqBufferLDeleteItem (error, bufL, index, 1);
    if (likely (bufL->cursize) && likely (index < bufL->cursize)) {
      MQ_STR tmp;
      MqErrorCheck (MqBufferGetC (error, bufL->data[index], &tmp));
      *var = mq_strdup(tmp);
      MqBufferLDeleteItem (error, bufL, index, 1);
    } else {
      return MqErrorDbV (MQ_ERROR_OPTION_ARG, opt);
    }
  }

  return MQ_OK;

error:
  return MqErrorStack (error);
}

enum MqErrorE
MqBufferLCheckOptionU (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_STRB const * const opt,
  MQ_BUF * const var
)
{
  struct MqBufferS * lbuf;
  MQ_SIZE index;
  MQ_SIZE len;

	if (!bufL) return MQ_OK;

  index = 0;
  len = strlen(opt);

  while ((index = MqBufferLSearchC (bufL, opt, len, index)) != -1) {
    MqBufferLDeleteItem (error, bufL, index, 1);
    if (likely (bufL->cursize) && likely (index < bufL->cursize)) {
      lbuf = bufL->data[index];
      if (*var) MqBufferDelete(var);
      *var = bufL->data[index];
      MqBufferLDeleteItem (error, bufL, index, 0);
    } else {
      return MqErrorDbV (MQ_ERROR_OPTION_ARG, opt);
    }
  }

  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                             bufferL_misc                                  */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
MqBufferLDeleteItem (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  const MQ_SIZE index,
  const MQ_BOL doDelete
)
{
  register MQ_SIZE const cursize = bufL->cursize;
  register struct MqBufferS * *arg = bufL->data + index;

  if (index < 0 || index >= cursize) {
    return MqErrorV (error, __func__, -1,
	 "wrong index '" MQ_FORMAT_Z "', allowed 0 <= index <= " MQ_FORMAT_Z, 
	    index, cursize - 1);
  }

  if (doDelete) {
    MqBufferDelete ((struct MqBufferS * *) arg);
  }

  memmove (arg, (arg+1), sizeof(struct MqBufferS *)*(cursize-index-1));
  bufL->cursize--;
  bufL->cur--;

  return MQ_OK;
}

MQ_SIZE
MqBufferLSearchC (
  struct MqBufferLS const * const bufL,
  MQ_STRB const * const str,
  MQ_SIZE const len,
  MQ_SIZE const startindex
)
{
  register struct MqBufferS ** arg = bufL->data + startindex;
  register struct MqBufferS ** const end = bufL->data + bufL->cursize;

  while (arg < end) {
    register struct MqBufferS const * const val = *arg;
    if (val->cursize == len && val->type == MQ_STRT) {
      if (!strcmp (val->cur.C, str)) {
        return bufL->cursize - (end - arg);
      }
    }
    arg++;
  }

  return -1;
}

enum MqErrorE
MqBufferLGetU (
  struct MqErrorS * const error,
  struct MqBufferLS * const bufL,
  MQ_SIZE const index,
  struct MqBufferS ** const var
)
{
  if (likely (bufL->cursize) && likely (index < bufL->cursize)) {
    if (*var) 
      MqBufferDelete(var);
    *var = bufL->data[index];
    MqBufferLDeleteItem (error, bufL, index, 0);
  } else {
    return MqErrorV (error, __func__, -1, "index '" MQ_FORMAT_Z "' is not available", index);
  }

  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                                debugging                                  */
/*                                                                           */
/*****************************************************************************/

#if defined _DEBUG
void
MqBufferLLog (
  struct MqS const * const msgque,
  struct MqBufferLS const * const bufL,
  MQ_STRB const * const func
)
{
  char tmp[20];
  int i;

  if (unlikely(bufL == NULL)) {
    MqDLogV (msgque, func, 0, "%s\n", "++++ MqBufferLS (NULL)");
    return;
  }

  MqDLogV (msgque, func, 0, ">>>> MqBufferLS (%p) (" MQ_FORMAT_Z " items)\n", bufL, bufL->cursize);
  for (i = 0; i < bufL->cursize; i++) {
    sprintf (tmp, "# %i ", i);
    MqBufferLog (msgque, bufL->data[i], tmp);
    if (i < bufL->cursize - 1)
      MqDLogV (msgque, func, 0, "-----------------------------\n");
  }
  MqDLogV (msgque, func, 0, "<<<< MqBufferLS <<<<\n");
}

void
MqBufferLLogS (
  struct MqS const * const msgque,
  struct MqBufferLS const * const bufL,
  MQ_STRB const * const func
)
{
  char tmp[20];
  int i;

  if (unlikely(bufL == NULL)) {
    MqDLogV (msgque, func, 0, "%s\n", "++++ MqBufferLS (NULL)");
    return;
  }

  MqDLogV (msgque, func, 0, ">>>> MqBufferLS (%p) (" MQ_FORMAT_Z " items)\n", bufL, bufL->cursize);
  for (i = 0; i < bufL->cursize; i++) {
    mq_snprintf (tmp, 20, "# %i ", i);
    MqBufferLogS (msgque, bufL->data[i], tmp);
  }
  MqDLogV (msgque, func, 0, "<<<< MqBufferLS <<<<\n");
}
#endif /* _DEBUG */

