/**
 *  \file       src/generic_io.c
 *  \brief      \$Id: generic_io.c 1183 2009-05-18 09:58:23Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1183 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"

#include "generic_io.h"
#include "io_private.h"
#include "buffer.h"
#include "bufferL.h"
#include "error.h"
#include "sys.h"
#include "log.h"
#include "send.h"

#define MQ_ERROR_S    generiC->error
#define MSGQUE	    generiC->io->msgque

/*****************************************************************************/
/*                                                                           */
/*                                 init                                      */
/*                                                                           */
/*****************************************************************************/

#ifdef MQ_IS_WIN32
#if defined(_MSC_VER)
#  define PREFIX
#else
#  define PREFIX static
PREFIX void sGenericCreate (
) __attribute__ ((constructor));

PREFIX void sGenericDelete (
) __attribute__ ((destructor));
#endif /* ! _MSC_VER */

PREFIX void
sGenericCreate (
  void
)
{
  SysWSAStartup (NULL);
}

PREFIX void
sGenericDelete (
  void
)
{
  SysWSACleanup (NULL);
}
#undef PREFIX
#endif

/*****************************************************************************/
/*                                                                           */
/*                                private                                    */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*                                public                                     */
/*                                                                           */
/*****************************************************************************/

/// \attention only the PARENT is using #GenericCreate
enum MqErrorE
GenericCreate (
  struct MqIoS * const io,
  struct MqBufferLS * const argv,
  struct GenericS ** const out
)
{
  struct MqS const * const msgque = io->msgque;
  struct MqErrorS * const error = msgque->error;

  // ok we have an TCP style communication
  register struct GenericS * const generiC = (struct GenericS *const) 
       MqSysCalloc (error, 1, sizeof (*generiC));

  // set the generic return
  if (likely (out != NULL))
    *out = generiC;

  generiC->io = io;
  generiC->error = error;
  generiC->sock = -1;

  return MQ_OK;
}

void
GenericDelete (
  struct GenericS ** const genericP
)
{
  struct GenericS * const generiC = *genericP;
  if (unlikely(generiC == NULL))	return;		// nothing to do
  SysCloseSocket (MQ_ERROR_IGNORE, MQ_YES, &generiC->sock);
  MqSysFree (*genericP);
}

/// \brief create a socket for \e TCP or \e UDS
enum MqErrorE
GenericCreateSocket (
  struct GenericS * const generiC,	///< [in] generic local storage
  int const domain,			///< [in] socket domain (should be: AF_INET or AF_UNIX)
  int const type,			///< [in] socket type (should be: SOCK_STREAM)
  int protocol				///< [in] the socket protocol (should be: 0)
)
{
  // create socket
  MqErrorCheck (SysSocket (MQ_ERROR_S, domain, type, protocol, &generiC->sock));

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

enum MqErrorE
GenericServer (
  struct GenericS * const generiC,	///< the current generic handle
  struct sockaddr * const sockaddr,	///< the address package
  socklen_t const sockaddrlen		///< the length of the address
) {
  struct MqS const * const msgque = MSGQUE;
  struct MqErrorS * const error = MQ_ERROR_S;
  MQ_SOCK child_sock;
  struct sockaddr mysockaddr; 
  socklen_t mysockaddrlen;
  // this is just a flag
  struct MqIdS id;

  // the parent server should "forget" his child's
  // server / child dependency management is done by libmsgque
  if (	    msgque->config.startAs == MQ_START_SPAWN
#if defined(HAVE_FORK)
	||  msgque->config.startAs == MQ_START_FORK
#endif
     ) {
    // do not create a "defunc" process
    MqErrorCheck (SysIgnorSIGCHLD(error));
  }

  // 1. create socket
  MqErrorCheck (GenericBind (generiC, sockaddr, sockaddrlen));

  // 2. listen on port
  MqErrorCheck (SysListen (error, generiC->sock, 128));
  do {
    id.type = MQ_ID_UNUSED;
    mysockaddrlen = sizeof(mysockaddr);
    // 3. accept incomming call
    MqErrorCheck (SysAccept (error, generiC->sock, &mysockaddr, &mysockaddrlen, &child_sock));
    // select: how to start the new task
    switch (msgque->config.startAs) {
      case MQ_START_FORK: {
#if defined(HAVE_FORK)
	MqErrorCheck (pIoStartServer(msgque->io,MQ_START_SERVER_AS_INLINE_FORK, &child_sock, NULL, &id));
#endif
	break;
      }
      case MQ_START_THREAD: {
#if defined(MQ_HAS_THREAD)
	MqErrorCheck (pIoStartServer(msgque->io,MQ_START_SERVER_AS_THREAD, &child_sock, NULL, &id));
	// keep the "child_sock" open, because it's still the same process
#endif
	break;
      }
      case MQ_START_SPAWN: {
	MqErrorCheck (pIoStartServer(msgque->io,MQ_START_SERVER_AS_SPAWN, &child_sock, NULL, &id));
	break;
      }
      case MQ_START_DEFAULT: {
	// server setup without "--fork,--thread or --spawn", this is just a one-time serve
	// of an incomming client-connection
	break;
      }
    }

  // for a NON --fork/--thread/--spawn server the pid will allways be '0' -> continue
  // for a --fork/--thread/--spawn parent the pid will be set by pIoStartServer to something else 
  //   than 0 -> loop
  // for a --fork/--thread/--spawn child the pid will be '0' -> continue
  } while (	(id.type == MQ_ID_PROCESS && id.val.process != 0) 
#if defined(MQ_HAS_THREAD)
	    ||	(id.type == MQ_ID_THREAD  && id.val.thread  != 0UL)	   
#endif
          );

  // fork child (pid=0) close parent socket
  MqErrorCheck (SysCloseSocket (error, MQ_NO, &generiC->sock));
  generiC->sock = child_sock;

  // 3. save socket pointer and listen on socket events
  pIoEventAdd(generiC->io, &generiC->sock);

  // 4. the fork-server should "allow" the SIGCHLD
  if (	    msgque->config.startAs == MQ_START_SPAWN
#if defined(HAVE_FORK)
	||  msgque->config.startAs == MQ_START_FORK
#endif
     ) {
    /*
     * TCL complain if SIGCHLD is ignored 
     * -> additional information from TCL source-code:
     * This changeup in message suggested by Mark Diekhans
     * to remind people that ECHILD errors can occur on
     * some systems if SIGCHLD isn't in its default state.
     */
    MqErrorCheck (SysAllowSIGCHLD(error));
  }

  // 5, finish
  return MQ_OK;

error:
  return MqErrorStack (error);
}

enum MqErrorE
GenericConnect (
  struct GenericS * const generiC,	    ///< [in] the current generic handle
  struct sockaddr * const sockaddr,	    ///< [in] the address package
  socklen_t const sockaddrlen,		    ///< [in] the length of the address
  MQ_TIME_T timeout			    ///< [in] timeout
)
{
  // 1. config socket
  MqErrorCheck (SysSetAsync (MQ_ERROR_S, generiC->sock));

  // 2. connect to server
  MqErrorCheck (SysConnect (MQ_ERROR_S, generiC->sock, sockaddr, sockaddrlen, timeout));

  // 3. add Event to Listen on
  pIoEventAdd(generiC->io, &generiC->sock);

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

enum MqErrorE
GenericBind (
  struct GenericS const * const generiC,    ///< the current generic handle
  struct sockaddr *sockaddr,		    ///< the address package
  socklen_t sockaddrlen 		    ///< the length of the address
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  int status = 1;
  MqErrorCheck (SysSetSockOpt
                (error, generiC->sock, SOL_SOCKET, SO_REUSEADDR,
                 (MQ_sockopt_optval_T) &status, sizeof (status)));
  MqErrorCheck (SysBind (error, generiC->sock, sockaddr, sockaddrlen));

  return MQ_OK;

error:
  return MqErrorStack (error);
}

/*****************************************************************************/
/*                                                                           */
/*                                set/get                                    */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
GenericGetSockName (
  struct GenericS const * const generiC,    ///< the current generic handle
  struct sockaddr *sockaddr,		    ///< the address package
  socklen_t *sockaddrlen 		    ///< the length of the address
)
{
  return SysGetSockName (MQ_ERROR_S, generiC->sock, sockaddr, sockaddrlen);
}

enum MqErrorE
GenericMqGetPeerName (
  struct GenericS const * const generiC,    ///< the current generic handle
  struct sockaddr *sockaddr,		    ///< the address package
  socklen_t *sockaddrlen		    ///< the length of the address
)
{
  return SysGetPeerName (MQ_ERROR_S, generiC->sock, sockaddr, sockaddrlen);
}

