/**
 *  \file       src/io.c
 *  \brief      \$Id: io.c 1182 2009-05-15 13:59:20Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1182 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"

#include "io_private.h"
#include "error.h"
#include "buffer.h"
#include "bufferL.h"
#include "sys.h"
#include "sys_com.h"
#include "uds_io.h"
#include "tcp_io.h"
#include "pipe_io.h"
#include "log.h"
#include "send.h"
#include "error.h"
#include "event.h"

/// \brief undefined socket, initial value for #MqIoS:sockP
static MQ_SOCK sockUndef = -1;

extern struct MqBufferLS * MqInitBuf;
extern MqMainSelectorF MqMainSelector;

/*****************************************************************************/
/*                                                                           */
/*                                io_init                                    */
/*                                                                           */
/*****************************************************************************/

#define MSGQUE io->msgque
#define MQ_ERROR_S MSGQUE->error
#define MQ_GENERIC io->generic

static enum MqErrorE
sIoCheckArg (
  register struct MqIoS * const io,
  struct MqBufferLS * const argv
)
{ 
  struct MqS * const msgque = MSGQUE;
  struct MqErrorS * const error = msgque->error;
  MQ_BOL mybool;
  MQ_INT dummy;

  // for the CHILD this is the default too
  const char *com = "PIPE";
  io->com = IO_PIPE;

  MqErrorCheck (MqBufferLCheckOptionW (error, argv, "--timeout", &io->timeout));
  MqErrorCheck (MqBufferLCheckOptionI (error, argv, "--buffersize", &io->buffersize));
  io->msgque->config.buffersize = io->buffersize;
  io->msgque->config.timeout = io->timeout;

  if (MQ_IS_SERVER(msgque)) {
    // test if the process was started by "pIoStartServer" 
    // if yes -> cleanup predefined "io"
    if (msgque->config.statusIs & MQ_STATUS_IS_DUP) {
      MqBufferLCheckOptionO (error, argv, "--uds" ,	    &mybool);
      MqBufferLCheckOptionO (error, argv, "--tcp" ,	    &mybool);
      MqBufferLCheckOptionO (error, argv, "--pipe",	    &mybool);
      MqBufferLCheckOptionI (error, argv, "--buffersize",  &dummy);
#if defined(MQ_IS_POSIX)
      UdsCleanup (error, argv);
#endif
      TcpCleanup (error, argv);
      com = "PIPE";
      io->com = IO_PIPE;
      goto done;
    }
  }

  MqErrorCheck (MqBufferLCheckOptionO (error, argv, "--uds", &mybool));
  if (mybool) {
#if defined(MQ_IS_WIN32)
    return MqErrorDbV2(MQ_ERROR_S,MQ_ERROR_OPTION_FORBIDDEN,"win32","--uds");
#else
    io->com = IO_UDS;
    com = "UDS";
      MqBufferLCheckOptionO (error, argv, "--tcp" , &mybool);
      MqBufferLCheckOptionO (error, argv, "--pipe", &mybool);
      TcpCleanup (error, argv);
      PipeCleanup (error, argv);
    goto done;
#endif /* MQ_IS_WIN32 */
  }

  MqErrorCheck (MqBufferLCheckOptionO (error, argv, "--tcp", &mybool));
  if (mybool) {
    io->com = IO_TCP;
    com = "TCP";
      MqBufferLCheckOptionO (error, argv, "--pipe", &mybool);
      PipeCleanup (error, argv);
    goto done;
  }

  MqErrorCheck (MqBufferLCheckOptionO (error, argv, "--pipe", &mybool));
  if (mybool) {
    com = "PIPE";
    io->com = IO_PIPE;
    goto done;
  }

done:
  MqDLog (msgque, 4, "option:  io->com = %s\n", com);

  return MQ_OK;

error:
  return MqErrorStack (error);
}

enum MqErrorE
pIoCreate (
  struct MqS * const msgque,
  struct MqBufferLS * const argv,
  struct MqIoS ** const out
)
{
  struct MqErrorS * error = msgque->error;
  struct MqIoS * const io = (struct MqIoS *) MqSysCalloc (error, 1, sizeof (*io));
  if (likely (out != NULL))
    *out = io;

  io->msgque = msgque;
  // parent and child get both this initial value
  // !!attention if child get an other value then it will wait forever (SysWait)
  // during child cleanup
  io->id.type = MQ_ID_UNUSED;
  io->buffersize = msgque->config.buffersize;
  io->timeout = msgque->config.timeout;

  // search for communication type
  if (MQ_IS_PARENT (msgque)) {

    // init parent
    io->sockP = (MQ_SOCK *)&sockUndef;

    // check for arguments
    MqErrorCheck (sIoCheckArg (io, argv));

    // create event-structure
    MqErrorCheck (pEventCreate (msgque, &io->event));

    // create communication layer
    switch (io->com) {
#if defined(MQ_IS_POSIX)
      case IO_UDS:
        MqErrorCheck (UdsCreate (io, argv, &io->iocom.udsSP));
        break;
#endif
      case IO_TCP:
        MqErrorCheck (TcpCreate (io, argv, &io->iocom.tcpSP));
        break;
      case IO_PIPE:
        MqErrorCheck (PipeCreate (io, argv, &io->iocom.pipeSP));
        break;
    }
  } else {
    struct MqIoS * parent = msgque->config.parent->io;
    io->iocom = parent->iocom;
    io->sockP = parent->sockP;
    io->timeout = parent->timeout;
    io->event = parent->event;
  }

  // reset fdset
  FD_ZERO (&io->fdset);

  // create the server socket
  if (MQ_IS_SERVER_PARENT (msgque)) {
    switch (io->com) {
#if defined(MQ_IS_POSIX)
      case IO_UDS:
        MqErrorCheck (UdsServer (io->iocom.udsSP));
        break;
#endif
      case IO_TCP:
        MqErrorCheck (TcpServer (io->iocom.tcpSP));
        break;
      case IO_PIPE:
        MqErrorCheck (PipeServer (io->iocom.pipeSP));
        break;
    }
  }

  return MQ_OK;

error:
  return MqErrorStack (error);
}

void
pIoDelete (
  struct MqIoS ** const ioP
)
{
  struct MqIoS const * io;
  struct MqS const * msgque;

  if (unlikely(ioP == NULL || *ioP == NULL)) return;               // nothing to do
  io = *ioP;
  msgque = io->msgque;

  // ...
  if (MQ_IS_PARENT (msgque)) {

    // drop event
    pEventDelete (io->msgque);

    // cleanup the io data
    switch (io->com) {
#if defined(MQ_IS_POSIX)
      case IO_UDS:
        UdsDelete ((struct UdsS **) & io->iocom);
        break;
#endif
      case IO_TCP:
        TcpDelete ((struct TcpS **) & io->iocom);
        break;
      case IO_PIPE:
        PipeDelete ((struct PipeS **) & io->iocom);
        break;
    }
  }

  MqSysFree (*ioP);

  return;
}

/*****************************************************************************/
/*                                                                           */
/*                                io_atom                                    */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
pIoRead (
  struct MqIoS const * const io,
  register struct MqBufferS * const buf,
  MQ_SIZE const numBytes
)
{
  // alloc space
  pBufferNewSizeDyn (buf, numBytes);
  // reset the buffer
  buf->cur.B = buf->data;
  // read into the buffer
  MqErrorReturn (SysRecv (MQ_ERROR_S, *io->sockP, buf->data, numBytes, &buf->cursize, io->timeout));
}

MQ_INLINE enum MqErrorE
pIoSend (
  struct MqIoS const * const io,
  struct MqBufferS const * const buf
)
{
  MqErrorReturn (SysSend (MQ_ERROR_S, *io->sockP, buf->data, buf->cursize, 0, io->timeout));
}

enum MqErrorE
pIoConnect (
  struct MqIoS * const io
)
{
  switch (io->com) {
#if defined(MQ_IS_POSIX)
    case IO_UDS:
      MqErrorCheck (UdsConnect (io->iocom.udsSP));
      break;
#endif
    case IO_TCP:
      MqErrorCheck (TcpConnect (io->iocom.tcpSP));
      break;
    case IO_PIPE:
      MqErrorCheck (PipeConnect (io->iocom.pipeSP));
      break;
  }

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

MQ_INLINE enum MqErrorE
pIoSelect (
  struct MqIoS * const io,
  enum MqIoSelectE const typ,
  struct timeval * const timeout
)
{
  // set in GenericCreateSocket or in PipeCreate -> SysConnect need this
  register MQ_SOCK const sock = *io->sockP;
  enum MqErrorE ret;
  fd_set *fds, *readfds = NULL, *writefds = NULL;

  if (sock < 0)
    return MQ_CONTINUE;	    // nothing found

  fds = &io->fdset;

  // build fds
  FD_SET (sock, fds);

  // compose
  switch (typ) {
    case MQ_SELECT_RECV:
      readfds = fds;
      break;
    case MQ_SELECT_SEND:
      writefds = fds;
      break;
  }

  // do the select
  MqErrorCheck (ret = SysSelect (MQ_ERROR_S, (sock+1), readfds, writefds, timeout));
  return ret;

error:
  return MqErrorStack (MQ_ERROR_S);
}

MQ_INLINE enum MqErrorE
pIoSelectAll (
  struct MqIoS * const io,
  enum MqIoSelectE const typ,
  struct timeval * const timeout
)
{
  return pEventCheck (MQ_ERROR_S, io->event, typ, timeout);
}

MQ_INLINE enum MqErrorE
pIoSelectStart (
  struct MqIoS * const io,
  struct timeval const * const timeout,
  EventCreateF const proc
)
{
  return pEventStart (MQ_ERROR_S, io->event, timeout, proc);
}

MQ_INLINE enum MqErrorE
pIoShutdown (
  register struct MqIoS const * const io
)
{
  if (io != NULL) {
    MqErrorCheck(SysCloseSocket (MQ_ERROR_S, MQ_YES, io->sockP));
  }

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

/*****************************************************************************/
/*                                                                           */
/*                                 io_misc                                   */
/*                                                                           */
/*****************************************************************************/

static void
sIoFillArgvC (
  struct MqIoS * const io,	    ///< [in] the current io object
  struct MqBufferLS * alfa1,	    ///< [in] befor first #MQ_ALFA
  struct MqBufferLS * alfa2,	    ///< [in] after first #MQ_ALFA
  const MQ_SOCK sock,		    ///< [in] socket for the \e new parent
  char ** arg			    ///< [out] the new argv array to fill
) {
  struct MqS const * const msgque = io->msgque;
  struct MqBufferS **start, **end;

  // 2. copy everything befor the first MQ_ALFA
  if (alfa1) {
    start = alfa1->data;
    end = start+alfa1->cursize;
    for (; start < end; start++) {
      *arg++ = mq_strdup((*start)->cur.C);
    }
  }

  // 3. add PIPE arguments
  *arg++ = mq_strdup("---duplicate");
  *arg++ = mq_strdup("--socket");
  *arg = (char*) MqSysMalloc(MQ_ERROR_PANIC,20);
  mq_snprintf((char*)*arg++,20,"%i",sock);
  *arg++ = mq_strdup("--buffersize");
  *arg = (char*) MqSysMalloc(MQ_ERROR_PANIC,20);
  mq_snprintf((char*)*arg++,20,"%i",io->buffersize);
  *arg++ = mq_strdup("--timeout");
  *arg = (char*) MqSysMalloc(MQ_ERROR_PANIC,20);
  mq_snprintf((char*)*arg++,20,"%lli",io->timeout);

  // 4. add msgque stuff
  if (msgque->config.isSilent) {
    *arg++ = mq_strdup("--silent");
  }
  if (msgque->config.debug) {
    *arg++ = mq_strdup("--debug");
    *arg = (char*) MqSysMalloc(MQ_ERROR_PANIC,5);
    mq_snprintf ((char*)*arg++, 5, "%i", msgque->config.debug);
  }
  switch (msgque->config.startAs) {
    case MQ_START_DEFAULT:
      break;
    case MQ_START_THREAD:
      *arg++ = mq_strdup("--thread");
      break;
    case MQ_START_FORK:
      *arg++ = mq_strdup("--fork");
      break;
    case MQ_START_SPAWN:
      *arg++ = mq_strdup("--spawn");
      break;
  }

  // 5. copy the left over parts starting with the first MQ_ALFA
  if (alfa2) {
    *arg++ = mq_strdup(MQ_ALFA_STR);
    start = alfa2->data;
    end = start+alfa2->cursize;
    for (; start < end; start++,arg++) {
      *arg = mq_strdup((*start)->cur.C);
    }
  }
  *arg = NULL;
}

#if defined(MQ_HAS_THREAD) || defined(HAVE_FORK)
static void
sIoFillArgvU (
  struct MqIoS * const io,	    ///< [in] the current io object
  const MQ_SOCK sock,		    ///< [in] socket for the \e new parent
  struct MqBufferLS * argv	    ///< [out] the new argv array to fill
) {
  struct MqS const * const msgque = io->msgque;
  MqBufferLAppendC(argv, "---duplicate");
  MqBufferLAppendC(argv, "--socket");
  MqBufferLAppendI(argv, sock);
  MqBufferLAppendC(argv, "--buffersize");
  MqBufferLAppendI(argv, io->buffersize);
  MqBufferLAppendC(argv, "--timeout");
  MqBufferLAppendW(argv, io->timeout);
  if (msgque->config.isSilent) {
    MqBufferLAppendC(argv, "--silent");
  }
  if (msgque->config.debug) {
    MqBufferLAppendC(argv, "--debug");
    MqBufferLAppendI(argv, msgque->config.debug);
  }
  switch (msgque->config.startAs) {
    case MQ_START_DEFAULT:
      break;
    case MQ_START_THREAD:
      MqBufferLAppendC(argv, "--thread");
      break;
    case MQ_START_FORK:
      MqBufferLAppendC(argv, "--fork");
      break;
    case MQ_START_SPAWN:
      MqBufferLAppendC(argv, "--spawn");
      break;
  }
}
#endif

enum MqErrorE
pIoStartServer (
  struct MqIoS * const io,
  enum IoStartServerE startType,
  MQ_SOCK * sockP,
  struct MqBufferLS ** alfaP,
  struct MqIdS * idP
) {
  struct MqS * msgque = (struct MqS *) io->msgque;
  struct MqErrorS * const error = msgque->error;
  struct MqBufferLS *alfa1 = NULL, *alfa2 = NULL;
#if defined(MQ_HAS_THREAD) || defined(HAVE_FORK)
  struct MqConfigS * config = NULL;
#endif
#if defined(MQ_HAS_THREAD)
# if defined(HAVE_PTHREAD)
    int thread_status = PTHREAD_CREATE_DETACHED;
# elif defined(MQ_IS_WIN32)
    int thread_status = 0;
# endif
#endif

  // MainSelector related variables
  MqCreateF fMain = NULL;
  const char * name = NULL;

  // alfa is now owned by "pIoStartServer"
  if (alfaP != NULL) {
    alfa1 = *alfaP;
    *alfaP = NULL;
  }

//I0
//MqBufferLLogS(msgque,alfa1,"alfa1");
//printI(startType)

  // select the code
rescan:
  switch (startType) {
    case MQ_START_SERVER_AS_PIPE: {
	int isSELF = MQ_NO;
//printLC("MQ_START_SERVER_AS_PIPE:")
	// case 2: this is used if the function is called from MqCreate. The
	// msgque->alfa argument has the startup arguments of the IO_PIPE server.
	// IO_TCP and IO_UDS is ignored and it is an error if msgque->alfa is
	// available too.
	if (io->com != IO_PIPE) {
	  // for non "PIPE" io the "alfa" have to be empty
	  if (alfa1 != NULL)
	    return MqErrorDbV2 (error, MQ_ERROR_CAN_NOT_START_NON_PIPE_SERVER, pIoLogCom(io->com));
	  // for non "PIPE" io we have nothing to do
	  return MQ_OK;
	}
	if (alfa1 == NULL) {
	  return MqErrorDb (MQ_ERROR_CAN_NOT_START_CLIENT);
	}
	// setup alfa1/2
	pBufferLSplitAlfa(&alfa1, &alfa2);
	// get the socket from the pipe
	sockP = PipeGetServerSocket(io->iocom.pipeSP);
	// get the id storage
	idP = &io->id;
	// this is the first entry in alfa1
	name = alfa1->data[0]->cur.C;
	// check if we use the "SELF" keyword
	if (msgque->isSELF || msgque->isWORKER) {
	  isSELF = MQ_YES;
	  // replace "SELF" with "MqInitBuf" data
	  name = MqInitBuf->data[0]->cur.C;
	  // replace "SELF" itself on position "0"
	  MqErrorCheck (MqBufferLDeleteItem (error, alfa1, 0, MQ_NO));
	  // add startup entry function (only used for a WORKER)
	  if (msgque->isWORKER)
	    fMain = (msgque->config.fParentCreate != NULL ? msgque->config.fParentCreate : MqCreate);
	} else {
	  // does we use MqMainSelector and is the first entry in alfa1 found?
	  fMain = (MqMainSelector ? MqMainSelector(name) : NULL);

	  // if yes then use MqMainToolName (if available)
	  // > atool split ... @ cut ... @ join ...
	  // name is "cut" and next line will replace the name with "atool"
	  name = ( fMain && MqInitBuf ? MqInitBuf->data[0]->cur.C : name );
	}

//printLC(name)

	if (fMain) {
#if defined(HAVE_FORK)
	  // continue with "fork"
	  if (msgque->config.ignoreFork == MQ_NO && (
		msgque->config.startAs == MQ_START_DEFAULT || 
		msgque->config.startAs == MQ_START_FORK)) {
//printLC("fork")
	    startType = MQ_START_SERVER_AS_FORK;
	    // if isSELF than the startup is like a "GenericServer" and not like a "pipe"
	    if (isSELF == MQ_YES) {
	      MqBufferLAppend(alfa1, MqBufferCreateC(name), 0);
	    } else {
	      config = MqConfigInit();
	      config->debug = msgque->config.debug;
	      config->isSilent = msgque->config.isSilent;
	      config->startAs = msgque->config.startAs;
	    }
	    goto rescan;
	  }
#endif
#if defined(MQ_HAS_THREAD)
	  // continue with "thread"
	  if (msgque->config.ignoreThread == MQ_NO && (
		msgque->config.startAs == MQ_START_DEFAULT || 
		msgque->config.startAs == MQ_START_THREAD)) {
//printLC("thread")
	    startType = MQ_START_SERVER_AS_THREAD;
	    // if isSELF than the startup is like a "GenericServer" and not like a "pipe"
	    if (isSELF == MQ_YES) {
	      MqBufferLAppend(alfa1, MqBufferCreateC(name), 0);
	    } else {
	      config = MqConfigInit();
	      config->debug = msgque->config.debug;
	      config->isSilent = msgque->config.isSilent;
	      config->startAs = msgque->config.startAs;
	    }
#if defined(HAVE_PTHREAD)
	    thread_status = PTHREAD_CREATE_JOINABLE;
#endif
	    goto rescan;
	  }
#endif
	}
	// continue with "spawn" (works allways)
	if (msgque->config.ignoreSpawn == MQ_YES) {
	  MqErrorDbV(MQ_ERROR_CAN_NOT_START_NON_PIPE_SERVER,(name == NULL ? "unknown" : name));
	}
//printLC("spawn")
	startType = MQ_START_SERVER_AS_SPAWN;
	// only SPAWN need all arguments from MqInitBuf
	if (isSELF == MQ_YES || fMain != NULL) {
	  MqBufferLAppendL(alfa1, MqInitBuf, 0);
	}

	goto rescan;
      }
      break;
#if defined(MQ_HAS_THREAD)
    case MQ_START_SERVER_AS_THREAD: {
//printLC("MQ_START_SERVER_AS_THREAD:")
	// copy the configuration from the "PARENT" server
	if (config == NULL) {
	  config = MqConfigDup(&msgque->config);
	  config->isServer = MQ_YES;
	}
	// find the entry-point
	if (!fMain) fMain = (msgque->config.fParentCreate?msgque->config.fParentCreate:MqCreate);

	// MqMainToolName set in "sMqCheckArg" or at application main startup
	if (name == NULL) {
	  if (MqInitBuf == NULL) {
	    return MqErrorDb(MQ_ERROR_NO_INIT);
	  } else {
	    name = MqInitBuf->data[0]->cur.C;
	  }
	}

	// empty alfa1 (GenericServer) starup -> init with MqInitBuf data
	if (alfa1 == NULL) {
	  alfa1 = MqBufferLCreate(20);
	  MqBufferLAppendC(alfa1, name);
	}

	// fill arg with system-arguments
	sIoFillArgvU(io,*sockP,alfa1);
	// start the server

/*
I2
MqBufferLLogS(msgque, alfa1, "alfa1");
MqBufferLLogS(msgque, alfa2, "alfa2");
*/

	MqErrorCheck (SysServerThread (error, config, alfa1, alfa2, fMain, name, thread_status, idP));
	// alfa1/2 is now owned by the thread
	alfa1 = alfa2 = NULL;
      }
      break;
#endif /* MQ_HAS_THREAD */
#if defined(HAVE_FORK)
    case MQ_START_SERVER_AS_FORK: {
//printLC("MQ_START_SERVER_AS_FORK:")
	// copy the configuration from the "PARENT" server
	if (config == NULL) {
	  config = MqConfigDup(&msgque->config);
	  config->isServer = MQ_YES;
	}

	// find the entry-point
	if (!fMain) fMain = (msgque->config.fParentCreate?msgque->config.fParentCreate:MqCreate);

	// MqMainToolName set in "sMqCheckArg" or at application main startup
	if (name == NULL) {
	  if (MqInitBuf == NULL) {
	    return MqErrorDb(MQ_ERROR_NO_INIT);
	  } else {
	    name = MqInitBuf->data[0]->cur.C;
	  }
	}

	// empty alfa1 (GenericServer) starup -> init with MqInitBuf data
	if (alfa1 == NULL) {
	  alfa1 = MqBufferLCreate(20);
	  MqBufferLAppendC(alfa1, name);
	}

	// fill arg with system-arguments
	sIoFillArgvU(io,*sockP,alfa1);

	// start the server
	MqErrorCheck (SysServerFork (error, config, alfa1, alfa2, fMain, name, idP));
	// cleanup socket of the child
	MqErrorCheck(SysCloseSocket (error, MQ_NO, sockP));
      }
      break;
#endif   /* HAVE_FORK */
    case MQ_START_SERVER_AS_SPAWN: {
//printLC("MQ_START_SERVER_AS_SPAWN:")
	char **argV, **arg;
	// add 20 item's as additional space
	argV = arg = (char **) MqSysMalloc (error, sizeof(*alfa1) * (
	    (alfa1 != NULL ? alfa1->cursize : 0) + (alfa2 ? alfa2->cursize : 0) + 20 
	));

	// init main
	if (name == NULL) {
	  if (MqInitBuf == NULL) {
	    return MqErrorDb(MQ_ERROR_NO_INIT);
	  } else {
	    name = MqInitBuf->data[0]->cur.C;
	  }
	}

	// empty alfa1 (GenericServer) starup -> init with MqInitBuf data
	if (alfa1 == NULL) {
          int i;
	  for (i=0; i<MqInitBuf->cursize; i++) {
	    *arg++ = mq_strdup(MqInitBuf->data[i]->cur.C);
	  }
	}

	// fill arg with system-arguments
	sIoFillArgvC(io, alfa1, alfa2, *sockP, arg);
	// start the server

/*
//printLC(name)
int i;
char ** xarg = argV;
for (i=0; *xarg != NULL; xarg++, i++) {
  MqDLogV (msgque, __func__, 0, "alfa1[%2i]=%s\n",i, *xarg);
}
*/

	MqErrorCheck (SysServerSpawn (error, argV, name, idP));
	// cleanup array's
	SysFreeArgvArray (&argV);
	// cleanup socket of the child
	MqErrorCheck(SysCloseSocket (error, MQ_NO, sockP));
      }
      break;
#if defined(HAVE_FORK)
    case MQ_START_SERVER_AS_INLINE_FORK: {
//printLC("MQ_START_SERVER_AS_INLINE_FORK:")
	if (name == NULL && MqInitBuf != NULL) {
	  name = MqInitBuf->data[0]->cur.C;
	} else {
	  return MqErrorDb(MQ_ERROR_NO_INIT);
	}
	// case: the function GenericServer is listen on a "tcl" or "uds" connection
	// and start for every incomming connection a new process with "fork". !no!
	// initialization have to be done because we reuse the current state.
	MqErrorCheck (SysFork (error, idP));
	// if pid != 0 -> this is the parent just continue with default "client" code
	if ((*idP).val.process == 0) {
	  // pid = 0 -> this is the fork child
	  // the daemon_pid_file only belongs to the parent
	  msgque->config.statusIs = (enum MqStatusIsE) (msgque->config.statusIs | MQ_STATUS_IS_FORK);
	  msgque->daemon_pid_file = NULL; 
	  // we don't need to cleanup argv1/2 because the INLINE_FORK is done
	  // with an empty "alfa"
	  return MQ_OK;
	}
	// cleanup socket of the child
	MqErrorCheck(SysCloseSocket (error, MQ_NO, sockP));
      }
      break;
#endif   /* HAVE_FORK */
  }

  // ok 2: inform the user
  if (likely(msgque != NULL) && unlikely(msgque->config.debug >= 4)) {
    MQ_STRB buf[50];
    // !attention "name" can be "alfa1[0]"
    MqDLogV(msgque,__func__,4,"finish start server '%s' as '%s' with ID '%s'\n",
		name, pIoLogStartType(startType), pIoLogId(buf,50,*idP));
  }

  // cleanup alfa1/2
  MqBufferLDelete (&alfa1);
  MqBufferLDelete (&alfa2);

  return MQ_OK;

error:
  return MqErrorStack (error);
}

MQ_BOL
pIoCheck (
  struct MqIoS * const io
) {
  return (io && io->sockP ? (*(io->sockP) >= 0) : MQ_NO);
}

const char*
pIoLogStartType (
  enum IoStartServerE startType
) {
  switch (startType) {
    case MQ_START_SERVER_AS_PIPE:	    return "pipe";
#if defined(HAVE_FORK)
    case MQ_START_SERVER_AS_FORK:	    return "fork";
    case MQ_START_SERVER_AS_INLINE_FORK:    return "MQ_INLINE fork";
#endif
#if defined(MQ_HAS_THREAD)
    case MQ_START_SERVER_AS_THREAD:	    return "thread";
#endif
    case MQ_START_SERVER_AS_SPAWN:	    return "spawn";
  }
  return "";
}

/// \brief helper to covert the #MqIdS into a human readable string
/// \retval the input-buffer (\e buf) filled with the string-representation of the \e id
MQ_STR
pIoLogId (
  MQ_STR buf,		    ///< [in,out] input-buffer used as result-buffer too
  MQ_INT size,		    ///< [in] size of the input-buffer
  struct MqIdS id	    ///< [in] value to print
) {
  switch (id.type) {
    case MQ_ID_PROCESS:
      mq_snprintf(buf,size,"%i",id.val.process);
      break;
#if defined(MQ_HAS_THREAD)
    case MQ_ID_THREAD:
      mq_snprintf(buf,size,"%lu",(long unsigned int)id.val.thread);
      break;
#endif
    case MQ_ID_UNUSED:
      strcpy(buf,"unknown");
      break;
  }
  return buf;
}

const char *
pIoLogCom (
  enum IoComE com
) {
  switch (com) {
#if defined(MQ_IS_POSIX)
    case IO_UDS:   return "UDS";
#endif
    case IO_TCP:   return "TCP";
    case IO_PIPE:  return "PIPE";
  }
  return "";
}

/*****************************************************************************/
/*                                                                           */
/*                           io_event_atom                                   */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
pIoEventAdd(
  struct MqIoS * const io,	///< [in] io interface data
  MQ_SOCK * sockP		///< [in] socket pointer
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  MQ_INT const buffersize = io->buffersize;
  const MQ_SOCK sock = *sockP;
  MQ_INT current;
  socklen_t len;

  // 1. set the buffersize
  len = sizeof (MQ_INT);
  MqErrorCheck (SysGetSockOpt (error, sock, SOL_SOCKET, SO_SNDBUF, (MQ_sockopt_optval_T) &current, &len));
//MqDLogV(MSGQUE,__func__,0,"buffersize<%i>, current<%i>\n", buffersize, current);
  if (current < buffersize) {
    len = sizeof (MQ_INT);
    MqErrorCheck (SysSetSockOpt (error, sock, SOL_SOCKET, SO_SNDBUF, (MQ_sockopt_optval_T) &buffersize, len));
    MqErrorCheck (SysGetSockOpt (error, sock, SOL_SOCKET, SO_SNDBUF, (MQ_sockopt_optval_T) &current, &len));
    MqDLog (MSGQUE, 0, "set SO_SNDBUF to '%d'\n", current);
  }
  len = sizeof (MQ_INT);
  MqErrorCheck (SysGetSockOpt (error, sock, SOL_SOCKET, SO_RCVBUF, (MQ_sockopt_optval_T) &current, &len));
  if (current < buffersize) {
    len = sizeof (MQ_INT);
    MqErrorCheck (SysSetSockOpt (error, sock, SOL_SOCKET, SO_RCVBUF, (MQ_sockopt_optval_T) &buffersize, len));
    MqErrorCheck (SysGetSockOpt (error, sock, SOL_SOCKET, SO_RCVBUF, (MQ_sockopt_optval_T) &current, &len));
    MqDLog (MSGQUE, 0, "set SO_RCVBUF to '%d'\n", current);
  }

  // 2. set to asyncrone
  MqErrorCheck (SysSetAsync (error, sock));

  // 3. save
  pEventAdd (io->msgque, sockP);

  return MQ_OK;

error:
  return MqErrorStack(error);
}

MQ_INLINE MQ_TIME_T
pIoGetTimeout(
  struct MqIoS * const io	///< [in] io interface data
)
{
  return io->timeout;
}

MQ_BOL 
pIoIsRemote (
  struct MqIoS * const	io
)
{
  switch (io->com) {
#if defined(MQ_IS_POSIX)
    case  IO_UDS:
#endif
    case  IO_TCP:   return MQ_YES;
    case  IO_PIPE:  return MQ_NO;
  }
  return MQ_NO;
}

