/**
 *  \file       src/msgque_private.h
 *  \brief      \$Id: msgque_private.h 1184 2009-05-18 11:54:15Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1184 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#ifndef MQ_PRIVATE_H
#define MQ_PRIVATE_H

#if defined(MQ_HAS_THREAD)
# if defined(HAVE_PTHREAD)
#  include <pthread.h>
#  define mqthread_t pthread_t
# elif defined(MQ_IS_WIN32)
#  define mqthread_t unsigned long
# endif
#endif /* MQ_HAS_THREAD */

#include <sys/types.h>

/*****************************************************************************/
/*                                                                           */
/*                            msgque/definition                              */
/*                                                                           */
/*****************************************************************************/

struct UdsS;
struct TcpS;
struct PipeS;
struct GenericS;
struct MqCacheS;
struct MqTransItemS;

#define MQ_IS_SERVER(msgque) (msgque->config.isServer == MQ_YES)
#define MQ_IS_SERVER_PARENT(msgque) (MQ_IS_SERVER(msgque) && MQ_IS_PARENT(msgque))
#define MQ_IS_CLIENT(msgque) (msgque->config.isServer == MQ_NO)
#define MQ_IS_CLIENT_PARENT(msgque) (MQ_IS_CLIENT(msgque) && MQ_IS_PARENT(msgque))
#define MQ_IS_CHILD(msgque)  (msgque->config.parent != NULL)
#define MQ_IS_SLAVE(msgque)  (msgque->config.master != NULL)
#define MQ_IS_PARENT(msgque) (msgque->config.parent == NULL)
#define MQ_IS_FILTER(msgque) (msgque->config.isFilter == MQ_YES)

#define MQ_STRING_TYPE(isString) (unlikely(isString) ? MQ_STRT : MQ_BINT)

/// \brief get a \e MqErrorS object from a \e MgS in a \e save manner
#define MQ_SAVE_ERROR(msgque) ((msgque)&&(msgque)->error?(msgque)->error:MQ_ERROR_IGNORE)

/// \brief separator between Message-Queue-Package-Data
#define BUFFER_CHAR         '+'
#define BUFFER_CHAR_S       "+"

/// \brief data type for a socket handle
typedef MQ_INT		    MQ_SOCK;

enum MqIdSE {
  MQ_ID_UNUSED,
  MQ_ID_PROCESS,
#if defined(MQ_HAS_THREAD)
  MQ_ID_THREAD
#endif
};

union MqIdSU {
  pid_t	process;	
#if defined(MQ_HAS_THREAD)
  mqthread_t	thread;
#endif
};

/// \brief data type for process/thread identification
///  
/// The values of \e pid_t and #mqthread_t have to fit into this storage
struct MqIdS {
  enum MqIdSE type;
  union MqIdSU val;
};


/*****************************************************************************/
/*                                                                           */
/*                          protocol reference                               */
/*                                                                           */
/*****************************************************************************/

/**
\defgroup protocol INTERNAL: Package - Protocol - Reference
\{
Every data-package send from a client to the server and back is a combination of
a \b fix size HEADER and a \b variable size BODY.

The fix size HEADER (defined in \ref HdrS) was defined as:

\verbatim
HDR+CONTEXT+BODYSIZE+TOKEN+TRANSACTION+RETURNCODE+
\endverbatim

If "BODYSIZE > 0" the BODY_ITEM_HEADER (defined in \ref BdyS) is appended to HEADER:

\verbatim
BDY+BODY_ITEM_SIZE+BODY_ITEM_LIST
\endverbatim

with BODY_ITEM_SIZE as the number of items in the BODY_ITEM_LIST.

If "BODY_ITEM_SIZE > 0" the BODY_ITEM_SIZE * BODY_DATA_HEADER are appended to BODY_ITEM_HEADER

\verbatim
BODY_DATA_TYPE+BODY_DATA_SIZE+BODY_DATA
\endverbatim

with BODY_DATA_SIZE as the number of bytes in BODY_DATA.

If RETURNCODE == #MQ_RETURN_ERROR (defined in #MqReturnE) the \e last BODYITEM is defined 
as RETURNITEM with a RETURNITEM_HEADER:

\verbatim
RETURN_NUMBER+RETURN_MESSAGE_SIZE+...data...
\endverbatim

*/

/// \brief max size of a native handle (Type: #MQ_INT - 4 byte)
#define HDR_HDL_LEN	4
/// \brief max size of an hex integer in characters (Type: #MQ_INT)
#define HDR_INT_LEN	8
/// \brief max size of an pToken (Type: char[4])
#define HDR_TOK_LEN	4
/// \brief size of the return code (N)othing, (O)k, (W)arning, (E)rror (Type: char)
#define HDR_RET_LEN	1

/*
 *	Header-Prefix
 */

#define HDR_CtxId_S	    (3					+ 1)        ///< 4  : "HDR" string[3]
#define	HDR_BdySize_S	    (HDR_CtxId_S	+ HDR_INT_LEN	+ 1)        ///< 13 : "CONTEXT" integer
#define HDR_Token_S	    (HDR_BdySize_S	+ HDR_INT_LEN	+ 1)        ///< 22 : "BODYSIZE" integer
#define HDR_Trans_S	    (HDR_Token_S	+ HDR_TOK_LEN	+ 1)        ///< 27 : "TOKEN" string[4]
#define HDR_Code_S	    (HDR_Trans_S	+ HDR_HDL_LEN	+ 1)        ///< 32 : "TRANSACTION" pointer
#define HDR_SIZE	    (HDR_Code_S		+ HDR_RET_LEN	+ 1)        ///< 34 : "RETURNCODE" char

/*
 *	Body-Prefix
 */

#define BDY_NumItems_S	    (3				    + 1)    ///< 4  : "BDY" string[3]
#define BDY_SIZE	    (BDY_NumItems_S + HDR_INT_LEN)	    ///< 12 : "BODY_ITEM_SIZE" integer

/*
 *      Return-Header (RET)
 */

#define RET_Num_S	   (0                                 )	    ///< 0  : Start of the "RETURNITEM"
#define RET_NumItems_S     (RET_Num_S      + HDR_INT_LEN   + 1)	    ///< 9  : "RETURM_NUMBER" integer
#define RET_SIZE           (RET_NumItems_S + HDR_INT_LEN      )	    ///< 17 : "RETURN_MASSAGE_SIZE" integer

/*
 *      Body-Object-Header (BOH)
 */

/// \brief the HEADER size of an BODYITEM

//  Definition:                 CHAR + TYP + CHAR + LENGTH      + CHAR 
#   define BUFFER_P2_PRENUM    (1    + 1   + 1    + HDR_INT_LEN + 1   )
 
//  Definition:                 CHAR + TYP + CHAR  
#   define BUFFER_P2_NATIVE    (1    + 1   + 1    )

/*
 *	Misc stuff
 */ 

#define validateHDR(buf) \
    printf("VALIDATE(HDR) : %c%c%c %c ctxId %c size %c token %c transptr %c\n", \
	buf->data[0], buf->data[1], buf->data[2], \
	buf->data[HDR_CtxId_S - 1], \
	buf->data[HDR_BdySize_S - 1], buf->data[HDR_Token_S - 1], \
	buf->data[HDR_Trans_S - 1], buf->data[HDR_SIZE - 1])


/*****************************************************************************/
/*                                                                           */
/*                                msgque/packed                              */
/*                                                                           */
/*****************************************************************************/

#pragma pack(1)

/// \brief \e integer representation of a BODYITEM size
union HdrBdyIU {
    MQ_INT	B		; ///< \e binary representation of a BODYITEM size
    MQ_STRB	S[6]		; ///< \e string representation of a BODYITEM size (!ATTENTION max 999999 bytes)
};

/// \brief \e integer representation
union HdrIU {
    MQ_INT	B		; ///< \e binary representation of a integer
    MQ_STRB	S[HDR_INT_LEN]	; ///< \e string representation of a integer
};

/// \brief The protocol HEADER mapped to a \e struct
struct HdrS {
    MQ_STRB	    ID1	    ;   ///< <TT>size: 1 .............. -> total: 1 .. -> character .. -> "H"</TT>
    MQ_STRB	    ID2	    ;   ///< <TT>size: 1 .............. -> total: 2 .. -> character .. -> "D"</TT>
    MQ_STRB	    ID3	    ;   ///< <TT>size: 1 .............. -> total: 3 .. -> character .. -> "R"</TT>
    MQ_STRB	    charI   ;   ///< <TT>size: 1 .............. -> total: 4 .. -> character .. -> "+"</TT>
    union HdrIU	    ctxId   ;   ///< <TT>size: HDR_INT_LEN=8 .. -> total: 12 . -> integer .... -> "CONTEXT"</TT>
    MQ_STRB	    charS   ;   ///< <TT>size: 1 .............. -> total: 13 . -> character .. -> "+"</TT>
    union HdrIU	    size    ;   ///< <TT>size: HDR_INT_LEN=8 .. -> total: 21 . -> integer .... -> "BODYSIZE"</TT>
    MQ_STRB	    charO   ;   ///< <TT>size: 1 .............. -> total: 22 . -> character .. -> "+"</TT>
    MQ_STRB	    tok[4]  ;   ///< <TT>size: 4 .............. -> total: 26 . -> string[4] .. -> "TOKEN" </TT>
    MQ_STRB	    charT   ;   ///< <TT>size: 1 .............. -> total: 27 . -> character .. -> "+"</TT>
    MQ_HDL	    trans   ;   ///< <TT>size: 4 .............. -> total: 31 . -> handle ..... -> "TRANSACTION"</TT>
    MQ_STRB	    charC   ;   ///< <TT>size: 1 .............. -> total: 32 . -> character .. -> "+"</TT>
    MQ_STRB	    code    ;   ///< <TT>size: 1 .............. -> total: 33 . -> character .. -> "RETURNCODE"</TT>
    MQ_STRB	    charE   ;   ///< <TT>size: 1 .............. -> total: 34 . -> character .. -> "+"</TT>
};

/// \brief The protocol HEADER-BODY part mapped to a \e struct
struct BdyS {
    MQ_STRB	    ID[3]   ;   ///< <TT>size: 3 .............. -> total: 3 .. -> string[3] .. -> "BDY" </TT>
    MQ_STRB	    charN   ;   ///< <TT>size: 1 .............. -> total: 4 .. -> character .. -> "+"</TT>
    union HdrIU	    num	    ;   ///< <TT>size: HDR_INT_LEN=8 .. -> total: 12 . -> integer .... -> "BODYITEM_SIZE"</TT>
    MQ_STRB	    charE   ;   ///< <TT>size: 1 .............. -> total: 13 . -> character .. -> "+"</TT>
};

/// \brief the entire HEADER 
struct HdrSendS {
    struct HdrS	    hdr	    ;   ///< <TT>size: 38 ............. -> total: 38 . -> HdrS ....... -> "HEADER" </TT>
    struct BdyS	    bdy	    ;   ///< <TT>size: 13 ............. -> total: 59 . -> BdyS ....... -> "HEADER-BODY" </TT>
};

#pragma pack()

/// \}

/*****************************************************************************/
/*                                                                           */
/*                           private data struc/union                        */
/*                                                                           */
/*****************************************************************************/

/** \brief the main \libmsgque object.
 *
 * \copydoc msgque_api
 */
struct MqS {

  // public variables
  struct MqConfigS config;	    ///< public #MqS configuration

  struct MqBufferS * temp;	    ///< misc temporary \e MqBufferS object

  struct MqSendS  * send;	    ///< object for sending a Msgque packet
  struct MqReadS  * read;	    ///< object for reading a Msgque packet
  struct MqIoS    * io;		    ///< object for management of the 'socket' infrastructure
  struct MqErrorS * error;	    ///< object used to store error messages

  // private variables
  struct MqS * myFilter;	    ///< link to the #MqS object of the filter (filter only)
  MQ_BOL endian;		    ///< a endian switch have to be done? (boolean: MQ_YES or MQ_NO)

  // context-management variables
  MQ_SIZE   ctxId;		    ///< the ctxId of this MqS object
  struct MqS *	    ctxIdP;	    ///< the inital (first) context (home of the ctxIdA)

  // private variables
  struct MqTokenS * srvT;	    ///< identifer for the 'service' token handle

  MQ_BOL onExit;		    ///< is allready an exit ongoing?
  struct MqS * exit_msgque;	    ///< msgque object got and "_SHD" request (only used at the parent)
  MQ_BOL onDelete;		    ///< is allready a "exit" ongoing?
  MQ_BOL onShutdown;		    ///< is allready a "shutdown" ongoing?
  MQ_STR daemon_pid_file;	    ///< the argument of the \c --daemon option

  struct MqCacheS * readCache;	    ///< cache for MqReadS

  // the next 3 items are !!only!! used in the parent
  MQ_SIZE   ctxIdR;		    ///< the largest currently used ctxId number
  MQ_SIZE   ctxIdZ;		    ///< the size of the ctxIdA array
  struct MqS ** ctxIdA;		    ///< array of struct MqS * pointer's

  // the next 3 items are used to map the transactionID (int) to the transaction pointer
  struct MqTransS * trans;	    ///< link to the trans object
  MQ_HDL _trans;		    ///< storage for the Transaktion object from the package header

  // the following lines manage the link between the parent and the child,
  // to be able to delete all childs if the parent is deleted
  struct pChildS * childs;	    ///< linked list of childs
  struct pChildS * self;	    ///< my own child storage

  // master/slave relationship
  struct MqSlaveS * slave;	    ///< link to the SLAVE object
  MQ_BOL  isSELF;		    ///< is alfa[0] is "SELF"
  MQ_BOL  isWORKER;		    ///< is alfa[0] is "WORKER"
};

/*****************************************************************************/
/*                                                                           */
/*                                msgque/event                               */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE pWaitOnEvent (
  struct MqS * const msgque,
  enum MqIoSelectE const what,
  MQ_TIME_T const timeout
);

enum MqErrorE pUSleep (
  register struct MqS const * const msgque,
  long const usec
);

MQ_INLINE struct MqS * pMqGetFirstParent(
  struct MqS * const msgque
);

void pMqShutdown (
  struct MqS * const msgque
);

#if (HDR_TOK_LEN+1) == 5
#define ppTokenCopy(i1,i2) memcpy(i1,i2,5);
#else
#   error invalid TOKEN_LEN
#endif

#endif /* MQ_PRIVATE_H */



