/**
 *  \file       src/read.c
 *  \brief      \$Id: read.c 1183 2009-05-18 09:58:23Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1183 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "main.h"
#include "read.h"
#include "error.h"
#include "buffer.h"
#include "sys.h"
#include "mq_io.h"
#include "log.h"
#include "send.h"
#include "token.h"
#include "swap.h"
#include "cache.h"

/*****************************************************************************/
/*                                                                           */
/*                                read_init                                  */
/*                                                                           */
/*****************************************************************************/

#define MQ_ERROR_S read->msgque->error
#define MSGQUE read->msgque
#define TOKEN_LEN (HDR_TOK_LEN+1)

/*****************************************************************************/
/*                                                                           */
/*                                read_native                                */
/*                                                                           */
/*****************************************************************************/

/// \brief special data needed to save the current-status for a LST object
struct ReadSaveS {
  struct ReadSaveS * save;	    ///< this is needed for recursion
  struct MqBufferS bdy;		    ///< used for BDY data (dynamic)
};

struct MqReadS {
  struct MqS * msgque;
  struct MqBufferS * hdr;	    ///< used for HDR data (static)
  struct MqBufferS * bdy;	    ///< used for BDY data (dynamic)
  struct MqBufferS * cur;	    ///< used as reference on BUF with the current data
  enum MqReturnE returnCode;	    ///< Return-Code (O)K, (E)RROR or (W)ARNING
  MQ_INT returnNum;		    ///< Return-Number
  struct MqCacheS * saveCache;
  struct ReadSaveS * save;	    ///< need for List objects
  enum MqTypeE type;		    ///< type of the item stored into the data-segment (InitialSet)
  MQ_BOL canUndo;		    ///< is an MqReadUndo allowed ?
};

/*****************************************************************************/
/*                                                                           */
/*                               read_static                                 */
/*                                                                           */
/*****************************************************************************/

static struct MqReadS*
sReadC (
  struct MqErrorS * error
)
{
  return (struct MqReadS *) MqSysMalloc (error, sizeof (struct ReadSaveS));
}

static void
sReadD (
  struct MqReadS ** const ptr
)
{
  MqSysFree (*ptr);
}

///pReadCreate have to fit into 'CacheCF' function
struct MqReadS*
pReadCreate (
  register struct MqS * const msgque
)
{
  struct MqErrorS * error = msgque->error;
  register struct MqReadS * const read = (struct MqReadS *const) MqSysCalloc (error, 1, sizeof (*read));
  read->msgque = msgque;

//MqDLogV(read->msgque,__func__,-1,"read<%p>\n", read);

  read->hdr = MqBufferCreate (HDR_SIZE + 1);
  read->bdy = MqBufferCreate (10000);
  read->canUndo = MQ_NO;

  // references
  read->cur = pBufferCreateRef (read->bdy);

  read->hdr->cursize = HDR_SIZE;

  // create the lst Cache data
  pCacheCreate (error, (CacheCF) sReadC, (CacheDF) sReadD, error, &read->saveCache);

  // cache the current package-type
  read->type = MQ_STRING_TYPE(msgque->config.isString);

  return read;
}

void
pReadDelete (
  register struct MqReadS ** const readP
)
{
  struct MqReadS * const read = *readP;

  if (unlikely (read == NULL))
    return;

  // rollback all open LST/RET blocks
  while (read->save != NULL) {
    MqReadL_END (read);
  }

  // delete cache
  pCacheDelete (&read->saveCache);

  MqBufferDelete (&read->hdr);
  MqBufferDelete (&read->bdy);

  // references
  MqBufferDelete (&read->cur);

  MqSysFree (*readP);
}

enum MqErrorE
MqReadL_START (
  struct MqReadS * const read,
  MQ_BUF buf
)
{
  struct ReadSaveS * save;
  register struct MqBufferS * const bdy = read->bdy;

  // get the buffer
  if (buf == NULL)
    MqErrorCheck(MqReadU(read,&buf));

  // check buffer type
  if (buf->type != MQ_LSTT)
    return MqErrorDbV(MQ_ERROR_TYPE, MqLogTypeName(MQ_LSTT), MqLogTypeName(buf->type));

  save = (struct ReadSaveS *) pCachePop (read->saveCache);
  
//MqDLog(msgque,0,"read<%p>, save<%p>\n", read, save);

  save->save = read->save;
  save->bdy  = *bdy;
  read->save = save;

  bdy->type = read->type;
  bdy->data = bdy->cur.B = buf->data;
  bdy->size = buf->size;
  bdy->cursize = buf->cursize;
  bdy->numItems = (read->type==MQ_BINT ? MqBufU2INT(bdy->cur) : str2int(bdy->cur.C,NULL,16));
  bdy->cur.B += (HDR_INT_LEN + 1);
  
  return MQ_OK;

error:
  return MqErrorStack(MQ_ERROR_S);
}

void
MqReadL_END (
  struct MqReadS * const read
)
{
  struct ReadSaveS * save = read->save;
  if (unlikely(save == NULL)) return;

  read->save = save->save;
  *read->bdy = save->bdy;

//MqDLog(MSGQUE,0,"read<%p>, save<%p>\n", read, save);
  pCachePush (read->saveCache, save);
}

MQ_INLINE enum MqErrorE
MqReadL_STARTM (
  struct MqS const * const msgque,
  MQ_BUF buf
)
{
  return MqReadL_START (msgque->read, buf);
}

MQ_INLINE void
MqReadL_ENDM (
  struct MqS const * const msgque
)
{
  MqReadL_END (msgque->read);
}

enum MqErrorE
pRead_RET_START (
  register struct MqReadS * const read
)
{
  struct MqBufferS * buf;
  struct ReadSaveS * save;
  register struct MqBufferS * const bdy = read->bdy;

  MqErrorCheck (MqReadU (read, &buf));

  // check buffer type
  if (buf->type != MQ_RETT)
    return MqErrorDbV(MQ_ERROR_TYPE, MqLogTypeName(MQ_RETT), MqLogTypeName(buf->type));

  save = (struct ReadSaveS *) pCachePop (read->saveCache);
  
//MqDLog(msgque,0,"read<%p>, save<%p>\n", read, save);

  save->save = read->save;
  save->bdy  = *bdy;
  read->save = save;

  // get -> returnNum
  if (likely (read->returnCode == MQ_RETURN_OK)) {
    read->returnNum = -1;
  } else {
    read->returnNum = (read->type==MQ_BINT ? MqBufU2INT(buf->cur) : str2int(buf->cur.C,NULL,16));
    buf->cur.B += (HDR_INT_LEN + 1);
    buf->cursize -= (HDR_INT_LEN + 1);
  }

  // 3. fix ret
  bdy->type = read->type;
  bdy->data = bdy->cur.B = buf->cur.B;
  bdy->size = buf->size;
  bdy->cursize = buf->cursize;
  bdy->numItems = (read->type==MQ_BINT ? MqBufU2INT(bdy->cur) : str2int(bdy->cur.C,NULL,16));
  bdy->cur.B += (HDR_INT_LEN + 1);
  
  return MQ_OK;

error:
  return MqErrorStack(MQ_ERROR_S);
}

// \attention "pRead_RET_END" is a #define

/*****************************************************************************/
/*                                                                           */
/*                                read_native                                */
/*                                                                           */
/*****************************************************************************/

static enum MqErrorE
sReadFIX (
  struct MqReadS const * const read,
  register struct MqBufferS * const buf,
  MQ_STRB const * const prefix
)
{
  // 1. read
  MqErrorSwitch (pIoRead (MSGQUE->io, buf, buf->cursize));

  // 2. read 'HDR'
#ifdef MQ_SAVE
  if (unlikely (strncmp (buf->cur.C, prefix, 3))) {
    MqErrorV (MQ_ERROR_S, __func__, "expect '%3s' but got '%s'", prefix, buf->cur.C);
    goto error;
  }
#endif
  buf->cur.B += HDR_CtxId_S;

  return MQ_OK;

error:
  return MqErrorStack (MQ_ERROR_S);
}

enum MqErrorE
pReadHDR (
  register struct MqReadS * read,
  struct MqS ** a_msgque
)
{
  register struct MqS * msgque = read->msgque;
  MQ_SIZE ctxId;
  MQ_SIZE size;
  MQ_HDL trans;
  register int const string = msgque->config.isString;
  register struct HdrS * cur;
  int debug;
  register struct MqBufferS * bdy;
  register struct MqTokenS * srvT;

  // 1. preset msgque
  *a_msgque = msgque;

  // 2. read
  MqErrorSwitch (pIoRead (msgque->io, read->hdr, sizeof(struct HdrS)));
  cur = (struct HdrS *) read->hdr->cur.B;

//MqDLogV (msgque, __func__, 0, "START (%c%c%c%c)\n", cur->tok[0], cur->tok[1], cur->tok[2], cur->tok[3]);

  // check "HDR"
  if (cur->ID1 != 'H' || cur->ID2 != 'D' || cur->ID3 != 'R') {
    MQ_BUF save = MqBufferCreate (1000);
    do {
      MqBufferAppendC (save, (MQ_STR) read->hdr->data);
    } while (pIoRead (msgque->io, read->hdr, sizeof(struct HdrS)) == MQ_OK);
    *(read->hdr->data+read->hdr->cursize) = '\0';
    MqBufferAppendC (save, (MQ_STR) read->hdr->data);
    MqErrorDb (MQ_ERROR_INVALID_HEADER);
    MqErrorSAppendV (MQ_ERROR_S, "but got: %s", (MQ_STR) save->data);
    MqBufferDelete (&save);
    return MQ_ERROR;
  }

  // 3. read binary data
  if (unlikely(string)) {
    ctxId  = str2int(cur->ctxId.S,NULL,16);
    size   = str2int(cur->size.S,NULL,16);
  } else {
    ctxId  = cur->ctxId.B;
    size   = cur->size.B;
    // 3.b fix byte-order
    if (msgque->endian) {
      pSwap4 (&ctxId);
      pSwap4 (&size);
    }
  }

  trans = cur->trans;

  // 4. set my msgque
  if (ctxId != msgque->ctxId) {
    // check if ctxId is out of range
    if (ctxId > msgque->ctxIdP->ctxIdR) 
	return MqErrorV (MQ_ERROR_S, __func__, -1, 
	    "invalid context-id '%i'", ctxId);
    // do the context-switch if the message does !!not!! belongs to the
    // original msgque
    msgque = msgque->ctxIdP->ctxIdA[ctxId];
    // check if entry was allready deleted
    if (!msgque)
	return MqErrorV (MQ_ERROR_S, __func__, -1, 
	    "context with id '%i' was allready deleted !!", ctxId);
    *a_msgque = msgque;
  }
//MqDLogV (msgque, __func__, 0, "cur->ctxId.B<%i>, msgque->ctxId<%i>\n", cur->ctxId.B, msgque->ctxId);
//MqDLogV (msgque, __func__, 0, "hex\n<%s>\n", pLogHEX (read->hdr->data, sizeof (struct HdrS)));
  debug = msgque->config.debug;
  msgque->_trans = trans;

  // 6. log message
  if (unlikely (debug >= 5)) {
    read->hdr->type = MQ_STRING_TYPE(string);
    pLogHDR (msgque, __func__, 5, read->hdr);
  }

  // 7. setup read
  read = msgque->read;
  read->returnCode = (enum MqReturnE) cur->code;
  read->returnNum = 0;
  bdy = read->bdy;

  // 8. read token
  srvT = msgque->srvT;
  cur->charT = '\0';
  pTokenSetCurrent (srvT, cur->tok);

  // 9. read BDY
  if ((bdy->cursize = size)) {
    bdy->type = MQ_STRING_TYPE(string);
    if (unlikely (debug >= 6))
      MqDLogV (msgque, __func__, 6, "BDY -> " MQ_FORMAT_Z " bytes\n", size);

    // 1. read
    MqErrorCheck (sReadFIX (read, bdy, "BDY"));

    // 2. read NumItems
    if (unlikely(string)) {
      bdy->numItems = str2int(bdy->cur.C,NULL,16);
    } else {
      if (msgque->endian) {
	pSwapBDY(bdy->data);
      }
      bdy->numItems = MqBufU2INT(bdy->cur);
    }

    bdy->cur.B = (bdy->data + BDY_SIZE + 1);    // BUFFER_CHAR

    if (unlikely (debug >= 7 && size > BDY_SIZE))
      pLogBDY (msgque, __func__, 7, bdy);
  }

  // reset data
  read->canUndo = MQ_NO;
  read->save = NULL;
  
  // 11. check if it's a System-pToken
  // this line have to use the 'return' because pTokenCheckSystem usually return
  // MQ_CONTINUE if a system token (_*) was found
  return pTokenCheckSystem (srvT);

error:
  return MqErrorStack (msgque->error);
}

#define pReadWordM(error, buf, out) { \
  register union MqBufferU bcur = buf->cur; \
	MQ_SIZE lsize; \
  /* check if an argument is available? and decrease the number of arguments */ \
  /* buf->numItems have to be an \b int !! because --0 have to be < 0 */ \
  if ((--buf->numItems) < 0) { \
    return MqErrorDb2 (error,MQ_ERROR_REQUEST_ARGUMENTS); \
  } \
 \
  /* read "type" of the value */ \
  out->type = (enum MqTypeE) *bcur.B; \
  bcur.B += 2; \
 \
  /* read "length" of the value */ \
  lsize = (out->type & MQ_TYPE_IS_NATIVE); \
  /* if type is native than "lsize" has the length of the native type */ \
  /* if type is *not* native than "lsize" has the value of '0' */ \
  if (lsize == 0) { \
    lsize = (buf->type == MQ_BINT ? MqBufU2INT(bcur) : str2int(bcur.C, NULL, 16)); \
    bcur.B += (HDR_INT_LEN + 1); \
  } \
 \
  /* read argument */ \
  out->data = out->cur.B = bcur.B; \
 \
  bcur.B += (out->cursize = out->size = lsize); \
  *bcur.B++ = '\0'; \
  buf->cur = bcur; \
}

enum MqErrorE
pReadWord (
  struct MqErrorS * const error,
  struct MqBufferS * const buf,
  register struct MqBufferS * const out
)
{
  pReadWordM(error, buf, out);
  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                                read_atom                                  */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
MqReadL (
  register struct MqReadS * const read,
  struct MqBufferLS ** const out
)
{
  struct MqBufferLS * const line = MqBufferLCreate (read->bdy->numItems);
  struct MqBufferS * buf;
  while (read->bdy->numItems) {
    MqErrorCheck (MqReadU (read, &buf));
    MqBufferLAppend (line, MqBufferDup(buf), -1);
  }
  *out = line;
  read->canUndo = MQ_NO;
  return MQ_OK;
error:
  return MqErrorStack (MQ_ERROR_S);
}

static enum MqErrorE
sReadA1 (
  struct MqReadS * const read,
  union MqBufferAtomU * const out,
  enum MqTypeE const type
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  struct MqBufferS * const buf = read->bdy;
  struct MqBufferS * const cur = read->cur;
  pReadWordM(error, buf, cur);
  //MqErrorCheck( pReadWord (error, buf, cur));
  MqErrorCheck( pBufferGetA1 (error, cur, out, type));
  read->canUndo = MQ_YES;
  return MQ_OK;

error:
  return MqErrorStack (error);
}

static enum MqErrorE
sReadA2 (
  struct MqReadS * const read,
  union MqBufferAtomU * const out,
  enum MqTypeE const type
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  struct MqBufferS * const buf = read->bdy;
  struct MqBufferS * const cur = read->cur;
  pReadWordM(error, buf, cur);
  //MqErrorCheck( pReadWord (error, buf, cur));
  MqErrorCheck( pBufferGetA2 (error, cur, out, type));
  read->canUndo = MQ_YES;
  return MQ_OK;

error:
  return MqErrorStack (error);
}

static enum MqErrorE
sReadA4 (
  struct MqReadS * const read,
  union MqBufferAtomU * const out,
  enum MqTypeE const type
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  struct MqBufferS * const buf = read->bdy;
  struct MqBufferS * const cur = read->cur;
  pReadWordM(error, buf, cur);
  //MqErrorCheck( pReadWord (error, buf, cur));
  MqErrorCheck( pBufferGetA4 (error, cur, out, type));
  read->canUndo = MQ_YES;
  return MQ_OK;

error:
  return MqErrorStack (error);
}

static enum MqErrorE
sReadA8 (
  struct MqReadS * const read,
  union MqBufferAtomU * const out,
  enum MqTypeE const type
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  struct MqBufferS * const buf = read->bdy;
  struct MqBufferS * const cur = read->cur;
  pReadWordM(error, buf, cur);
  //MqErrorCheck( pReadWord (error, buf, cur));
  MqErrorCheck( pBufferGetA8 (error, cur, out, type));
  read->canUndo = MQ_YES;
  return MQ_OK;

error:
  return MqErrorStack (error);
}

/*****************************************************************************/
/*                                                                           */
/*                               read functions                              */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
MqReadY (
  struct MqReadS * const read,
  MQ_BYT * const valP
)
{
  return sReadA1(read, (union MqBufferAtomU * const) valP, MQ_BYTT);
}

MQ_INLINE enum MqErrorE
MqReadO (
  struct MqReadS * const read,
  MQ_BOL * const valP
)
{
  return sReadA1(read, (union MqBufferAtomU * const) valP, MQ_BOLT);
}

MQ_INLINE enum MqErrorE
MqReadS (
  struct MqReadS * const read,
  MQ_SRT * const valP
)
{
  return sReadA2(read, (union MqBufferAtomU * const) valP, MQ_SRTT);
}

MQ_INLINE enum MqErrorE
MqReadI (
  struct MqReadS * const read,
  MQ_INT * const valP
)
{
  return sReadA4(read, (union MqBufferAtomU * const) valP, MQ_INTT);
}

MQ_INLINE enum MqErrorE
MqReadF (
  struct MqReadS * const read,
  MQ_FLT * const valP
)
{
  return sReadA4(read, (union MqBufferAtomU * const) valP, MQ_FLTT);
}

MQ_INLINE enum MqErrorE
MqReadW (
  struct MqReadS * const read,
  MQ_WID * const valP
)
{
  return sReadA8(read, (union MqBufferAtomU * const) valP, MQ_WIDT);
}

MQ_INLINE enum MqErrorE
MqReadD (
  struct MqReadS * const read,
  MQ_DBL * const valP
)
{
  return sReadA8(read, (union MqBufferAtomU * const) valP, MQ_DBLT);
}

enum MqErrorE
MqReadC (
  struct MqReadS * const read,
  MQ_STR * const out
)
{
  struct MqErrorS * const error = MQ_ERROR_S;
  struct MqBufferS * const buf = read->bdy;
  struct MqBufferS * const cur = read->cur;
  pReadWordM(error, buf, cur);
  MqErrorCheck (MqBufferGetC (error, cur, out));
  read->canUndo = MQ_YES;
  return MQ_OK;

error:
  return MqErrorStack (error);
}

enum MqErrorE
MqReadB (
  struct MqReadS * const read,
  MQ_BIN * const out,
  MQ_SIZE * const len
)
{
  struct MqBufferS * const buf = read->bdy;
  struct MqBufferS * const cur = read->cur;
  pReadWordM(MQ_ERROR_S, buf, cur);
  *out = cur->data;
  *len = cur->cursize;
  read->canUndo = MQ_YES;
  return MQ_OK;
}

enum MqErrorE
MqReadU (
  struct MqReadS * const read,
  MQ_BUF * const out
)
{
  register struct MqBufferS * const buf = read->bdy;
  register struct MqBufferS * const cur = read->cur;
  pReadWordM(MQ_ERROR_S, buf, cur);
  *out = cur;
  read->canUndo = MQ_YES;
  return MQ_OK;
}

/*****************************************************************************/
/*                                                                           */
/*                              readM functions                              */
/*                                                                           */
/*****************************************************************************/

MQ_INLINE enum MqErrorE
MqReadYM (
  struct MqS const * const msgque,
  MQ_BYT * const valP
)
{
  return sReadA1(msgque->read, (union MqBufferAtomU * const) valP, MQ_BYTT);
}

MQ_INLINE enum MqErrorE
MqReadOM (
  struct MqS const * const msgque,
  MQ_BOL * const valP
)
{
  return sReadA1(msgque->read, (union MqBufferAtomU * const) valP, MQ_BOLT);
}

MQ_INLINE enum MqErrorE
MqReadSM (
  struct MqS const * const msgque,
  MQ_SRT * const valP
)
{
  return sReadA2(msgque->read, (union MqBufferAtomU * const) valP, MQ_SRTT);
}

MQ_INLINE enum MqErrorE
MqReadIM (
  struct MqS const * const msgque,
  MQ_INT * const valP
)
{
  return sReadA4(msgque->read, (union MqBufferAtomU * const) valP, MQ_INTT);
}

MQ_INLINE enum MqErrorE
MqReadFM (
  struct MqS const * const msgque,
  MQ_FLT * const valP
)
{
  return sReadA4(msgque->read, (union MqBufferAtomU * const) valP, MQ_FLTT);
}

MQ_INLINE enum MqErrorE
MqReadWM (
  struct MqS const * const msgque,
  MQ_WID * const valP
)
{
  return sReadA8(msgque->read, (union MqBufferAtomU * const) valP, MQ_WIDT);
}

MQ_INLINE enum MqErrorE
MqReadDM (
  struct MqS const * const msgque,
  MQ_DBL * const valP
)
{
  return sReadA8(msgque->read, (union MqBufferAtomU * const) valP, MQ_DBLT);
}

MQ_INLINE enum MqErrorE
MqReadCM (
  struct MqS const * const msgque,
  MQ_STR * const out
)
{
  return MqReadC (msgque->read, out);
}

MQ_INLINE enum MqErrorE
MqReadBM (
  struct MqS const * const msgque,
  MQ_BIN * const out,
  MQ_SIZE * const len
)
{
  return MqReadB (msgque->read, out, len);
}

MQ_INLINE enum MqErrorE
MqReadUM (
  struct MqS const * const msgque,
  MQ_BUF * const out
)
{
  return MqReadU (msgque->read, out);
}

/*****************************************************************************/
/*                                                                           */
/*                                 read.high                                 */
/*                                                                           */
/*****************************************************************************/

enum MqErrorE
MqReadUndo (
  struct MqReadS * const read
)
{
  struct MqBufferS * const bdy = read->bdy;
  struct MqBufferS * const cur = read->cur;

  if (read->canUndo == MQ_NO) {
    return MqErrorDb(MQ_ERROR_UNDO_FORBIDDEN);
  }
  read->canUndo = MQ_NO;
  
  bdy->numItems++; 
  bdy->cur.B -= ( 2 + cur->cursize + 1 +
    (cur->type & MQ_TYPE_IS_NATIVE ? 0 : HDR_INT_LEN + 1)
  );
  return MQ_OK;
}

MQ_INLINE enum MqErrorE
MqReadUndoM (
  struct MqS const * const msgque
)
{
  return MqReadUndo(msgque->read);
}

void
pReadSetType(
  struct MqReadS * const read,
  MQ_BOL const type
)
{
  read->type = MQ_STRING_TYPE(type);
}

MQ_INLINE MQ_SIZE
MqReadGetNumItems (
  struct MqReadS const * const read
)
{
  return read->bdy->numItems;
}

MQ_INLINE MQ_SIZE
MqReadGetNumItemsM (
  struct MqS const * const msgque
)
{
  return msgque->read->bdy->numItems;
}

MQ_INLINE MQ_BOL
MqReadItemExists (
  struct MqReadS const * const read
)
{
  return (read->bdy->numItems != 0 ? MQ_YES : MQ_NO);
}

MQ_INLINE MQ_BOL
MqReadItemExistsM (
  struct MqS const * const msgque
)
{
  return (msgque->read->bdy->numItems != 0 ? MQ_YES : MQ_NO);
}

MQ_INLINE enum MqReturnE
pReadGetReturnCode (
  struct MqReadS const * const read
)
{
  return read->returnCode;
}

MQ_INLINE MQ_INT
pReadGetReturnNum (
  struct MqReadS const * const read
)
{
  return read->returnNum;
}

MQ_INLINE struct MqS *
MqReadGetMsgque (
  struct MqReadS const * const read
)
{
  return read->msgque;
}

/*****************************************************************************/
/*                                                                           */
/*                                 debugging                                 */
/*                                                                           */
/*****************************************************************************/

#ifdef _DEBUG

void
pReadLog (
  MQ_STRB const * const prefix,
  register struct MqReadS const * const read
)
{
  struct MqS * msgque = read->msgque;

  MqDLogV (msgque, prefix, 0, ">>>> MqReadS (%p)\n", read);
  MqDLogV (msgque, prefix, 0, "msgque     = " MQ_FORMAT_P "\n", read->msgque);
  MqDLogV (msgque, prefix, 0, "returnCode = " "%c" "\n", read->returnCode);
  MqDLogV (msgque, prefix, 0, "returnNum  = " MQ_FORMAT_I "\n", read->returnNum);
  MqDLogV (msgque, prefix, 0, "  >>>> read->hdr\n");
  MqBufferLog (msgque, read->hdr, prefix);
  MqDLogV (msgque, prefix, 0, "  >>>> read->bdy\n");
  MqBufferLog (msgque, read->bdy, prefix);
  MqDLogV (msgque, prefix, 0, "  >>>> read->cur\n");
  MqBufferLog (msgque, read->cur, prefix);
  MqDLogV (msgque, prefix, 0, "<<<< MqReadS\n");
}

#endif



