/**
 *  \file       tclmsgque/context_tcl.c
 *  \brief      \$Id: context_tcl.c 1187 2009-05-19 07:56:13Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1187 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

#include "msgque_tcl.h"

#define MSGQUE msgque

typedef int (
  *LookupKeywordF
) (
  Tcl_Interp  *		      interp,
  struct TclContextS * const  context,
  int			      skip,
  int			      objc,
  struct Tcl_Obj *const *     objv
);

struct LookupKeyword {
  const char	  *key;
  LookupKeywordF  keyF;
};

/*****************************************************************************/
/*                                                                           */
/*                              context_init                                 */
/*                                                                           */
/*****************************************************************************/

static int NS(ContextCmd) ( ClientData, Tcl_Interp*, int, Tcl_Obj * const []);
static void NS(DeleteCmd) ( ClientData );

/*****************************************************************************/
/*                                                                           */
/*                                 helper                                    */
/*                                                                           */
/*****************************************************************************/

/// \brief link \e Tcl event-queue with \libmsgque event-queue
///
/// \libmsgque using the \e -fevent option to get a 'C' procedure for calling an
/// external event-queue.
///
/// \param[in] context object to work on
///
static void
NS(EventLink) (
  struct MqContextS * const context
)
{
  Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT);
}

/// \brief helper: used to simplify the code
#define CheckForAdditionalArg(txt) \
    objc--; objv++; \
    if (objc<1) { \
	Tcl_SetResult(interp, "no additional argument available for '" #txt "'", TCL_STATIC); \
	return TCL_ERROR; \
    }

/*****************************************************************************/
/*                                                                           */
/*                                 subcommands                               */
/*                                                                           */
/*****************************************************************************/

static
HD(RenameTo)
{
  Tcl_Obj * lobjv[3], *new;
  int ret;

  CHECK_OBJ(new)
  CHECK_NOARGS

  lobjv[0] = Tcl_NewStringObj("rename",-1);
  lobjv[1] = context->self;
  lobjv[2] = new;

  Tcl_IncrRefCount(new);
  Tcl_IncrRefCount(lobjv[0]);
  ret = Tcl_EvalObjv (interp, 3, lobjv, TCL_EVAL_GLOBAL);
  Tcl_DecrRefCount(lobjv[0]);
  TclErrorCheck(ret);
  Tcl_DecrRefCount(context->self);

  context->self = new;
  RETURN_TCL
}

static
HD(Delete)
{
  CHECK_NOARGS
  NS(DeleteCmd)(context);
  RETURN_TCL
}

static
HD(ProcessEvent)
{
  SETUP_msgque
  int timeout = -1;
  int FLAGS = MQ_WAIT_NO;
  int iA;
  static const char *optA[] = { "-timeout", "-wait", NULL };
  static const char *optB[] = { "NO", "ONCE", "FOREVER", NULL };
  enum optE { TIMEOUT, WAIT };

  // look for options
  objc -= skip;
  objv += skip;
  while (objc) {
    TclErrorCheck (Tcl_GetIndexFromObj (interp, objv[0], optA, "option", 0, &iA));
    switch ((enum optE) iA) {
      case WAIT:
        CheckForAdditionalArg (-wait);
	TclErrorCheck (Tcl_GetIndexFromObj (interp, objv[0], optB, "-wait", 0, &FLAGS));
	objv++;objc--;
	break;
      case TIMEOUT:
        CheckForAdditionalArg (-timeout);
        TclErrorCheck (Tcl_GetIntFromObj (interp, objv[0], &timeout));
	objv++;objc--;
        break;
    }
  }

  ErrorMqToTclWithCheck (MqProcessEvent (msgque, timeout, FLAGS));
  RETURN_TCL
}

static
HD(ErrorC)
{
  MQ_INT errnum;
  MQ_STR prefix, error_message;

  SETUP_msgque
  CHECK_C(prefix)
  CHECK_I(errnum)
  CHECK_C(error_message)
  CHECK_NOARGS
  MqErrorC(msgque->error, prefix, errnum, error_message);
  ErrorMqToTcl();
error:
  return TCL_ERROR;
}

/*****************************************************************************/
/*                                                                           */
/*                              context_basic                                */
/*                                                                           */
/*****************************************************************************/

/** \brief delete a child-context
 *
 *  \param[in,out] msgqueP the context to delete
 **/ 
static void
NS(ChildDelete) (
  struct MqS ** msgqueP
)
{
  struct TclContextS *context;

  // return if the object was allread deleted
  if (!msgqueP && !(*msgqueP)) return;

  context = (struct TclContextS *) (*msgqueP)->config.myCtx;

  if (context != NULL) {
    // the onDelete is necessary to avoid the double freeing of memory.
    // if a parent is deleted all childs are deleted as well.
    // during child delete the command-object is deleted (Tcl_DeleteCommandFromToken).
    // if the command-object is deleted the "NS(ChildDelete)" is called again -> double free
    if (context->onDelete == MQ_YES) return;
    context->onDelete = MQ_YES;

    if (context->configProc != NULL) Tcl_DecrRefCount (context->configProc);
    if (context->self != NULL) Tcl_DecrRefCount (context->self);
    if (context->command != NULL) Tcl_DeleteCommandFromToken (context->interp, context->command);
  }

  MqDelete (msgqueP);
  MqSysFree (context);
}

/** \brief delete a parent-context in server mode
 *
 *  \param[in,out] msgqueP the context to delete
 **/ 
static void
NS(ParentDelete) (
  struct MqS ** msgqueP
)
{
  struct TclContextS *context;
  struct MqS *msgque;

  // return if the object was allread deleted
  if (!msgqueP && !(*msgqueP)) return;
  msgque = *msgqueP;

  context = (struct TclContextS *) msgque->config.myCtx;

  if (context != NULL) {
    // delete Exit-Handler
    Tcl_DeleteExitHandler(NS(DeleteCmd), context);

    // the onDelete is necessary to avoid the double freeing of memory.
    // if a parent is deleted all childs are deleted as well.
    // during child delete the command-object is deleted (Tcl_DeleteCommandFromToken).
    // if the command-object is deleted the "ContextDelete" is called again -> double free
    if (context->onDelete == MQ_YES) return;

    // cleanup the event-source
    //Tcl_DeleteEvents (tclEventDelete, context);
    //Tcl_DeleteEventSource (tclEventSetup, tclEventCheck, context);
  }

  // finally use child-delete to clean the context
  NS(ChildDelete)(msgqueP);
}

/** \brief delete a parent process or thread
 *
 *  \param[in] isThread bool value set to \c 1 (yes) if it is a thread and \c 0 (no) if not
 *  \param[in] num the exit number
 *
 *  \e libmsgque has the \e thread created -> \e libmsgque
 *  have to delete the thread -> no \e thread action in this handler
 *
 **/ 
static void
NS(ParentExit) (
  int isThread,
  int num
)
{
  // Tcl_Finalize is calling Tcl_FinalizeThread too
  if (isThread) {
    Tcl_FinalizeThread();
  } else {
    Tcl_Finalize();
  }
}

/** \brief delete a \e ContextS object (called by "rename $ctx {}")
 *
 *  \param[in] clientData command specific data, \e ContextS in this case
 **/
static void
NS(DeleteCmd) (
  ClientData clientData
)
{
  struct TclContextS * context;
  struct MqS * msgque;
  
  if (!clientData) return;

  context = (struct TclContextS *) clientData;
  if (context->onDelete == MQ_YES) return;
  msgque = context->msgque;

  // 1. print message
  MqDLog (msgque, 4, "DELETE context<%p>, msgque<%p>\n", context, msgque);

  // 2. cleanup
  MqDelete(&msgque);
}

/** \brief create a new child-context in server mode
 *
 * \param[in,out] config configuration data
 * \param[in] argv server specific configuration data
 * \param[in] alfa this should allways be NULL
 * \param[out] msgqueP the new created #MqS object
 * \retMqErrorE
 **/
static enum MqErrorE
NS(ChildCreate) (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS ** msgqueP
) {
  // this is the default for:
  //  1. MQ_STATUS_IS_INITIAL
  //  2. MQ_STATUS_IS_DUP & MQ_STATUS_IS_FORK
  struct TclContextS * context = (struct TclContextS *) config->myCtx;
  struct MqS * msgque;
  struct MqBufferS *cmdname;

  // 1. setup the SERVER+CHILD context using PARENT data
  if (config->parent != NULL && context == NULL) {
    struct TclContextS * parentContext = (struct TclContextS *) config->parent->config.myCtx;
    context = MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (*context));
    config->myCtx = (struct MqContextS *) context;
    memcpy(context,parentContext,sizeof(*context));
    if (context->configProc) Tcl_IncrRefCount (context->configProc);
    context->self = NULL;	// the command belongs to the PARENT
    context->command = NULL;	// the command belongs to the PARENT
  }

  // start dummyTimer and activate the event-handling
  config->fEvent = NS(EventLink);

  // 2. create the Message-Queue
  MqErrorCheck (MqCreate (config, argv, alfa, msgqueP));
  msgque = *msgqueP;

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(msgque, &argv));

  if ((msgque->config.statusIs & MQ_STATUS_IS_FORK) && msgque->config.parent == NULL) {
    // A thread enabled tcl has an independent thread as "notifier"
    // forking a process does !not! create this thread again 
    // -> Tcl_InitNotifier start the notifier thread
    Tcl_InitNotifier();
  }

  // 3. create the Tcl command
  cmdname = msgque->temp;
  MqBufferSetV (cmdname, "<%s-%i|%p>", msgque->config.name, MqGetCtxId(msgque), msgque);
  context->command = Tcl_CreateObjCommand (context->interp, cmdname->cur.C, NS(ContextCmd),
                                           (ClientData) context, NS(DeleteCmd));
  context->self = Tcl_NewStringObj(cmdname->cur.C,-1);
  Tcl_IncrRefCount(context->self);

  // 4. if configProc is available (only in server mode) configure the server
  if (msgque->config.isServer == MQ_YES && context->configProc != NULL) {
    Tcl_DString cmd;
    Tcl_DStringInit(&cmd);
    Tcl_DStringAppendElement(&cmd,Tcl_GetString(context->configProc));
    Tcl_DStringAppendElement(&cmd,cmdname->cur.C);
    if (Tcl_EvalEx(context->interp, Tcl_DStringValue(&cmd), 
	Tcl_DStringLength(&cmd), TCL_EVAL_GLOBAL) != TCL_OK) {
      Tcl_DStringFree(&cmd);
      return MqErrorC(msgque->error,__func__,-1,Tcl_GetStringResult(context->interp));
    }
    Tcl_DStringFree(&cmd);
  }

  if (unlikely(msgque->config.debug >= 2) && unlikely(msgque->config.parent != NULL))
    MqDLogV(msgque,__func__,2,"CREATE context<%p>, msgque<%p>\n",context,msgque);

  Tcl_SetResult (context->interp,cmdname->cur.C, TCL_VOLATILE);

  // everything is OK
  return MQ_OK;

error:
  return MqErrorStack ((*msgqueP)->error);
}

/** \brief create a TCL parent context
 *
 */
static enum MqErrorE
NS(ParentCreate) (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS       ** msgqueP
)
{
  // this is the default for:
  //  1. MQ_STATUS_IS_INITIAL
  //  2. MQ_STATUS_IS_DUP & MQ_STATUS_IS_FORK
  struct TclContextS * context = (struct TclContextS *) config->myCtx;
  struct MqErrorS * error = MQ_ERROR_PRINT;
  struct MqS * msgque;

  // The thread get a new interp 
  if (config->statusIs & MQ_STATUS_IS_DUP && config->statusIs & MQ_STATUS_IS_THREAD) {
    Tcl_Obj *argvO;
    Tcl_Interp * interp;
    int i;

    // 0. prepare
    if (MqInitBuf == NULL || MqInitBuf->cursize < 2)
      return MqErrorC(error, __func__, 1, "unable to start thread, 'script' not available");

    // 1. get new interpreter
    interp = Tcl_CreateInterp();
    TclErrorCheck3 (Tcl_Init(interp));

    // 2. fill "argv" list
    argvO = Tcl_NewListObj(0, NULL);
    for (i=1; i<argv->cursize; i++) {
      MQ_STR str;
      if (MqBufferGetC(error, argv->data[i], &str) == MQ_ERROR) return MQ_ERROR;
      TclErrorCheck3 (Tcl_ListObjAppendElement(interp, argvO, Tcl_NewStringObj(str, -1)));
    }
    Tcl_SetVar2Ex(interp, "argv", NULL, argvO, TCL_GLOBAL_ONLY);

    // 3. cleanup argv, alfa
    MqBufferLDelete (&argv);
    MqBufferLDelete (&alfa);

    // 4. evaluate the MqMainScriptName -> this command will enter the event-loop
    TclErrorCheck3 (Tcl_EvalFile(interp, MqInitBuf->data[1]->cur.C));

    // 5. end interpreter
    Tcl_DeleteInterp(interp);

    return MQ_OK;
error1:
    MqErrorC(error,__func__,-1,Tcl_GetStringResult(interp));
    Tcl_DeleteInterp(interp);
    return MQ_ERROR;
  }

  // create the server-context
  MqErrorCheck (NS(ChildCreate)(config, argv, alfa, msgqueP));
  msgque = *msgqueP;

  // add exit handler
  Tcl_CreateExitHandler(NS(DeleteCmd), context);

  MqDLog (msgque, 2, "CREATE context<%p>, msgque<%p>\n", context, msgque);

  // everything is OK
  return MQ_OK;

error:
  return MqErrorStack((*msgqueP)->error);
}

/** \brief create a context 
 *
 *  \param[in] interp current Tcl interpreter
 *  \param[in] objc number of objects in \e objv
 *  \param[in] objv array of \e Tcl_Obj objects
 *  \return Tcl error-code
 **/
int
NS(Context_Init) (
  Tcl_Interp * interp,
  int objc,
  Tcl_Obj * const objv[]
)
{
  register struct TclContextS * context = MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (*context));
  struct MqConfigS * config = MqConfigInit();
  struct MqS * msgque;

  struct MqBufferLS * const argv = MqBufferLCreate (10);
  struct MqBufferLS * const alfa = MqBufferLCreate (10);
  MQ_STR val;
  int i,len;
  char valp2;

  // add name of the client as first argument
  MqBufferLAppendC (argv, (const MQ_STR) Tcl_GetNameOfExecutable());

  // 1. parse command-line
  for (i = 0; i < objc; i++) {
    val = Tcl_GetStringFromObj (objv[i], &len);

    // only search until MQ_ALFA because the following appends have to be
    // !!before!! the MQ_ALFA token
    if (*val == MQ_ALFA && len == 1) break;

    // now check for "known" arguments
    if (len >= 5 && *val == '-' && *(val + 1) == '-') {
	valp2 = *(val + 2);
	if (valp2 == 'p' && !strcmp (val, "--parent")) {
	  MQ_PTR parent = NULL;
	  struct MqS * parent_msgque = NULL;
	  if (++i>=objc || !Tcl_GetCommandFromObj (interp, objv[i])) {
	    Tcl_WrongNumArgs (interp, 0, NULL, "tclmsgque MqS --parent CONTEXT");
	    return TCL_ERROR;
	  }
	  TclErrorCheck (NS(GetClientData) (interp, objv[i], &parent));
	  parent_msgque = ((struct TclContextS *) parent)->msgque;
	  config = MqConfigDup (&parent_msgque->config);
	  MqConfigInitParent (config, parent_msgque);
	  continue;
	} else if (valp2 == 'm' && !strncmp (val, "--master", 8)) {
	  MQ_PTR master = NULL;
	  int master_id = 0;
	  struct MqS * master_msgque = NULL;
	  if (++i>=objc || !Tcl_GetCommandFromObj (interp, objv[i])) {
	    Tcl_WrongNumArgs (interp, 0, NULL, "tclmsgque MqS --master CONTEXT");
	    return TCL_ERROR;
	  }
	  TclErrorCheck (NS(GetClientData) (interp, objv[i], &master));
	  master_msgque = ((struct TclContextS *) master)->msgque;
	  config = MqConfigDup (&master_msgque->config);
	  // check if next item is "--master-id"
	  i++;
	  val = Tcl_GetStringFromObj (objv[i], &len);
	  if (strncmp (val, "--master-id", 11)) {
	    i--;
	  } else {
	    if (++i>=objc) {
	      Tcl_WrongNumArgs (interp, 0, NULL, "tclmsgque MqS --master CONTEXT --master-id NUM");
	      return TCL_ERROR;
	    }
	    TclErrorCheck(Tcl_GetIntFromObj(interp, objv[i], &master_id));
	  }
	  MqConfigInitMaster (config, master_msgque, master_id);
	  // use "--config" from master
	  context->configProc = ((struct TclContextS *) master)->configProc;
	  if (context->configProc != NULL) Tcl_IncrRefCount (context->configProc);
	  continue;
	} else if (valp2 == 's' && !strcmp (val, "--server")) {
	  config->isServer = MQ_YES;
	  continue;
	} else if (valp2 == 'c' && !strcmp (val, "--config")) {
	  //Tcl_DString cmd;
	  if (++i>=objc) {
	    Tcl_WrongNumArgs (interp, 0, NULL, "tclmsgque MqS --config tclProc");
	    return TCL_ERROR;
	  }
	  TclErrorCheck(NS(ProcCheck)(interp, objv[i], "tclmsgque MqS --config tclProc"));
	  context->configProc = objv[i];
	  Tcl_IncrRefCount (context->configProc);
	  continue;
	} else if (valp2 == 'f' && !strcmp (val, "--filter")) {
	  config->isFilter = MQ_YES;
	  continue;
	} else if (valp2 == 'F' && !strcmp (val, "--FTR")) {
	  if (++i>=objc) {
	    Tcl_WrongNumArgs (interp, 0, NULL, "tclmsgque MqS --FTR tclProc");
	    return TCL_ERROR;
	  }
	  TclErrorCheck(NS(ProcCheck)(interp, objv[i], "tclmsgque MqS --FTR tclProc"));
	  Tcl_IncrRefCount (objv[i]);
	  config->fFTR = NS(ProcCall);
	  config->FTRdata = objv[i];
	  config->FTRfree = NS(ProcFree);
	  continue;
	} else if (valp2 == 'E' && !strcmp (val, "--EOF")) {
	  if (++i>=objc) {
	    Tcl_WrongNumArgs (interp, 0, NULL, "tclmsgque MqS --EOF tclProc");
	    return TCL_ERROR;
	  }
	  TclErrorCheck(NS(ProcCheck)(interp, objv[i], "tclmsgque MqS --EOF tclProc"));
	  Tcl_IncrRefCount (objv[i]);
	  config->fEOF = NS(ProcCall);
	  config->EOFdata = objv[i];
	  config->EOFfree = NS(ProcFree);
	  continue;
	}
    }

    // 3. add other items to the argument \e MqBufferLS
    MqBufferLAppendC (argv, val);
  }

  // add configoration dependencies
  if (Tcl_GetVar2Ex(interp,"tcl_platform","threaded",TCL_GLOBAL_ONLY) != NULL) {
    config->ignoreFork = MQ_YES;
  }

  // 5. Finish the appends to alfa, everything after first MQ_ALFA
  for (i++; i < objc; i++) {
    MqBufferLAppendC (alfa, Tcl_GetString (objv[i]));
  }

  // 6. set additional data
  context->interp = interp;
  config->fChildCreate = NS(ChildCreate);
  config->fChildDelete = NS(ChildDelete);
  config->fParentCreate = NS(ParentCreate);
  config->fParentDelete = NS(ParentDelete);
  config->fParentExit = NS(ParentExit);
  config->myCtx = (struct MqContextS *) context;

  // 7. create Context
  if (config->parent == NULL) {
    MqErrorCheck(NS(ParentCreate)(config, argv, alfa, &msgque));
  } else {
    MqErrorCheck(NS(ChildCreate)(config, argv, alfa, &msgque));
  }

  return TCL_OK;

error:
  if (msgque->config.isServer == MQ_NO) {
    // set error-message
    ErrorMqToTcl();
    // allways delete the msgque object because this is the last chance to do it
    // !!attention!! never use "?Parentdelete" ot "?ChildeDelete" directly
    // "MqDelete" will call the functions from above
    MqDelete(&msgque);
  } else {
    MqExit(msgque);
  }
  return TCL_ERROR;
}

/*****************************************************************************/
/*                                                                           */
/*                                 public                                    */
/*                                                                           */
/*****************************************************************************/

// public interfaces

HD(ReadY);
HD(ReadO);
HD(ReadS);
HD(ReadI);
HD(ReadF);
HD(ReadW);
HD(ReadD);
HD(ReadC);
HD(ReadB);
HD(ReadU);
HD(ReadL_START);
HD(ReadL_END);
HD(ReadGetNumItems);
HD(ReadItemExists);
HD(ReadUndo);
HD(ReadALL);

HD(SendSTART);
HD(SendEND);
HD(SendEND_AND_WAIT);
HD(SendEND_AND_CALLBACK);
HD(SendRETURN);
HD(SendFTR);
HD(SendEOF);
HD(SendY);
HD(SendO);
HD(SendS);
HD(SendI);
HD(SendF);
HD(SendW);
HD(SendD);
HD(SendC);
HD(SendB);
HD(SendU);
HD(SendL_START);
HD(SendL_END);

HD(GetY);
HD(GetO);
HD(GetS);
HD(GetI);
HD(GetF);
HD(GetW);
HD(GetD);
HD(GetC);
HD(GetB);
HD(GetL_START);
HD(GetL_END);
HD(GetType);

HD(ConfigGetIsString);
HD(ConfigGetIsSilent);
HD(ConfigGetIsServer);
HD(ConfigGetIsParent);
HD(ConfigGetIsSlave);
HD(ConfigGetDebug);
HD(ConfigGetToken);
HD(ConfigGetCtxId);
HD(ConfigGetName);
HD(ConfigGetParent);
HD(ConfigGetMaster);

HD(ServiceCreate);
HD(ServiceDelete);

HD(SlaveCreate);
HD(SlaveDelete);
HD(SlaveGet);

/** \brief create the \b $CTX tcl command
 *
 *  \tclmsgque
 *
 * \param[in] clientData command specific data, \e ContextS in this case
 * \param[in] interp the current Tcl interpreter
 * \param[in] objc number of items in the \e objv array
 * \param[in] objv array with arguments of the \c $CTX 'Tcl' command
 * \return Tcl error-code
 **/
static int
NS(ContextCmd) (
  ClientData clientData,
  Tcl_Interp * interp,
  int objc,
  Tcl_Obj * const objv[]
)
{
  int index;
  struct TclContextS * context = (struct TclContextS *) clientData;

  struct LookupKeyword keys[] = {

// READ

    { "ReadY",		      NS(ReadY)			},
    { "ReadO",		      NS(ReadO)			},
    { "ReadS",		      NS(ReadS)			},
    { "ReadI",		      NS(ReadI)			},
    { "ReadF",		      NS(ReadF)			},
    { "ReadW",		      NS(ReadW)			},
    { "ReadD",		      NS(ReadD)			},
    { "ReadC",		      NS(ReadC)			},
    { "ReadB",		      NS(ReadB)			},
    { "ReadU",		      NS(ReadU)			},
    { "ReadL_START",	      NS(ReadL_START)		},
    { "ReadL_END",	      NS(ReadL_END)		},
    { "ReadGetNumItems",      NS(ReadGetNumItems)	},
    { "ReadItemExists",	      NS(ReadItemExists)	},
    { "ReadUndo",	      NS(ReadUndo)		},
    { "ReadAll",	      NS(ReadALL)		},

// SEND

    { "SendSTART",	      NS(SendSTART)		},
    { "SendEND",	      NS(SendEND)		},
    { "SendEND_AND_WAIT",     NS(SendEND_AND_WAIT)	},
    { "SendEND_AND_CALLBACK", NS(SendEND_AND_CALLBACK)	},
    { "SendRETURN",	      NS(SendRETURN)		},
    { "SendFTR",	      NS(SendFTR)		},
    { "SendEOF",	      NS(SendEOF)		},
    { "SendY",		      NS(SendY)			},
    { "SendO",		      NS(SendO)			},
    { "SendS",		      NS(SendS)			},
    { "SendI",		      NS(SendI)			},
    { "SendF",		      NS(SendF)			},
    { "SendW",		      NS(SendW)			},
    { "SendD",		      NS(SendD)			},
    { "SendC",		      NS(SendC)			},
    { "SendB",		      NS(SendB)			},
    { "SendU",		      NS(SendU)			},
    { "SendL_START",	      NS(SendL_START)		},
    { "SendL_END",	      NS(SendL_END)		},

// GET

    { "GetY",		      NS(GetY)			},
    { "GetO",		      NS(GetO)			},
    { "GetS",		      NS(GetS)			},
    { "GetI",		      NS(GetI)			},
    { "GetF",		      NS(GetF)			},
    { "GetW",		      NS(GetW)			},
    { "GetD",		      NS(GetD)			},
    { "GetC",		      NS(GetC)			},
    { "GetB",		      NS(GetB)			},
    { "GetL_START",	      NS(GetL_START)		},
    { "GetL_END",	      NS(GetL_END)		},
    { "GetType",	      NS(GetType)		},

// CONFIG

    { "ConfigGetIsString",    NS(ConfigGetIsString)	},
    { "ConfigGetIsSilent",    NS(ConfigGetIsSilent)	},
    { "ConfigGetIsServer",    NS(ConfigGetIsServer)	},
    { "ConfigGetIsParent",    NS(ConfigGetIsParent)	},
    { "ConfigGetIsSlave",     NS(ConfigGetIsSlave)	},
    { "ConfigGetDebug",	      NS(ConfigGetDebug)	},
    { "ConfigGetToken",	      NS(ConfigGetToken)	},
    { "ConfigGetCtxId",	      NS(ConfigGetCtxId)	},
    { "ConfigGetName",	      NS(ConfigGetName)		},
    { "ConfigGetParent",      NS(ConfigGetParent)	},
    { "ConfigGetMaster",      NS(ConfigGetMaster)	},

// SERVICE

    { "ServiceCreate",	      NS(ServiceCreate)		},
    { "ServiceDelete",	      NS(ServiceDelete)		},

// CONTEXT

    { "ProcessEvent",	      NS(ProcessEvent)		},
    { "RenameTo",	      NS(RenameTo)		},
    { "Delete",		      NS(Delete)		},
    { "ErrorC",		      NS(ErrorC)		},

// SLAVE

    { "SlaveCreate",	      NS(SlaveCreate)		},
    { "SlaveDelete",	      NS(SlaveDelete)		},
    { "SlaveGet",	      NS(SlaveGet)		},

    { NULL,		      NULL			}
  };

  // read the index
  if (objc < 2) {
    Tcl_WrongNumArgs (interp, 1, objv, "subcommand ...");
    return TCL_ERROR;
  }

  TclErrorCheck (Tcl_GetIndexFromObjStruct (interp, objv[1], &keys, 
      sizeof(struct LookupKeyword), "subcommand", 0, &index));

  return (*keys[index].keyF) (context->interp, context, 2, objc, objv);
}

