/**
 *  \file       tclmsgque/misc_tcl.c
 *  \brief      \$Id: misc_tcl.c 1178 2009-05-15 07:25:57Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1178 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

/** \defgroup io Misc Helper Commands
 *  \ingroup tcl
 *  \{
 *  \brief used for reading and sending \e Tcl_Obj data items.
 */

#include "msgque_tcl.h"
#include <math.h>

#define MSGQUE msgque

/** \brief helper to get the \c ClientData from a tcl command
 *
 * \param[in] interp the current Tcl interpreter
 * \param[in] obj the Tcl command to get the \c ClientData from
 * \param[out] out store the \c ClientData
 * \return Tcl error-code
 */
int
NS(GetClientData) (
  Tcl_Interp * interp,
  Tcl_Obj * obj,
  MQ_PTR *out
)
{
  Tcl_CmdInfo infoPtr;
  Tcl_Command command;

  command = Tcl_GetCommandFromObj (interp, obj);
  if (command == NULL)
    return TCL_ERROR;
  Tcl_GetCommandInfoFromToken (command, &infoPtr);
  *out = infoPtr.objClientData;
  return TCL_OK;
}

void NS(pErrorFromMq) (
  Tcl_Interp *	      interp,
  const char *	      func,
  const char *	      file,
  const int	      line,
  struct MqS * const  msgque
)
{
  SETUP_context;
  if (context->callLevel != 0) {
    Tcl_SetResult(interp, "MQ-ERROR", TCL_STATIC);
  } else {
    Tcl_SetResult(interp, MqErrorGetText(msgque->error), TCL_VOLATILE);
    MqErrorReset(msgque->error);
  }
}

/** \brief helper to link a \libmsgque service package with a Tcl procedure
 *
 * \param[in] msgque the object to work on
 * \param[in] data the Tcl procedure as \e Tcl_Obj
 * \return \libmsgque error-code
 **/
enum MqErrorE
NS(ProcCall) (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  SETUP_context
  Tcl_Obj *lobjv[2];
  Tcl_Interp *interp = context->interp;
  int ret;

  MqErrorReset(msgque->error);

  // 1. add service handler
  lobjv[0] = (Tcl_Obj *) data;

  // 2. setup Command (e.g. the ContextS)
  lobjv[1] = context->self;

  // 3. evaluate the script
  Tcl_IncrRefCount(lobjv[0]);
  Tcl_IncrRefCount(lobjv[1]);
  context->callLevel++;
  ret = Tcl_EvalObjv (interp, 2, lobjv, TCL_EVAL_GLOBAL);
  context->callLevel--;
  Tcl_DecrRefCount(lobjv[1]);
  Tcl_DecrRefCount(lobjv[0]);


  // 4. if TCL_OK than return MQ_OK
  if (ret == TCL_OK) return MQ_OK;

  // now we start to check the error

  // 5. if the 'error' is from tcl (MqErrorGetCode (msgque->error) == MQ_OK)
  // convert the tcl into a msgque error
  if (MqErrorGetCode (msgque->error) == MQ_OK) {
    return MqErrorC (msgque->error,__func__,-1,Tcl_GetVar (interp, "errorInfo", TCL_GLOBAL_ONLY));
  }

  // 6. check to be in a transaction or not
  if (MqIsTransaction(msgque) == MQ_NO) {
    // 7. no transaction
    // return server error from a MqSendEND call asyncrone
    return MqSendERROR (msgque->send);
  } else if (MqCurrentTokenIs(msgque,"_FTR") || MqCurrentTokenIs(msgque,"_EOF")) {
    // 9. the _FTR and _EOF token have a "build-in" MqSendRETURN
    return MQ_OK;
  } else {
    // 10. in a default transaction
    // return a "server" error from a MqSendEND_AND_WAIT or MqSendEND_AND_callback call
    MqSendSTART(msgque->send);
    return MqSendRETURN (msgque->send);
  }
}

/** \brief free data located in service handles
 *  \msgque
 *  \param[in] data the objext used to store the tcl-procedure
 *  \return Tcl error-code
 **/
void
NS(ProcFree) (
  struct MqS const * const msgque,
  MQ_PTR data
)
{
  Tcl_DecrRefCount ( (Tcl_Obj*) data);
}


/// \brief check a \c tclProc argument to be valid
/// \param[in] interp current Tcl interpreter
/// \param[in] cmdObj the command name to check
/// \param[in] wrongNrStr wrong number error message
int
NS(ProcCheck) (
  Tcl_Interp * interp,
  struct Tcl_Obj * cmdObj,
  char const * const wrongNrStr
)
{
  int ret,len;
  Tcl_DString cmd;
  if (!Tcl_GetCommandFromObj (interp, cmdObj)) {
    Tcl_WrongNumArgs (interp, 0, NULL, wrongNrStr);
    return TCL_ERROR;
  }
  Tcl_DStringInit(&cmd);
  Tcl_DStringAppendElement(&cmd,"info");
  Tcl_DStringAppendElement(&cmd,"args");
  Tcl_DStringAppendElement(&cmd,Tcl_GetString(cmdObj));
  ret = Tcl_EvalEx(interp, Tcl_DStringValue(&cmd), Tcl_DStringLength(&cmd), TCL_EVAL_GLOBAL);
  Tcl_DStringFree(&cmd);
  TclErrorCheck(ret);
  TclErrorCheck(Tcl_ListObjLength(interp, Tcl_GetObjResult(interp), &len));
  if (len != 1) {
    Tcl_DString msg;
    Tcl_DStringInit(&msg);
    Tcl_DStringAppend(&msg,"wrong # args: ", -1);
    if (len > 1) Tcl_DStringAppend(&msg,"only ", -1);
    Tcl_DStringAppend(&msg,"one argument for procedure \"", -1);
    Tcl_DStringAppend(&msg,Tcl_GetString(cmdObj), -1);
    Tcl_DStringAppend(&msg,"\" is required", -1);
    Tcl_DStringResult(interp, &msg);
    Tcl_DStringFree(&msg);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/** \} */


