/**
 *  \file       tests/client.c
 *  \brief      \$Id: client.c 1175 2009-05-14 13:44:36Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1175 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

/** \ingroup validation
 *  \defgroup client client
 *  \{
 *  \brief client \client_desc
 *
 *  \verbinclude client.help
 */

#include "msgque.h"
#include "config.h"
#include "stat.h"
#include "debug.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>

#define NUM_TRANS   50000
#define NUM_PARENT  200
#define NUM_CHILD   1000
#define MSGQUE	    msgque

/*****************************************************************************/
/*                                                                           */
/*                                     main                                  */
/*                                                                           */
/*****************************************************************************/

static void ClientHelp ( const char * ) __attribute__ ((noreturn));

static MQ_INT callnum = 0;

static enum MqErrorE 
RET_ECUL
(
  struct MqS * msgque,
  MQ_PTR data
) {
  register struct MqReadS * const read = msgque->read;
  MQ_INT valI;
  MQ_WID valW;
  MQ_DBL valD;
  MQ_BUF buf;
  callnum++;
  MqErrorCheck (MqReadI (read, &valI));
  MqErrorCheck (MqReadW (read, &valW));
  MqErrorCheck (MqReadD (read, &valD));
  MqErrorCheck (MqReadU (read, &buf));
  return MQ_OK;

error:
  return MqErrorStack(msgque->error);
}

/// \brief display help using \b -h or \b --help command-line option
/// \param base the executable usually: <tt>basename(argv[0])</tt>
static void
ClientHelp ( const char * base  )
{
    fprintf(stderr, "usage: %s [OPTION]... [ARGUMENT]...\n", base);
    fputs("\n", stderr);
    fputs("  This tool is the client part of the performance test toolkit and expect\n", stderr);
    fputs("  the libmsgque performance test and validation 'server' as argument.\n", stderr);
    fputs("  The following test are defined:\n",stderr);
    fputs("\n",stderr);
    fputs("  --all\n",stderr);
    fputs("    Do all the following tests.\n", stderr);
    fputs("\n",stderr);
    fputs("  --send-perf\n",stderr);
    fputs("    Creation of '--num' (default: " MQ_CPPXSTR(NUM_TRANS) ") transactions.\n", stderr);
    fputs("    The data is send from the client to the server using four different\n", stderr);
    fputs("    kind of packages:\n", stderr);
    fputs("\n", stderr);
    fputs("      1. an MQ_INT integer between INT_MIN and INT_MAX\n", stderr);
    fputs("      2. an MQ_WID integer between WID_MIN and WID_MAX\n", stderr);
    fputs("      3. a  MQ_DBL double  between DBL_MIN and DBL_MAX\n", stderr);
    fputs("      4. a  MQ_BIN binary  of size between 1 and 1000 bytes\n", stderr);
    fputs("\n", stderr);
    fputs("  --send-and-wait-perf  (this is the DEFAULT)\n",stderr);
    fputs("    Same as '--send-perf' but use 'MqSendEND_AND_WAIT'\n", stderr);
    fputs("\n", stderr);
    fputs("  --send-and-callback-perf\n",stderr);
    fputs("    Same as '--send-perf' but use 'MqSendEND_AND_CALLBACK'\n", stderr);
    fputs("\n", stderr);
    fputs("  --parent-perf\n",stderr);
    fputs("    Creation of '--num' (default: " MQ_CPPXSTR(NUM_PARENT) ") parent-context\n", stderr);
    fputs("\n", stderr);
    fputs("  --child-perf\n",stderr);
    fputs("    Creation of '--num' (default: " MQ_CPPXSTR(NUM_CHILD) ") child-context\n", stderr);
    fputs("\n", stderr);
    fprintf(stderr, "  %s [ARGUMENT]... syntax:\n", base);
    fprintf(stderr, "    %s [OPTION]... %c server [OPTION]... [ARGUMENT]\n", base, MQ_ALFA);
    fputs("\n", stderr);
    fputs(MqHelp (NULL), stderr);
    fputs("\n", stderr);
    fprintf(stderr,"  %s [OPTION]:\n", base);
    fputs("    --num NUMBER     number of transactions\n", stderr);
    fputs("    -h, --help       print this help\n", stderr);
    fputs("\n", stderr);

    exit(EXIT_SUCCESS);
}

// help doxygen to build external references to the "main" proc of a tool
#define ClientMain main

// START-C-CLIENT-1
/// \brief main entry-point for the tool
/// \param argc the number of command-line arguments
/// \param argv the command-line arguments as an array of strings
/// \return the exit number
int
ClientMain (
  int argc,
  char **argv
)
{
  struct MqS * msgque = NULL;			// the initial context
  struct MqErrorS * error = MQ_ERROR_PANIC;     // the error object
  struct MqBufferS * buf;			// a object for return values
  // the commandline-arguments (before and after the first MQ_ALFA)
  struct MqBufferLS * largv, * lalfa;

  // what should be tested ?
  MQ_BOL sendB;		// test MqSendEND time?
  MQ_BOL sendAndWait;	// test MqSendEND_AND_WAIT round-trip time?
  MQ_BOL sendAndCall;	// test MqSendEND_AND_CALLBACK round-trip time?
  MQ_BOL parent;	// test parent-context creation time?
  MQ_BOL child;		// test child-context creation time?
  MQ_BOL all;		// test all
  MQ_INT num = -1;
#define SIZE 1000
  MQ_BIN data = MqSysMalloc (error, (SIZE));

  // make a backup of largv and lalfa for every test
  struct MqBufferLS *parentAlfa = NULL, *parentArgv = NULL;

  // create the client context and start the server
  struct MqConfigS * config = MqConfigInit();

  // fast link
  register struct MqSendS * send;
  register struct MqReadS * read;

  // split argv into largv and lalfa
  MqBufferLCreateArgv(argc, argv, &largv, &lalfa);
  parentArgv = MqBufferLDup(largv);
  parentAlfa = MqBufferLDup(lalfa);

  config->fHelp = ClientHelp;

  MqErrorCheck (MqCreate (config, largv, lalfa, &msgque));
  error = msgque->error;

  // read application specific arguments
  MqBufferLCheckOptionO (error, largv, "--send-perf", &sendB);
  MqBufferLCheckOptionO (error, largv, "--send-and-wait-perf", &sendAndWait);
  MqBufferLCheckOptionO (error, largv, "--send-and-callback-perf", &sendAndCall);
  MqBufferLCheckOptionO (error, largv, "--parent-perf", &parent);
  MqBufferLCheckOptionO (error, largv, "--child-perf", &child);
  MqBufferLCheckOptionO (error, largv, "--all", &all);
  if (all) {
    sendB = sendAndWait = sendAndCall = parent = child = MQ_YES;
  } else if (sendB == MQ_NO && sendAndWait == MQ_NO && sendAndCall == MQ_NO
		&& parent == MQ_NO && child == MQ_NO) {
    sendAndWait = MQ_YES;
  }

  // the user can supply --num to change the number of iterations
  MqBufferLCheckOptionI(error, largv,"--num",&num);

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(msgque, &largv));

  send = msgque->send;
  read = msgque->read;

  // initialize memory, just run one test-case to 
  // initialize dynamic data
  memset (data, 'A', SIZE);
  MqSendSTART (send);
  MqSendB (send, data, SIZE);
  MqErrorCheck (MqSendEND_AND_WAIT (send, "ECOU", MQ_TIMEOUT10));
  MqErrorCheck (MqReadU (read, &buf));

  // start the MqSendEND_AND_WAIT transaction-performance test
  MqDLogV (msgque, __func__, 0, "start: --------------------------------------\n");

  if (sendB) {

    // if necessary apply the default number of transactions
    MQ_INT lnum = (num == -1 ? NUM_TRANS : num);

    StatTimerSP itemT = StatCreate (msgque);

    {
      StatCtxSP stat = StatCtxCreate (msgque, "MqSendEND", 0);
      const MQ_INT stepI = ((INT_MAX/lnum)*2);
      const MQ_WID stepW = ((LLONG_MAX/lnum)*2);
      const MQ_DBL stepD = (DBL_MAX/lnum)    ;
      MQ_INT valI = INT_MIN;
      MQ_WID valW = LLONG_MIN;
      MQ_DBL valD = DBL_MIN;
      int i; 
      StatInit (itemT);
	for (i=0; i<lnum; i++) {
	  MqSendSTART (send);
	  MqSendI (send, valI);
	  MqSendW (send, valW);
	  MqSendD (send, valD);
	  MqSendB (send, data, ((i%SIZE)+1));
	  MqErrorCheck (MqSendEND (send, "RDUL"));
	  valI += stepI;
	  valW += stepW;
	  valD += stepD;
	};
	// just sync with server
	MqSendSTART (send);
	MqSendI(send, valI);
	MqErrorCheck (MqSendEND_AND_WAIT (send, "ECOI", 10));
	MqErrorCheck (MqReadI(msgque->read,&valI));
      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);
    }

    // cleanup
    StatDelete (&itemT);

  } /* finish the MqSendEND test */

  if (sendAndCall) {

    // if necessary apply the default number of transactions
    MQ_INT lnum = (num == -1 ? NUM_TRANS : num);

    StatTimerSP itemT = StatCreate (msgque);

    {
      StatCtxSP stat = StatCtxCreate (msgque, "MqSendEND_AND_CALLBACK", 0);
      const MQ_INT stepI = ((INT_MAX/lnum)*2);
      const MQ_WID stepW = ((LLONG_MAX/lnum)*2);
      const MQ_DBL stepD = (DBL_MAX/lnum)    ;
      MQ_INT valI = INT_MIN;
      MQ_WID valW = LLONG_MIN;
      MQ_DBL valD = DBL_MIN;
      int i; 
      StatInit (itemT);
	for (i=0; i<lnum; i++) {
	  MqSendSTART (send);
	  MqSendI (send, valI);
	  MqSendW (send, valW);
	  MqSendD (send, valD);
	  MqSendB (send, data, ((i%SIZE)+1));
	  MqErrorCheck (MqSendEND_AND_CALLBACK (send, "ECUL", RET_ECUL, NULL, NULL));
	  valI += stepI;
	  valW += stepW;
	  valD += stepD;
	  // don't flood the socket buffer with unread messages
	  if ((i % 7) == 0) {
	    while (MqProcessEvent(msgque, 3, MQ_WAIT_NO) == MQ_OK);
	  }
	};

      // wait untill all callbacks are processed
      while (callnum != lnum)
	MqProcessEvent(msgque, 3, MQ_WAIT_ONCE);

      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);
    }

    // cleanup
    StatDelete (&itemT);

  } /* finish the MqSendEND_AND_CALLBACK test */

  if (sendAndWait) {

    // if necessary apply the default number of transactions
    MQ_INT lnum = (num == -1 ? NUM_TRANS : num);

    StatTimerSP itemT = StatCreate (msgque);

    {
      StatCtxSP stat = StatCtxCreate (msgque, "MqSendEND_AND_WAIT", 0);
      const MQ_INT stepI = ((INT_MAX/lnum)*2);
      const MQ_WID stepW = ((LLONG_MAX/lnum)*2);
      const MQ_DBL stepD = (DBL_MAX/lnum)    ;
      MQ_INT valI = INT_MIN;
      MQ_WID valW = LLONG_MIN;
      MQ_DBL valD = DBL_MIN;
      MQ_BUF buf;
      int i; 
      StatInit (itemT);
	for (i=0; i<lnum; i++) {
	  MqSendSTART (send);
	  MqSendI (send, valI);
	  MqSendW (send, valW);
	  MqSendD (send, valD);
	  MqSendB (send, data, ((i%SIZE)+1));
	  MqErrorCheck (MqSendEND_AND_WAIT (send, "ECUL", MQ_TIMEOUT10));
	  MqErrorCheck (MqReadI (read, &valI));
	  MqErrorCheck (MqReadW (read, &valW));
	  MqErrorCheck (MqReadD (read, &valD));
	  MqErrorCheck (MqReadU (read, &buf));
	  valI += stepI;
	  valW += stepW;
	  valD += stepD;
	};
      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);
    }

    // cleanup
    StatDelete (&itemT);

  } /* finish the transaction-performance test */

  // start the parent-context creation test
  if (parent) {
    int n;
    StatTimerSP itemT;

    // if necessary apply the default number of transactions
    MQ_INT lnum = (num == -1 ? NUM_PARENT : num);
    
    struct MqS** msgqueA = (struct MqS**) MqSysMalloc(error, lnum * sizeof(struct MqS*));
    struct MqBufferLS** lalfaA = (struct MqBufferLS**) MqSysMalloc(error, lnum * sizeof(struct MqBufferLS*));
    struct MqBufferLS** largvA = (struct MqBufferLS**) MqSysMalloc(error, lnum * sizeof(struct MqBufferLS*));

    // copy the argument vector to create 'num' parent context
    for (n=0; n<lnum; n++) {
      lalfaA[n] = MqBufferLDup(parentAlfa);
      largvA[n] = MqBufferLDup(parentArgv);
    }

    itemT = StatCreate (msgque);

    {
      // loop to create the parent context
      StatCtxSP stat = StatCtxCreate (msgque, "parent create", 0);
      StatInit (itemT);
      for (n=0; n<lnum; n++) {
	if (MqCreate (MqConfigDup(&msgque->config), largvA[n], lalfaA[n], &msgqueA[n]) == MQ_ERROR) {
	  MqDelete (&msgque);
	  msgque = msgqueA[n];
	  lnum = n-1;
	  for (n=0; n < lnum; n++) {
	    MqDelete (&msgqueA[n]);
	  }
	  goto error;
	}
	MqBufferLDelete(&largvA[n]);
      }
      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);

      // loop to delete the parent context
      stat = StatCtxCreate (msgque, "parent delete", 0);
      StatInit (itemT);
      for (n=0; n<lnum; n++) {
	MqDelete (&msgqueA[n]);
      }
      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);
    }

    // cleanup
    MqSysFree(lalfaA);
    MqSysFree(largvA);
    MqSysFree(msgqueA);
    StatDelete (&itemT);
  }

  // start the child-context creation test
  if (child) {
    int n;
    StatTimerSP itemT;

    // if necessary apply the default number of transactions
    MQ_INT lnum = (num == -1 ? NUM_CHILD : num);
    
    struct MqS** contextA = (struct MqS**) MqSysMalloc(error, lnum * sizeof(struct MqS*));

    // fill template configuration
    struct MqConfigS *template = MqConfigInit();
    struct MqConfigS *itemConfig;
    template->name = strdup("child");
    template->debug = msgque->config.debug;

    // copy the argument vector to create 'num' parent context
    itemT = StatCreate (msgque);

    {
      // loop to create the parent context
      StatCtxSP stat = StatCtxCreate (msgque, "child create", 0);
      StatInit (itemT);
      for (n=0; n<lnum; n++) {
	itemConfig = MqConfigDup(template);
	itemConfig->parent = msgque;
	if (MqCreate (itemConfig, NULL, NULL, &contextA[n]) == MQ_ERROR) {
	  MqErrorCopy(msgque->error,contextA[n]->error);
	  goto error;
	}
      }
      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);

      // loop to delete the parent context
      stat = StatCtxCreate (msgque, "child delete", 0);
      StatInit (itemT);
      for (n=0; n<lnum; n++) {
	MqDelete (&contextA[n]);
      }
      StatCtxCalc (stat, itemT);
      stat->val /= lnum;
      StatCtxPrint (stat);
      StatCtxDelete (&stat);
    }

    // cleanup
    MqSysFree(template);
    MqSysFree(contextA);
    StatDelete (&itemT);
  }

// cleanup
  MqSysFree(data);
  MqDLogV (msgque, __func__, 0, "end: ----------------------------------------\n");

// START-C-CLIENT-4
  MqBufferLDelete(&parentAlfa);
  MqBufferLDelete(&parentArgv);
  MqExit (msgque);

error:
  MqErrorSAppendC(msgque->error, "use '-h' or '--help' for usage");
  MqExit (msgque);
}
// END-C-CLIENT-4

/** \} client */
