/**
 *  \file       tests/server_context.c
 *  \brief      \$Id: server_context.c 1178 2009-05-15 07:25:57Z dev1usr $
 *  
 *  (C) 2004 - LibMsgque - Project - Group
 *  
 *  \version    \$Rev: 1178 $
 *  \author     EMail: aotto1968 at users.sourceforge.net
 *  \attention  this software has GPL permissions to copy
 *              please contact AUTHORS for additional information
 */

/** \addtogroup server
 * \{
 */

#include "server_context.h"
#include <stdlib.h>
#include <string.h>
#include "debug.h"

/// link to the MqErrorS object
#define MQ_ERROR_S context->msgque->error
#define MSGQUE msgque

/*****************************************************************************/
/*                                                                           */
/*                               Request Handler                             */
/*                                                                           */
/*****************************************************************************/

/// \brief display help using \b -h or \b --help command-line option
/// \param base the executable usually: <tt>basename(argv[0])</tt>
static void __attribute__ ((noreturn))
ServerHelp (const char * base)
{
    fprintf(stderr, "usage: %s [OPTION]... [ARGUMENT]...\n", base);
    fputs("\n", stderr);
    fputs("  This tool is the server part of the performance test toolkit and expect\n", stderr);
    fputs("  packages from the libmsgque performance test and validation client.\n", stderr);
    fputs("\n", stderr);
    fprintf(stderr, "  %s [ARGUMENT]... syntax:\n", base);
    fprintf(stderr, "    aclient [OPTION]... %c %s [OPTION]... [ARGUMENT]\n", MQ_ALFA, base);
    fputs("\n", stderr);
    fputs(MqHelp (NULL), stderr);
    fputs("\n", stderr);
    fprintf(stderr,"  %s [OPTION]:\n", base);
    fputs("    -h, --help       print this help\n", stderr);
    fputs("\n", stderr);

    exit(EXIT_SUCCESS);
}

/// \brief test a non recoverable error (panic)
/// \service
static enum MqErrorE
Ot_ERR1 (
  struct MqS * const msgque,
  void *data
)
{
  MqPanicC (msgque->error, __func__, -1, "this is a non recoverable error");

  // this line will never reached
  return MQ_ERROR;
}

/// \brief test a recoverable error using an error return package
/// \service
static enum MqErrorE
Ot_ERR2 (
  struct MqS * const msgque,
  void *data
)
{
  register struct MqSendS * const send = msgque->send;

  MqSendSTART (send);
  MqSendC (send, "some data");
  MqErrorC (msgque->error, __func__, 10, "some error");

  return MqSendRETURN (send);
}

/// \brief double return message
/// \service
static enum MqErrorE
Ot_ERR3 (
  struct MqS * const msgque,
  void *data
)
{
  register struct MqSendS * const send = msgque->send;

  MqSendSTART (send);
  // double MqSendRETURN
  MqSendRETURN (send);

  return MqSendRETURN (send);
}

/// \brief echo a MqBufferS object
/// \service
static enum MqErrorE
Ot_ECOU (
  struct MqS * const msgque,
  void *data
)
{
  register struct MqSendS * const send = msgque->send;

  struct MqBufferS *buf;
  MqSendSTART (send);
  MqErrorCheck (MqReadU (msgque->read, &buf));

  MqSendU (send, buf);

error:
  return MqSendRETURN (send);
}

/// \brief echo the server-name
/// \service
static enum MqErrorE
Ot_ECON (
  struct MqS * const msgque,
  void *data
)
{
  register struct MqSendS * const send = msgque->send;

  MQ_STR  str;
  MqSendSTART (send);
  MqErrorCheck (MqReadC (msgque->read, &str));
  MqBufferSetV(msgque->temp,"%s-%s", str, msgque->config.name);
  MqSendC (send, msgque->temp->cur.C);

error:
  return MqSendRETURN (send);
}

// START-C-SERVICE-PROC
/// \brief echo a #MQ_INT object
/// \service
static enum MqErrorE
Ot_ECOI (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_INT l;
  MqSendSTART (send);
  MqErrorCheck (MqReadI (msgque->read, &l));
  MqSendI (send, l);

error:
  return MqSendRETURN (send);
}
// END-C-SERVICE-PROC

/// \brief echo a #MQ_SRT object
/// \service
static enum MqErrorE
Ot_ECOS (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_SRT l;
  MqSendSTART (send);
  MqErrorCheck (MqReadS (msgque->read, &l));
  MqSendS (send, l);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_BYT object
/// \service
static enum MqErrorE
Ot_ECOY (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_BYT l;
  MqSendSTART (send);
  MqErrorCheck (MqReadY (msgque->read, &l));
  MqSendY (send, l);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_BOL object
/// \service
static enum MqErrorE
Ot_ECOO (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_BOL l;
  MqSendSTART (send);
  MqErrorCheck (MqReadO (msgque->read, &l));
  MqSendO (send, l);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_WID object
/// \service
static enum MqErrorE
Ot_ECOW (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_WID w;
  MqSendSTART (send);
  MqErrorCheck (MqReadW (msgque->read, &w));
  MqSendW (send, w);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_DBL object
/// \service
static enum MqErrorE
Ot_ECOD (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_DBL d;
  MqSendSTART (send);
  MqErrorCheck (MqReadD (msgque->read, &d));
  MqSendD (send, d);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_FLT object
/// \service
static enum MqErrorE
Ot_ECOF (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_FLT d;
  MqSendSTART (send);
  MqErrorCheck (MqReadF (msgque->read, &d));
  MqSendF (send, d);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_STR object
/// \service
static enum MqErrorE
Ot_ECOC (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_STR str;
  MqSendSTART (send);
  MqErrorCheck (MqReadC (msgque->read, &str));
  MqSendC (send, str);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MQ_BIN object
/// \service
static enum MqErrorE
Ot_ECOB (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_BIN bin;
  MQ_SIZE len;
  MqSendSTART (send);
  MqErrorCheck (MqReadB (msgque->read, &bin, &len));
  MqSendB (send, bin, len);

error:
  return MqSendRETURN (send);
}

/// \brief helper to increment/echo data objects
/// \read
/// \param incr bool for increment testing 0 = no else yes
/// \retMqErrorE
static enum MqErrorE
_ECOL_Item (
  struct MqReadS * const read,
  const int incr
)
{
  struct MqS * const msgque = MqReadGetMsgque(read);
  struct MqErrorS * const error = msgque->error;
  struct MqSendS * const send = msgque->send;
  struct MqBufferS * buf;
  MQ_INT i;
  while (MqReadGetNumItems (read)) {
    MqErrorCheck (MqReadU (read, &buf));
    if (buf->type == MQ_LSTT) {
      MqReadL_START (read, buf);
      MqSendL_START (send);
      MqErrorCheck (_ECOL_Item (read, incr));
      MqSendL_END (send);
      MqReadL_END (read);
    } else if (incr) {
      MqErrorCheck (MqBufferGetI (error, buf, &i));
      MqSendI (send, ++i);
    } else {
      MqSendU (send, buf);
    }
  }

  return MQ_OK;

error:
  return MqErrorStack(msgque->error);
}

/// \brief echo a #MQ_LST object
/// \service
static enum MqErrorE
Ot_ECLI (
  struct MqS * const msgque,
  void *data
)
{
  MQ_BUF buf;
  MQ_BOL doincr;

  MqErrorCheck( MqReadU(msgque->read,&buf));

  doincr = (buf->type == MQ_STRT && strncmp(buf->cur.C,"--incr",6) == 0);
  if (!doincr) MqReadUndo(msgque->read);
  
  MqSendSTART (msgque->send);
  MqErrorCheck (_ECOL_Item (msgque->read, doincr));

error:
  return MqSendRETURN (msgque->send);
}

/// \brief force echo a #MQ_LST object
/// \service
static enum MqErrorE
Ot_ECOL (
  struct MqS * const msgque,
  void *data
)
{
  register struct MqSendS * const send = msgque->send;
  register struct MqReadS * const read = msgque->read;

  MqSendSTART (send);
  MqErrorCheck (MqReadL_START (read, NULL));
  MqSendL_START (send);
  MqErrorCheck (_ECOL_Item (read, 0));
  MqSendL_END (send);
  MqReadL_END (read);

error:
  return MqSendRETURN (send);
}

/// \brief echo a #MqBufferLS object
/// \service
static enum MqErrorE
Ot_ECUL (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  MQ_INT i;
  MQ_WID w;
  MQ_DBL d;
  MQ_BUF b;

  MqErrorCheck(MqReadI(msgque->read, &i));
  MqErrorCheck(MqReadW(msgque->read, &w));
  MqErrorCheck(MqReadD(msgque->read, &d));
  MqErrorCheck(MqReadU(msgque->read, &b));

  MqSendSTART (send);

  MqSendI (send, i);
  MqSendW (send, w);
  MqSendD (send, d);
  MqSendU (send, b);

error:
  return MqSendRETURN (send);
}

/// \brief read a #MqBufferLS object
/// \service
static enum MqErrorE
Ot_RDUL (
  struct MqS * const msgque,
  void *data
)
{
  MQ_INT i;
  MQ_WID w;
  MQ_DBL d;
  MQ_BUF b;

  MqErrorCheck(MqReadI(msgque->read, &i));
  MqErrorCheck(MqReadW(msgque->read, &w));
  MqErrorCheck(MqReadD(msgque->read, &d));
  MqErrorCheck(MqReadU(msgque->read, &b));

error:
  return MQ_OK;
}

/// \brief echo a variable length buffer object
/// \service
static enum MqErrorE
Ot_ECOM (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MQ_BUF buf;

  MqSendSTART(send);
  while (MqReadGetNumItems(read)) {
    MqErrorCheck(MqReadU(read,&buf));
    MqSendU(send,buf);
  }

error:
  return MqSendRETURN (send);
}

/// \brief return debug, binary, and silent value
/// \service
static enum MqErrorE
Ot_MSQT (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;

  // Send
  MqSendSTART (send);
  if (msgque->config.debug) {
    MqSendC (send, "debug");
    MqSendI (send, msgque->config.debug);
  }
  if (!msgque->config.isString)
    MqSendC (send, "binary");
  if (msgque->config.isSilent)
    MqSendC (send, "silent");
  MqSendC (send, "sOc");
  MqSendC (send, MqLogServerOrClient (msgque));
  MqSendC (send, "pOc");
  MqSendC (send, MqLogParentOrChild (msgque));

  return MqSendRETURN (send);
}

/// \brief server sleeps n seconds
/// \service
static enum MqErrorE
Ot_SLEP (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MQ_INT time;

  MqSendSTART (send);
  MqErrorCheck (MqReadI (read, &time));
  MqErrorCheck (MqSysSleep (msgque->error, time));
  MqSendI (send, time);

error:
  return MqSendRETURN (send);
}

/// \brief server sleeps n microseconds
/// \service
static enum MqErrorE
Ot_USLP (
  struct MqS * const msgque,
  void *data
)
{
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MQ_INT time;

  MqSendSTART (send);
  MqErrorCheck (MqReadI (read, &time));
  MqErrorCheck (MqSysUSleep (msgque->error, time));
  MqSendI (send, time);

error:
  return MqSendRETURN (send);
}

/// \brief test client/server interaction
/// \service
static enum MqErrorE
Ot_CSV1 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MQ_INT num;

  // call an other service
  //    I don't need an transactionPtr because retTok = NULL
  MqSendSTART (send);
  MqErrorCheck (MqReadI (read, &num));
  num++;
  MqSendI (send, num);
  MqErrorCheck(MqSendEND_AND_WAIT (send, "CSV2", MQ_TIMEOUT));

  // read the answer and send the result back
  MqSendSTART (send);
  MqErrorCheck (MqReadI (read, &num));
  num++;
  MqSendI (send, num);

error:
  return MqSendRETURN (send);
}

/// \brief test mass startup of childs
/// \service
static enum MqErrorE
Ot_GTCX (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;

  MqSendSTART (send);
  MqSendI (send, MqGetCtxId(msgque));
  return MqSendRETURN (send);
}

/// \brief get the current token
/// \service
static enum MqErrorE
Ot_GTTO (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;

  MqSendSTART (send);
  MqSendC (send, MqGetToken(msgque));
  return MqSendRETURN (send);
}

/// \brief get the MqConfigS::statusIs field value
/// \service
static enum MqErrorE
Ot_GTSI (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;
  MqBufferReset(msgque->temp);

  MqSendSTART (send);
  if (msgque->config.statusIs & MQ_STATUS_IS_DUP)
    MqBufferAppendC(msgque->temp, "DUP-");
  if (msgque->config.statusIs & MQ_STATUS_IS_THREAD)
    MqBufferAppendC(msgque->temp, "THREAD-");
  if (msgque->config.statusIs & MQ_STATUS_IS_FORK)
    MqBufferAppendC(msgque->temp, "FORK-");
  if (msgque->config.statusIs & MQ_STATUS_IS_SPAWN)
    MqBufferAppendC(msgque->temp, "SPAWN-");
  if (msgque->config.statusIs & MQ_STATUS_IS_MANAGED)
    MqBufferAppendC(msgque->temp, "MANAGED-");
  MqSendU(send, msgque->temp);
  return MqSendRETURN (send);
}

/// \brief set context specific data
/// \service
static enum MqErrorE
Ot_SETU (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct ServerCtxS * context = (struct ServerCtxS *) msgque->config.myCtx;
  return MqReadU(msgque->read, &context->val);
}

/// \brief get context specific data
/// \service
static enum MqErrorE
Ot_GETU (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct ServerCtxS * context = (struct ServerCtxS *) msgque->config.myCtx;
  struct MqSendS * const send = msgque->send;
  MqSendSTART (send);
  MqSendU (send, context->val);
  return MqSendRETURN (send);
}

/// \brief "callback" used in #Ot_SND1
/// \service
static enum MqErrorE
Callback1 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct ClientCtxS * clientCtx = (struct ClientCtxS *) msgque->config.myCtx;
  MqErrorCheck(MqReadI(msgque->read, &clientCtx->sv->i));
  return MQ_OK;
error:
  return MqErrorStack(msgque->error);
}

/// \brief "callback" used in #Ot_SND2
/// \service
static enum MqErrorE
Callback2 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct ServerCtxS * const ctx = (struct ServerCtxS *const) msgque->config.master->config.myCtx;
  MqErrorCheck(MqReadI(msgque->read, &ctx->i));
  return MQ_OK;
error:
  return MqErrorStack(msgque->error);
}

/// \brief just read an MQ_INT
/// \service
static enum MqErrorE
Ot_REDI (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  MQ_INT i;
  MqErrorCheck(MqReadI(msgque->read, &i));
  return MQ_OK;
error:
  return MqErrorStack(msgque->error);
}

/// \brief test buffer #1
/// \service
static enum MqErrorE
Ot_BUF1 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  MQ_BUF buf;
  char str[2] = { '\0', '\0' };
  struct MqSendS * const send = msgque->send;
  MqErrorCheck(MqReadU(msgque->read, &buf));
  MqSendSTART(send);
  str[0] = MqBufferGetType(buf);
  MqSendC(send, str);
  switch (buf->type) {
    case MQ_BYTT:
      MqSendY(send, buf->cur.A->Y); break;
    case MQ_BOLT:
      MqSendO(send, buf->cur.A->O); break;
    case MQ_SRTT:
      MqSendS(send, buf->cur.A->S); break;
    case MQ_INTT:
      MqSendI(send, buf->cur.A->I); break;
    case MQ_FLTT:
      MqSendF(send, buf->cur.A->F); break;
    case MQ_WIDT:
      MqSendW(send, buf->cur.A->W); break;
    case MQ_DBLT:
      MqSendD(send, buf->cur.A->D); break;
    case MQ_BINT:
      MqSendB(send, buf->cur.B, buf->cursize); break;
    case MQ_STRT:
      MqSendC(send, buf->cur.C); break;
    case MQ_LSTT:
    case MQ_RETT:
      break;
  }
  return MqSendRETURN(send);
error:
  return MqErrorStack(msgque->error);
}

/// \brief test buffer #2
/// \service
static enum MqErrorE
Ot_BUF2 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  MQ_BUF buf;
  int i;
  char str[2] = {'\0', '\0'};
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MqSendSTART(send);
  for (i=0; i<3; i++) {
    MqErrorCheck(MqReadU(read, &buf));
    str[0] = MqBufferGetType(buf);
    MqSendC(send, str);
    MqSendU(send, buf);
  }
  return MqSendRETURN(send);
error:
  return MqErrorStack(msgque->error);
}

/// \brief test buffer #3
/// \service
static enum MqErrorE
Ot_BUF3 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  MQ_BUF buf;
  MQ_INT i;
  char str[2] = {'\0', '\0'};
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MqSendSTART(send);
  MqErrorCheck(MqReadU(read, &buf));
  str[0] = MqBufferGetType(buf);
  MqSendC(send, str);
  MqSendU(send, buf);
  MqErrorCheck(MqReadI(read, &i));
  MqSendI(send, i);
  MqSendU(send, buf);
  return MqSendRETURN(send);
error:
  return MqErrorStack(msgque->error);
}

/// \brief test list #1
/// \service
static enum MqErrorE
Ot_LST1 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;
  MqSendSTART(send);
  MqSendL_END(send);
  return MqSendRETURN(send);
}

/// \brief test list #2
/// \service
static enum MqErrorE
Ot_LST2 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;
  MqSendSTART(send);
  MqReadL_END(msgque->read);
  return MqSendRETURN(send);
}

/// \brief test the "MqSend*" feature
/// \service
static enum MqErrorE
Ot_SND1 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  #define LocalErrorCheckAndCopy(PROC) \
    MqErrorCheckAndCopy(PROC, serverCtx->cl[clid]->error, msgque->error)
  struct ServerCtxS * serverCtx = (struct ServerCtxS *) msgque->config.myCtx;
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MQ_STR s;
  MQ_BUF buf;
  MQ_INT clid;
  MqErrorCheck(MqReadC(read, &s));
  MqErrorCheck(MqReadI(read, &clid));
  MqSendSTART(send);
    if (!strcmp(s,"START")) {
      struct ClientCtxS * clientCtx = MqSysCalloc(msgque->error,1,sizeof(*clientCtx));
      struct MqConfigS *config = MqConfigInit();
      struct MqBufferLS *alfa = MqBufferLCreate(2);
      clientCtx->sv = serverCtx;
      config->debug = msgque->config.debug;
      config->isString = MQ_NO;
      config->isSilent = MQ_NO;
      config->myCtx = (struct MqContextS *) clientCtx;
      MqConfigInitName(config, "test-client");
      if (msgque->config.parent != NULL)
	config->parent = ((struct ServerCtxS *)msgque->config.parent->config.myCtx)->cl[clid];
      MqBufferLAppendC(alfa,"SELF");
      MqBufferLAppendC(alfa,"--name");
      MqBufferLAppendC(alfa,"test-server");
      if (MqCreate(config,NULL,alfa,&serverCtx->cl[clid]) == MQ_ERROR) {
	MqErrorCopy (msgque->error, serverCtx->cl[clid]->error);
	MqDelete(&serverCtx->cl[clid]);
	goto error;
      }
    } else if (!strcmp(s,"STOP")) {
      MqDelete(&serverCtx->cl[clid]);
    } else if (!strcmp(s,"SEND")) {
      MQ_STR TOK;
      struct MqSendS * const clsend = serverCtx->cl[clid]->send;
      MqSendSTART(clsend);
      MqErrorCheck(MqReadC(read, &TOK));
      MqErrorCheck(MqReadU(read, &buf));
      MqSendU(clsend, buf);
      LocalErrorCheckAndCopy(MqSendEND(clsend, TOK));
    } else if (!strcmp(s,"WAIT")) {
      struct MqSendS * const clsend = serverCtx->cl[clid]->send;
      MqSendSTART(clsend);
      MqErrorCheck(MqReadU(read, &buf));
      MqSendU(clsend, buf);
      LocalErrorCheckAndCopy(MqSendEND_AND_WAIT(clsend, "ECOI", 5));
      LocalErrorCheckAndCopy(MqReadI(serverCtx->cl[clid]->read, &serverCtx->i));
      MqSendI(send, serverCtx->i+1);
    } else if (!strcmp(s,"CALLBACK")) {
      struct MqSendS * const clsend = serverCtx->cl[clid]->send;
      serverCtx->i = -1;
      MqSendSTART(clsend);
      MqErrorCheck(MqReadU(read, &buf));
      MqSendU(clsend, buf);
      LocalErrorCheckAndCopy(MqSendEND_AND_CALLBACK(clsend, "ECOI", Callback1, NULL, NULL));
      LocalErrorCheckAndCopy(MqProcessEvent(serverCtx->cl[clid], 10, MQ_WAIT_ONCE));
      MqSendI(send, serverCtx->i+1);
    }

error:
  return MqSendRETURN (send);
}

/// \brief test the "MqSend*" feature
/// \service
static enum MqErrorE
Ot_SND2 (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct ServerCtxS * serverCtx = (struct ServerCtxS *) msgque->config.myCtx;
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  MQ_STR s;
  MQ_BUF buf;
  MQ_INT clid;
  struct MqS * cl;
  MqErrorCheck(MqReadC(read, &s));
  MqErrorCheck(MqReadI(read, &clid));
  cl = MqSlaveGet(msgque, clid);
  MqSendSTART(send);
    if (!strcmp(s,"CREATE")) {
struct MqS * master = msgque;
struct MqBufferLS * argv = MqBufferLCreate(5);
struct MqConfigS * const config = MqConfigDup (&master->config);
MqConfigInitMaster(config, master, clid);
MqConfigInitName (config, "test-client");
MqBufferLAppendC(argv, "worker-client");
MqBufferLAppendC(argv, "--srvname");
MqBufferLAppendC(argv, "test-server");
while (MqReadItemExists(read)) {
  MqErrorCheck(MqReadC(read, &s));
  MqBufferLAppendC(argv, s);
}
MqErrorCheck(ServerParentCreate(config, argv, NULL, &cl));
    } else if (!strcmp(s,"CREATE2")) {
      struct ClientCtxS * clientCtx = MqSysCalloc(msgque->error,1,sizeof(*clientCtx));
      struct MqConfigS *config = MqConfigInit();
      struct MqBufferLS *alfa = MqBufferLCreate(2);
      clientCtx->sv = serverCtx;
      config->debug = msgque->config.debug;
      config->myCtx = (struct MqContextS *) clientCtx;
      MqConfigInitName(config, "test-client");
      if (msgque->config.parent != NULL)
	config->parent = ((struct ServerCtxS *)msgque->config.parent->config.myCtx)->cl[clid];
      MqBufferLAppendC(alfa,"SELF");
      MqBufferLAppendC(alfa,"--name");
      MqBufferLAppendC(alfa,"test-server");
      if (MqCreate(config,NULL,alfa,&cl) == MQ_ERROR) {
	MqErrorCopy (msgque->error, cl->error);
	MqDelete(&cl);
	goto error;
      }
      MqErrorCheck (MqSlaveCreate (msgque, clid, cl));
    } else if (!strcmp(s,"DELETE")) {
      MqErrorCheck(MqSlaveDelete(msgque, clid));
      MqSendC (send, MqSlaveGet(msgque, clid) == NULL ? "OK" : "ERROR");
    } else if (!strcmp(s,"SEND")) {
      MQ_STR TOK;
      struct MqSendS * const clsend = cl->send;
      MqSendSTART(clsend);
      MqErrorCheck(MqReadC(read, &TOK));
      MqErrorCheck(MqReadU(read, &buf));
      MqSendU(clsend, buf);
      MqErrorCheck(MqSendEND(clsend, TOK));
    } else if (!strcmp(s,"WAIT")) {
      struct MqSendS * const clsend = cl->send;
      MqSendSTART(clsend);
      MqErrorCheck(MqReadU(read, &buf));
      MqSendU(clsend, buf);
      MqErrorCheck(MqSendEND_AND_WAIT(clsend, "ECOI", 5));
      MqErrorCheck(MqReadI(cl->read, &serverCtx->i));
      MqSendI(send, serverCtx->i+1);
    } else if (!strcmp(s,"MqSendEND_AND_WAIT")) {
      struct MqSendS * const clsend = cl->send;
      struct MqReadS * const clread = cl->read;
      MQ_STR token;
      MqErrorCheck(MqReadC(read, &token));
      MqSendSTART(clsend);
      while (MqReadItemExists(read)) {
	MqErrorCheck(MqReadU(read, &buf));
	MqSendU(clsend, buf);
      }
      MqErrorCheck(MqSendEND_AND_WAIT(clsend, token, 5));
      while (MqReadItemExists(clread)) {
	MqErrorCheck(MqReadU(clread, &buf));
	MqSendU(send, buf);
      }
    } else if (!strcmp(s,"MqSendEND")) {
      struct MqSendS * const clsend = cl->send;
      MQ_STR token;
      MqErrorCheck(MqReadC(read, &token));
      MqSendSTART(clsend);
      while (MqReadItemExists(read)) {
	MqErrorCheck(MqReadU(read, &buf));
	MqSendU(clsend, buf);
      }
      MqErrorCheck(MqSendEND(clsend, token));
      return MQ_OK;
    } else if (!strcmp(s,"CALLBACK")) {
      struct MqSendS * const clsend = cl->send;
      serverCtx->i = -1;
      MqSendSTART(clsend);
      MqErrorCheck(MqReadU(read, &buf));
      MqSendU(clsend, buf);
      MqErrorCheck(MqSendEND_AND_CALLBACK(clsend, "ECOI", Callback2, NULL, NULL));
      MqErrorCheck(MqProcessEvent(cl, 10, MQ_WAIT_ONCE));
      MqSendI(send, serverCtx->i+1);
    } else if (!strcmp(s,"ERR-1")) {
      struct MqS * master = msgque;
      struct MqBufferLS * alfa = MqBufferLCreate(5);
      struct MqConfigS * const config = MqConfigInit ();
      MqConfigInitMaster(config, master, clid);
      MqConfigInitName (config, "test-client");
      MqBufferLAppendC(alfa, "DUMMY");
      if (ServerParentCreate(config, NULL, alfa, &cl) == MQ_ERROR) {
	MqDelete (&cl);
      }
    } else if (!strcmp(s,"isSlave")) {
      MqSendO(send, MqConfigGetIsSlave(cl) == 1);
    }

error:
  return MqSendRETURN (send);
}

/// \brief test the "Init" feature
/// \service
static enum MqErrorE
Ot_INIT (
  struct MqS * const msgque,
  MQ_PTR data
)
{
  struct MqSendS * const send = msgque->send;
  struct MqReadS * const read = msgque->read;
  struct MqBufferLS * initB = MqInitCreate();
  MQ_STR str;
  MqSendSTART(send);
  while (MqReadItemExists(read)) {
    MqErrorCheck(MqReadC(read, &str));
    MqBufferLAppendC(initB, str);
  }
error:
  return MqSendRETURN (send);
}

/*****************************************************************************/
/*                                                                           */
/*                                context_init                               */
/*                                                                           */
/*****************************************************************************/

/// \brief create the ServerCtxS object
/// \msgqueP
/// \attDelete
static void
ServerDelete (
  struct MqS ** msgqueP
)
{
  if (unlikely(msgqueP == NULL || *msgqueP == NULL))
    return;
  free ((*msgqueP)->config.myCtx);
  MqDelete (msgqueP);
}

/// \tool_Create
static enum MqErrorE
ServerChildCreate (
  struct MqConfigS  * config,
  struct MqBufferLS * const argv,
  struct MqBufferLS * alfa,
  struct MqS       ** msgqueP
)
{
  struct MqS * msgque;

  // create the context object
  config->myCtx = MqSysCalloc (MQ_ERROR_PANIC, 1, sizeof (struct ServerCtxS));

  // create the Message-Queue
  MqErrorCheck (MqCreate (config, argv, alfa, msgqueP));
  msgque = *msgqueP;

  if (MqConfigGetIsSlave(msgque)) {
    // add "slave" services here
  } else {
    // add "master" services here
    MqErrorCheck (MqServiceCreate (msgque, "CSV1", Ot_CSV1, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ERR1", Ot_ERR1, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ERR2", Ot_ERR2, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ERR3", Ot_ERR3, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOU", Ot_ECOU, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECON", Ot_ECON, NULL, NULL));
    // START-C-SERVICE-DEFINITION
    MqErrorCheck (MqServiceCreate (msgque, "ECOI", Ot_ECOI, NULL, NULL));
    // END-C-SERVICE-DEFINITION
    MqErrorCheck (MqServiceCreate (msgque, "ECOY", Ot_ECOY, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOO", Ot_ECOO, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOS", Ot_ECOS, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOW", Ot_ECOW, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOD", Ot_ECOD, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOF", Ot_ECOF, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOC", Ot_ECOC, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOB", Ot_ECOB, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECLI", Ot_ECLI, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOL", Ot_ECOL, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECOM", Ot_ECOM, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "ECUL", Ot_ECUL, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "GETU", Ot_GETU, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "GTCX", Ot_GTCX, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "GTTO", Ot_GTTO, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "GTSI", Ot_GTSI, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "MSQT", Ot_MSQT, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "RDUL", Ot_RDUL, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "REDI", Ot_REDI, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "SETU", Ot_SETU, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "SLEP", Ot_SLEP, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "SND1", Ot_SND1, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "SND2", Ot_SND2, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "USLP", Ot_USLP, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "BUF1", Ot_BUF1, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "BUF2", Ot_BUF2, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "BUF3", Ot_BUF3, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "LST1", Ot_LST1, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "LST2", Ot_LST2, NULL, NULL));
    MqErrorCheck (MqServiceCreate (msgque, "INIT", Ot_INIT, NULL, NULL));
  }

  // everything is OK
  return MQ_OK;

error:
  return MqErrorStack((*msgqueP)->error);
}

/// \tool_Create
enum MqErrorE
ServerParentCreate (
  struct MqConfigS  * config,
  struct MqBufferLS * argv,
  struct MqBufferLS * alfa,
  struct MqS       ** msgqueP
)
{
  // add config data
  config->fChildCreate  = ServerChildCreate;
  config->fChildDelete  = ServerDelete;
  config->fParentCreate = ServerParentCreate;
  config->fParentDelete = ServerDelete;
  config->fHelp		= ServerHelp;
  config->isServer	= MQ_YES;

  // create the server-context
  MqErrorCheck (ServerChildCreate(config, argv, alfa, msgqueP));

  // check for wrong arguments
  MqErrorCheck (MqCheckForLeftOverArguments(*msgqueP, &argv));

  // everything is OK
  return MQ_OK;

error:
  return MqErrorStack((*msgqueP)->error);
}

/** \} server */

